// A class for euclidean codes

// (C) 1997-1998 Frank Paehlke <frank@paehlke.de>
// Created: 01/17/1997
// $Id: code.cc,v 1.1 2003/10/14 08:49:02 frank Exp $

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// $Log: code.cc,v $
// Revision 1.1  2003/10/14 08:49:02  frank
// moved from . to src/
//
// Revision 1.4  2003/10/14 08:41:58  frank
// ANSI C++ / gcc 3.x compatibility changes
//
// Revision 1.3  2002/08/02 14:45:06  paehlke
// added GPL comment
//
// Revision 1.2  1998/05/28 12:56:08  paehlke
// Final version for masters thesis
//
// Revision 1.1  1997/06/20 12:44:44  frank
// Initial revision
//

#ifndef _PSKCODE_H
#include "pskcode.h"
#endif

#ifndef _CODE_H
#include "code.h"
#endif

using std::endl;

Code::Code (istream & is)
{
  int i;
  is >> m >> n;
  assert (m > 0 && n > 0);
  cc = new double [mn = m*n];
  cv = new Vector [m];
  for (i=0; i<m; i++) {
    Vector v (n, &cc[i*n]);
    cv[i] = v;
  }
  for (i=0; i<mn; i++)
    is >> cc [i];
}

Code::Code (const PSKCode & psk)
{
  int i, j;
  double s = sqrt (1.0/psk.n);
  n = 2 * psk.n;
  m = psk.m;
  cc = new double [mn = m*n];
  cv = new Vector [m];
  for (i=0; i<m; i++) {
    Vector v (n, &cc[i*n]);
    cv[i] = v;
  }
  for (i=0; i<m; i++) {
    Vector & cvi = cv[i];
    const int * pvi = psk.cv[i];
    for (j=0; j<psk.n; j++) {
      double angle = 2.0 * M_PI * (pvi[j]+.5) / psk.quant;
      cvi[2*j] = s * cos (angle);
      cvi[2*j+1] = s * sin (angle);
    }
  }
}

Code & Code::operator = (const Code & code)
{
  int i;
  if (m != code.m || n != code.n) {
    if (cc != NULL) {
      delete [] cc;
      delete [] cv;
    }
    m = code.m;
    n = code.n;
    cc = new double [mn = m*n];
    cv = new Vector [m];
    for (i=0; i<m; i++) {
      Vector v (n, &cc[i*n]);
      cv[i] = v;
    }
  }
  for (i=0; i<mn; i++)
    cc[i] = code.cc[i];
  return *this;
}

double Code::MinSqrDist () const
{
  int i1, i2;
  double md = HUGE_VAL;
  for (i1=0; i1<m; i1++)
    for (i2=i1+1; i2<m; i2++) {
      double d = sqrdist (cv[i1], cv[i2]);
      if (d <= md)
      	md = d;
    }
  return md;
}

Code & Code::Quantize (int quant)
{
  assert ((quant&1) == 0); // quant must be even
  int i;
  quant /= 2;
  // Distance of quantized values
  double qu = sqrt (MaxEnergy ()) / quant;
  for (i=0; i<mn; i++) {
    int ival = (int) floor (cc[i]/qu);
    if (ival >= quant) ival = quant-1;
    else if (ival < -quant) ival = -quant;
    cc[i] = qu * (ival + .5);
  }
  return *this;
}

double Code::Decode3 (const Vector & v, int & i1, int & i2, int & i3) const
{
  int i;
  double md1, md2, md3;
  i1 = i2 = i3 = -1;
  md1 = md2 = md3 = HUGE_VAL;
  for (i=0; i<m; i++) {
    double d = sqrdist (v, cv[i]);
    if (d < md3) {
      if (d < md2) {
	i3 = i2; md3 = md2;
	if (d < md1) {
	  i2 = i1; md2 = md1;
	  i1 = i; md1 = d;
	}
	else { // md1 <= d < md2
	  i2 = i; md2 = d;
	}
      }
      else { // md2 <= d < md3
	i3 = i; md3 = d;
      }
    }
  }
  return md2 - md1;
}

double Code::Decode3Spherical
(const Vector & v, int & i1, int & i2, int & i3) const
{
  int i;
  double md1, md2, md3;
  double sav = sqrabs (v);
  i1 = i2 = i3 = -1;
  md1 = md2 = md3 = HUGE_VAL;
  for (i=0; i<m; i++) {
    double s = v * cv[i];
    if (s > 0) {
      double d = 1.0 - s * s / (sav * sqrabs(cv[i]));
      if (d < md3) {
	if (d < md2) {
	  i3 = i2; md3 = md2;
	  if (d < md1) {
	    i2 = i1; md2 = md1;
	    i1 = i; md1 = d;
	  }
	  else { // md1 <= d < md2
	    i2 = i; md2 = d;
	  }
	}
	else { // md2 <= d < md3
	  i3 = i; md3 = d;
	}
      }
    }
  }
  return md2 - md1;
}

double Code::Decode3APP (double var, const Vector & v,
			 int & i1, int & i2, int & i3) const
{
  int i;
  double mp1, mp2, mp3;
  i1 = i2 = i3 = -1;
  mp1 = mp2 = mp3 = 0.0;
  double psum = 0.0;
  for (i=0; i<m; i++) {
    double p = exp (-sqrdist (v, cv[i])/(2*var));
    psum += p;
    if (p > mp3) {
      if (p > mp2) {
	i3 = i2; mp3 = mp2;
	if (p > mp1) {
	  i2 = i1; mp2 = mp1;
	  i1 = i; mp1 = p;
	}
	else { // mp1 >= p > mp2
	  i2 = i; mp2 = p;
	}
      }
      else { // mp2 >= p > mp3
	i3 = i; mp3 = p;
      }
    }
  }
  return log (mp1 / mp2) / M_LN2;
}

double Code::LLHDecode (double var, const Vector & v, int & mdi,
			double * alpha) const
{
  int i;
  // first compute alpha[i] = ln (m * P[sent == cv[i] && rcvd == v])
  mdi = 0;
  double maxalpha = alpha[0] = -sqrdist (v, cv[0]) / (2*var);
  double expsum = exp(alpha[0]);
  for (i = 1; i < m; i++) {
    alpha[i] = -sqrdist (v, cv[i]) / (2*var);
    expsum += exp(alpha[i]);
    if (alpha[i] > maxalpha) {
      maxalpha = alpha[i];
      mdi = i;
    }
  }
  double lp_recv = log(expsum); // lp_recv == ln (m * P[rcvd == v])
  double beta = HUGE_VAL;
  maxalpha = (maxalpha - lp_recv) / M_LN2;
  for (i = 0; i < m; i++) {
    alpha[i] = (alpha[i] - lp_recv) / M_LN2;
    double beta_i = maxalpha - alpha[i];
    if (i != mdi && beta_i < beta)
      beta = beta_i;
  }
  assert (beta >= 0);
  return beta;
}

istream & operator >> (istream & is, Code & c)
{
  int m,n,i;
  is >> m >> n;
  assert (m > 0 && n > 0);
  if (m != c.m || n != c.n) {
    delete [] c.cc;
    delete [] c.cv;
    c.m = m;
    c.n = n;
    c.cc = new double [c.mn = m*n];
    c.cv = new Vector [m];
    for (i=0; i<m; i++) {
      Vector v (n, &c.cc[i*n]);
      c.cv[i] = v;
    }
  }
  for (i=0; i<c.mn; i++)
    is >> c.cc [i];
  return is;
}

ostream & operator << (ostream & os, const Code & c)
{
  int i,j;
  os << c.m << ' ' << c.n << endl;
  for (i=0; i<c.m; i++) {
    Vector cvi = c.cv[i];
    for (j=0; j<c.n; j++)
      os << cvi[j] << ' ';
    os << endl;
  }
  return os;
}
