#ifndef _CODE_H
#define _CODE_H

// A class for euclidean codes

// (C) 1997-1998 Frank Paehlke <frank@paehlke.de>
// Created: 01/17/1997
// $Id: code.h,v 1.1 2003/10/14 08:49:02 frank Exp $

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// $Log: code.h,v $
// Revision 1.1  2003/10/14 08:49:02  frank
// moved from . to src/
//
// Revision 1.4  2003/10/14 08:41:58  frank
// ANSI C++ / gcc 3.x compatibility changes
//
// Revision 1.3  2002/08/02 14:45:06  paehlke
// added GPL comment
//
// Revision 1.2  1998/05/28 12:56:41  paehlke
// Final version for masters thesis
//
// Revision 1.1  1997/06/20 12:44:44  frank
// Initial revision
//

#ifndef _ASSERT_H
#include <assert.h>
#endif

#ifndef _CPP_IOSTREAM
#include <iostream>
#endif

#ifndef _MATH_H
#include <math.h>
#endif

#ifndef _VECTOR_H
#include "vector.h"
#endif

#ifndef _PSKCODE_H
class PSKCode;
#endif

// --- BEGIN OF PUBLIC INTERFACE ---

class Code {
public:

  Code ();
  Code (int _m, int _n);
  Code (const Code &);
  Code (std::istream &);
  Code (const PSKCode &);

  Code & operator = (const Code &);

  int get_m () const;
  int get_n () const;
  double get_r () const;
  Vector & operator [] (int);
  const Vector & operator [] (int) const;

  double AvgEnergy () const;
  double MaxEnergy () const;
  double MinSqrDist () const;
  double MinDist () const;

  // Quantize (q) quantizes the code into a QAM configuration
  // using q equidistant values per coordinate

  Code & Quantize (int);

  // Decode3(v,i1,i2,i3) sets (i1,i2,i3) to the indices of those three
  // code vectors which are closest to v,
  // with d(v,cv[i1]) <= d(v,cv[i2]) <= d(v,cv[i3]).
  // The distances are computed as squared euclidean distances:
  // d(x,y) = sum ((x[i] - y[i])^2, i = 0..get_n()-1).
  // The difference d(v,cv[i2]) - d(v,cv[i3]) is returned.

  double Decode3 (const Vector &, int &, int &, int &) const;

  // Decode(v,i1) works exactly like Decode3(), except for that
  // it only sets the index i1.

  double Decode (const Vector &, int &) const;

  // Decode3Spherical() and Decode3Spherical() are similar to
  // Decode3() and Decode(), but use some kind of spherical distance:
  // d(x,y) = 1 - (x*y)^2 / ((x*x) * (y*y)),
  // where '*' denotes the dot product.
  // (Note: d(x,y) = sin(phi)^2, where phi is the angle
  // between the vectors x and y.

  double DecodeSpherical (const Vector &, int &) const;
  double Decode3Spherical (const Vector &, int &, int &, int &) const;

  double DecodeAPP (double, const Vector &, int &) const;
  double Decode3APP (double, const Vector &, int &, int &, int &) const;

  // LLHDecode (var,v,mdi,alpha) sets alpha[i] (i = 0..m-1)
  // to the logarithmic likelihood log (P[sent = cv[i] | rcvd = v]),
  // where var is the noise variance per coordinate.
  // mdi is set to the index i with the highest value of alpha[i],
  // and the difference between alpha[mdi] and the next highest alpha value
  // is returned.

  double LLHDecode (double, const Vector &, int &, double *) const;

  Code & operator *= (double);

  friend std::istream & operator >> (std::istream &, Code &);
  friend std::ostream & operator << (std::ostream &, const Code &);

// --- END OF PUBLIC INTERFACE ---

protected:

  Vector * cv;
  double * cc;
  int m,n,mn;
};


inline Code::Code ()
{
  cv = NULL;
  cc = NULL;
  m = n = mn = 0;
}

inline Code::Code (int _m, int _n)
{
  int i;
  assert (_m > 0 && _n > 0);
  m = _m;
  n = _n;
  cc = new double [mn = m*n];
  cv = new Vector [m];
  for (i=0; i<m; i++) {
    Vector v (n, &cc[i*n]);
    cv[i] = v;
  }
  for (i=0; i<mn; i++)
    cc[i] = 0;
}

inline Code::Code (const Code & code)
{
  int i;
  m = code.m;
  n = code.n;
  cc = new double [mn = m*n];
  cv = new Vector [m];
  for (i=0; i<m; i++) {
    Vector v (n, &cc[i*n]);
    cv[i] = v;
  }
  for (i=0; i<mn; i++)
    cc[i] = code.cc[i];
}

inline int Code::get_m () const
{
  return m;
}

inline int Code::get_n () const
{
  return n;
}

inline double Code::get_r () const
{
  return log(m) / (M_LN2 * n);
}

inline Vector & Code::operator [] (int i)
{
  return cv[i];
}

inline const Vector & Code::operator [] (int i) const
{
  return cv[i];
}

inline double Code::AvgEnergy () const
{
  int i;
  double ae = 0.0;
  for (i=0; i<mn; i++)
    ae += cc[i] * cc[i];
  return ae / m;
}

inline double Code::MaxEnergy () const
{
  int i;
  double me = 0.0;
  for (i=0; i<m; i++) {
    double e = sqrabs (cv[i]);
    if (e > me)
      me = e;
  }
  return me;
}

inline double Code::MinDist () const
{
  return sqrt (MinSqrDist ());
}

inline Code & Code::operator *= (double d)
{
  int i;
  for (i=0; i<mn; i++)
    cc[i] *= d;
  return *this;
}

inline double Code::Decode (const Vector & v, int & mdi) const
{
  int md2, md3;
  return Decode3 (v, mdi, md2, md3);
}

inline double Code::DecodeSpherical (const Vector & v, int & mdi) const
{
  int md2, md3;
  return Decode3Spherical (v, mdi, md2, md3);
}

inline double Code::DecodeAPP (double var, const Vector & v, int & mdi) const
{
  int md2, md3;
  return Decode3APP (var, v, mdi, md2, md3);
}

#endif
