#ifndef _DISTCODE_H
#define _DISTCODE_H

// A class for euclidean codes
// suited for minimal distance optimization

// (C) 1997-1998 Frank Paehlke <frank@paehlke.de>
// Created: 01/19/1997
// $Id: distcode.h,v 1.1 2003/10/14 08:49:02 frank Exp $

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// $Log: distcode.h,v $
// Revision 1.1  2003/10/14 08:49:02  frank
// moved from . to src/
//
// Revision 1.3  2002/08/02 14:45:06  paehlke
// added GPL comment
//
// Revision 1.2  1998/05/28 12:56:41  paehlke
// Final version for masters thesis
//
// Revision 1.1  1997/06/20 12:52:50  frank
// Initial revision
//

#ifndef _CODE_H
#include "code.h"
#endif

#ifndef _HEAP_H
#include "heap.h"
#endif

// --- BEGIN OF PUBLIC INTERFACE ---

class DistCode : public Code
{
public:

  DistCode (const Code &);

  // OptValue() returns the minimal distance between any 2 code vectors
  double OptValue ();

  // Do a change with extent delta
  DistCode & DoChange (double delta);

  // Undo the previous change
  DistCode & UndoChange ();

  // Do a small change
  DistCode & SteepestAscent (double);

  // --- END OF PUBLIC INTERFACE ---

protected:

  Heap<double> * DistanceHeap;

  // The distance between cv[i] and cv[j] where i > j is held
  // in DistanceEl (i-1)*i/2+j
  HeapElement<double> * DistanceEl;

  // for UndoChange() in order to work correctly:
  int k_moved;
  Vector cvk_old;

  DistCode & ComputeDistances ();
  DistCode & NotifyChange (int);
};

// ComputeDistances () is used to set up the distance heap correctly

inline DistCode & DistCode::ComputeDistances ()
{
  int i1,i2;
  for (i1=0; i1<m; i1++)
    for (i2=0; i2<i1; i2++)
      DistanceEl [(i1-1)*i1/2+i2].Data = sqrdist (cv[i1], cv[i2]);
  return *this;
}

// NotifyChange(k) is used to update the distance heap
// when cv[k] has been changed

inline DistCode & DistCode::NotifyChange (int k)
{
  int i;
  Vector & cvk = cv[k];
  // recalculate the distances between cv[k] and all other code vectors
  for (i=0; i<k; i++)
  {
    HeapElement<double> & he = DistanceEl [(k-1)*k/2+i];
    double d_old = he.Data;
    he.Data = sqrdist (cvk, cv[i]);
    if (he.Data < d_old)
      DistanceHeap->RiseUp (he);
    else if (he.Data > d_old)
      DistanceHeap->SinkDown (he);
  }
  for (i=k+1; i<m; i++)
  {
    HeapElement<double> & he = DistanceEl [(i-1)*i/2+k];
    double d_old = he.Data;
    he.Data = sqrdist (cvk, cv[i]);
    if (he.Data < d_old)
      DistanceHeap->RiseUp (he);
    else if (he.Data > d_old)
      DistanceHeap->SinkDown (he);
  }
  return *this;
}

inline DistCode::DistCode (const Code & c) : Code (c)
{
  // Initialize Distance Heap
  DistanceEl = new HeapElement<double> [(m-1)*m/2];
  ComputeDistances ();
  DistanceHeap = new Heap<double> (DistanceEl, (m-1)*m/2);
}

inline double DistCode::OptValue ()
{
  return DistanceHeap->SmallestElement().Data;
}

// DoChange(delta) alters every coordinate of a randomly chosen code vector
// by at most delta in any direction, normalizes the vector and rearranges
// the distance heap

inline DistCode & DistCode::DoChange (double delta)
{
  int i;
  int k = 1 + (int) floor ((m-1) * (double) rand() / (RAND_MAX+1.0));
  Vector & cvk = cv[k];
  // save the old vector for undoing
  k_moved = k;
  cvk_old.copy (cvk);
  // alter cv[k]
  for (i=0; i<n; i++)
    cvk[i] += 2 * delta * (rand() / (RAND_MAX + 1.0) - 0.5);
  // normalize cv[k]
  cvk /= abs (cvk);
  // readjust the distance heap
  NotifyChange (k);
  return *this;
}

inline DistCode & DistCode::UndoChange ()
{
  Vector & cvk = cv[k_moved];
  cvk.copy (cvk_old);
  NotifyChange (k_moved);
  return *this;
}

inline DistCode & DistCode::SteepestAscent (double epsilon)
{
  double old_ov = OptValue ();
  DoChange (epsilon);
  if (OptValue() <= old_ov)
    UndoChange ();
  return *this;
}

#endif
