#ifndef _DISTPSKCODE_H
#define _DISTPSKCODE_H

// A class for MPSK codes
// suited for minimal distance optimization

// (C) 1997-1998 Frank Paehlke <frank@paehlke.de>
// Created: 01/21/1997
// $Id: distpskcode.h,v 1.1 2003/10/14 08:49:02 frank Exp $

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// $Log: distpskcode.h,v $
// Revision 1.1  2003/10/14 08:49:02  frank
// moved from . to src/
//
// Revision 1.5  2002/08/02 14:45:06  paehlke
// added GPL comment
//
// Revision 1.4  1998/05/28 12:56:42  paehlke
// Final version for masters thesis
//
// Revision 1.3  1997/09/23 12:58:15  frank
// minor changes
//
// Revision 1.2  1997/06/20 14:11:01  frank
// fixed some bugs
//
// Revision 1.1  1997/06/20 12:53:20  frank
// Initial revision
//

#ifndef _PSKCODE_H
#include "pskcode.h"
#endif

#ifndef _HEAP_H
#include "heap.h"
#endif

#ifndef _FIXED_H
#include "fixed.h"
#endif

// --- BEGIN OF PUBLIC INTERFACE ---

class DistPSKCode : public PSKCode
{
public:

  DistPSKCode (const PSKCode &);

  // OptValue() returns the minimal distance between any 2 code vectors
  double OptValue ();

  // Do a change with extent delta
  DistPSKCode & DoChange (double delta);

  // Undo the previous change
  DistPSKCode & UndoChange ();

  // Do a small change
  DistPSKCode & SteepestAscent (double epsilon);

  // --- END OF PUBLIC INTERFACE ---

protected:

  Heap<Fixed> * DistanceHeap;

  // The distance between cv[i] and cv[j] where i > j is held
  // in DistanceEl (i-1)*i/2+j
  // All distance calculations are done with fixed point arithmetic
  // in order to prevent rounding errors

  HeapElement<Fixed> * DistanceEl;

  // for UndoChange() in order to work correctly:
  int k_moved, c_moved, cvkc_old;

  Fixed SqrDist (const int *, const int *);

  DistPSKCode & ComputeDistances ();
  DistPSKCode & NotifyChange (int, int);

  // The sine function is implemented by table lookup for speed reasons.
  // sine_table[i] holds 1/n*(2*sin(2*PI*(i/quant)/2))^2

  Fixed * SineTable;

  void InitSineTable ();
};

void DistPSKCode::InitSineTable ()
{
  int i;
  SineTable = new Fixed [quant];
  for (i=0; i<quant; i++) {
    double s = sin (i*M_PI/quant);
    SineTable[i] = Fixed (4.0*s*s/n);
  }
}

// SqrDist(v1,v2) calculates the squared distance between
// code vectors v1 and v2

Fixed DistPSKCode::SqrDist (const int * v1, const int * v2)
{
  int j;
  Fixed d = 0;
  for (j=0; j<n; j++)
  {
    int dj = v1[j] - v2[j];
    if (dj < 0)
      dj += quant;
    d += SineTable [dj];
  }
  return d;
}

// ComputeDistances() calculates the squared distances between any two
// points in p[] and stores the results in DistanceEl[]

inline DistPSKCode & DistPSKCode::ComputeDistances ()
{
  int i1,i2;
  for (i1=0; i1<m; i1++)
    for (i2=0; i2<i1; i2++)
      DistanceEl [(i1-1)*i1/2+i2].Data = SqrDist (cv[i1], cv[i2]);
  return *this;
}

// NotifyChange(k,c) is used to update the distance heap
// when cv[k][c] has been changed

inline DistPSKCode & DistPSKCode::NotifyChange (int k, int c)
{
  int i;
  int * cvk = cv[k];
  // recalculate the distances between cv[k] and all other code vectors
  for (i=0; i<k; i++) {
    HeapElement<Fixed> & he = DistanceEl [(k-1)*k/2+i];
    int cvic = cv[i][c];
    int dcold = cvkc_old - cvic;
    if (dcold < 0)
      dcold += quant;
    int dcnew = cvk[c] - cvic;
    if (dcnew < 0)
      dcnew += quant;
    Fixed d = SineTable[dcnew] - SineTable[dcold];
    he.Data += d;
    if (d < Fixed(0))
      DistanceHeap->RiseUp (he);
    else if (d > Fixed(0))
      DistanceHeap->SinkDown (he);
  }
  for (i=k+1; i<m; i++) {
    HeapElement<Fixed> & he = DistanceEl [(i-1)*i/2+k];
    int cvic = cv[i][c];
    int dcold = cvkc_old - cvic;
    if (dcold < 0)
      dcold += quant;
    int dcnew = cvk[c] - cvic;
    if (dcnew < 0)
      dcnew += quant;
    Fixed d = SineTable[dcnew] - SineTable[dcold];
    he.Data += d;
    if (d < Fixed(0))
      DistanceHeap->RiseUp (he);
    else if (d > Fixed(0))
      DistanceHeap->SinkDown (he);
  }
  return *this;
}

inline DistPSKCode::DistPSKCode (const PSKCode & c) : PSKCode (c)
{
  InitSineTable();
  // Initialize distance heap
  DistanceEl = new HeapElement<Fixed> [(m-1)*m/2];
  ComputeDistances ();
  DistanceHeap = new Heap<Fixed> (DistanceEl, (m-1)*m/2);
}

inline double DistPSKCode::OptValue ()
{
  return DistanceHeap->SmallestElement().Data;
}

// DoChange(delta) alters every coordinate of a randomly chosen code vector
// by at most delta*quant in any direction, normalizes the vector and
// rearranges the distance heap

inline DistPSKCode & DistPSKCode::DoChange (double delta)
{
  double dq = delta * quant;
  assert (dq + 0.5 < quant);
  if (dq < 1)
    dq = 1;
  // Determine which code vector to alter [1 .. m-1]
  int k = 1 + (int) floor ((m-1) * (double) rand() / (RAND_MAX+1.0));
  int * cvk = cv[k];
  // Determine which coordinate to alter [0 .. n-1]
  int c = (int) floor (((double) n) * rand() / (RAND_MAX + 1.0));
  // Determine the altering distance [-dq .. dq]
  int dist;
  do {
    dist = (int) floor ((2*dq+1) * (rand() / (RAND_MAX+1.0) - .5) + .5);
  } while (dist == 0);
  // save the old values for undoing
  k_moved = k;
  c_moved = c;
  cvkc_old = cvk[c];
  // alter cv[k][c]
  cvk[c] += dist;
  if (cvk[c] < 0)
    cvk[c] += quant;
  else if (cvk[c] >= quant)
    cvk[c] -= quant;
  // readjust the distance heap
  NotifyChange (k,c);
  return *this;
}

inline DistPSKCode & DistPSKCode::UndoChange ()
{
  int * cvk = cv[k_moved];
  int h = cvk[c_moved];
  cvk[c_moved] = cvkc_old;
  cvkc_old = h;
  NotifyChange (k_moved, c_moved);
  return *this;
}

inline DistPSKCode & DistPSKCode::SteepestAscent (double epsilon)
{
  double old_ov = OptValue ();
  DoChange (epsilon);
  if (OptValue() <= old_ov)
    UndoChange ();
  return *this;
}

#endif
