#ifndef _FUNC1CODE_H
#define _FUNC1CODE_H

// A class for euclidean codes
// suited for optimization of the function
// sum (-exp(-sqrdist(cv[i],cv[j])), i=0..m-1, j=0..m-1, i!=j);

// (C) 1997-1998 Frank Paehlke <frank@paehlke.de>
// Created: 01/19/1997
// $Id: func1code.h,v 1.1 2003/10/14 08:49:02 frank Exp $

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// $Log: func1code.h,v $
// Revision 1.1  2003/10/14 08:49:02  frank
// moved from . to src/
//
// Revision 1.3  2002/08/02 14:45:06  paehlke
// added GPL comment
//
// Revision 1.2  1998/05/28 12:56:43  paehlke
// Final version for masters thesis
//
// Revision 1.1  1997/06/20 12:52:37  frank
// Initial revision
//

#ifndef _MATH_H
#include <math.h>
#endif

#ifndef _CODE_H
#include "code.h"
#endif

#ifndef _FIXED_H
#include "fixed.h"
#endif

const int func1tablequant = 1024;

const double expfactor = 4.0;

// --- BEGIN OF PUBLIC INTERFACE ---

class Func1Code : public Code
{
public:

  Func1Code (const Code &);

  // OptValue() returns the minimal distance between any 2 code vectors
  double OptValue ();

  // Do a change with extent delta
  Func1Code & DoChange (double delta);

  // Undo the previous change
  Func1Code & UndoChange ();

  // Try a steepest descent optimization

  Func1Code & SteepestAscent (double);

  // --- END OF PUBLIC INTERFACE ---

protected:

  // The current value of func1=sum(-exp(-expfactord[i,j]^2))
  // is stored as fixed point to avoid rounding errors

  Fixed func1;

  // for UndoChange() in order to work correctly:
  int k_moved;
  Vector cvk_old;

  Func1Code & ComputeDistances ();

  Func1Code & NotifyChange (int);

  Fixed func1table [4*func1tablequant+1];

  // DistVector[i*(i-1)/2+j] where i>j holds the distance vector cv[i]-cv[j]
  Vector * DistVector;

  // Func1Dist[i] holds the func1 value of DistVector[i]
  Fixed * Func1Dist;
};

// ComputeDistances() is used to set up the DistVector[] and Func1Dist[]
// arrays and func1 correctly

inline Func1Code & Func1Code::ComputeDistances ()
{
  int i1,i2;
  func1 = 0;
  for (i1=1; i1<m; i1++) {
    Vector & cvi1 = cv[i1];
    for (i2=0; i2<i1; i2++) {
      int j = i1*(i1-1)/2 + i2; // index in DistVector[] and Func1Dist[]
      DistVector[j].copy (cvi1) -= cv[i2];
      int index = (int) floor (sqrabs (DistVector[j]) * func1tablequant);
      Func1Dist[j] = func1table [index];
      func1 += Func1Dist[j];
    }
  }
  return *this;
}

// NotifyChange(k) updates DistVector[], Func1Dist[] and func1
// when vector k has been changed

inline Func1Code & Func1Code::NotifyChange (int k)
{
  int i;
  Vector & cvk = cv[k];
  for (i=0; i<k; i++) {
    int j = k*(k-1)/2+i; // index in DistVector[] and Func1Dist[]
    // Calculate new distance vector
    DistVector[j].copy (cvk) -= cv[i];
    // Update func1 and Func1Dist[]
    func1 -= Func1Dist[j];
    int index = (int) floor (sqrabs (DistVector[j]) * func1tablequant);
    Func1Dist[j] = func1table [index];
    func1 += Func1Dist[j];
  }
  for (i=k+1; i<m; i++) {
    int j = i*(i-1)/2+k; // index in DistVector[] and Func1Dist[]
    // Calculate new distance vector
    DistVector[j].copy (cv[i]) -= cvk;
    // Update func1 and Func1Dist[]
    func1 -= Func1Dist[j];
    int index = (int) floor (sqrabs (DistVector[j]) * func1tablequant);
    Func1Dist[j] = func1table [index];
    func1 += Func1Dist[j];
  }
  return *this;
}

inline Func1Code::Func1Code (const Code & c) : Code (c)
{
  int i;
  // Setup the func1 table
  for (i=0; i < 4*func1tablequant+1; i++) {
    double d2 = ((double) i) / func1tablequant;
    func1table[i] = Fixed (-exp(-expfactor*d2));
  }
  // Setup the DistVector[] and Func1Dist[] arrays
  int m2 = m*(m-1)/2;
  DistVector = new Vector [m2];
  Func1Dist = new Fixed [m2];
  ComputeDistances ();
}

inline double Func1Code::OptValue ()
{
  return (double) func1;
}

// DoChange(delta) alters every coordinate of a randomly chosen code vector
// by at most delta in any direction, normalizes the vector and rearranges
// the distance heap

inline Func1Code & Func1Code::DoChange (double delta)
{
  int i;
  int k = 1 + (int) floor ((m-1) * (double) rand() / (RAND_MAX+1.0));
  Vector & cvk = cv[k];
  // save the old vector for undoing
  k_moved = k;
  cvk_old.copy (cvk);
  // alter cv[k]
  for (i=0; i<n; i++)
    cvk[i] += 2 * delta * (rand() / (RAND_MAX + 1.0) - 0.5);
  // normalize cv[k]
  cvk /= abs (cvk);
  // Recalculate the distances
  NotifyChange (k);
  return *this;
}

inline Func1Code & Func1Code::UndoChange ()
{
  Vector & cvk = cv[k_moved];
  cvk.copy (cvk_old);
  NotifyChange (k_moved);
  return *this;
}

// func1 = -exp(-expfactor*sqrdist)
// has the derivative (D func1) / (D sqrdist) = -expfactor*func1
// steepest ascent optimization:
// cv[i] -> cv[i] - epsilon * 2 * dist * (-expfactor*func1)

inline Func1Code & Func1Code::SteepestAscent (double epsilon)
{
  int i1,i2;
  // Scale the DistVectors in such a way that the steepest ascent step
  // can be done simply by adding them to the code vectors
  for (i1=1; i1<m; i1++)
    for (i2=0; i2<i1; i2++) {
      int j = i1*(i1-1)/2+i2; // index in DistVector[] and Func1Dist[]
      DistVector [j] *= -epsilon * expfactor * (double) Func1Dist[j];
    }
  // Calculate the new code vectors
  for (i1=0; i1<m; i1++) {
    Vector & cvi1 = cv[i1];
    for (i2=0; i2<i1; i2++) {
      int j = i1*(i1-1)/2+i2; // index in DistVector[] and Func1Dist[]
      cvi1 += DistVector[j];
    }
    for (i2=i1+1; i2<m; i2++) {
      int j = i2*(i2-1)/2+i1; // index in DistVector[] and Func1Dist[]
      cvi1 -= DistVector[j];
    }
    // Normalize cv[i1]
    cvi1 /= abs (cvi1);
  }
  // Update DistVector[], Func1Dist[] and func1
  ComputeDistances ();
  return *this;
}

#endif
