#ifndef _GBCHCODE_H
#define _GBCHCODE_H

// A class for generalized BCH encoding and decoding
// where the code symbols come from a Galois field GF = GF(q),
// but where the coefficients of the parity check matrix
// are from the extension field EXT = GF(q^m).
// The code word length is n <= q^m - 1.

// (C) 1997-1998 Frank Paehlke <frank@paehlke.de>
// Created: 03/19/1997
// $Id: gbchcode.h,v 1.1 2003/10/14 08:49:02 frank Exp $

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// $Log: gbchcode.h,v $
// Revision 1.1  2003/10/14 08:49:02  frank
// moved from . to src/
//
// Revision 1.5  2002/08/02 14:45:06  paehlke
// added GPL comment
//
// Revision 1.4  1998/05/28 12:56:44  paehlke
// Final version for masters thesis
//
// Revision 1.3  1998/01/29 09:34:09  paehlke
// added support for sequential decoding
//
// Revision 1.2  1997/10/19 16:45:28  frank
// general cleanup and simplification
// second template <class EXT> instead of parameter "bits"
// arguments of Encode() and Decode(): (GF *) instead of (UBYTE *)
// added support for shortened GBCH codes
//
// Revision 1.1  1997/06/20 18:41:39  frank
// Initial revision
//

#ifndef _ASSERT_H
#include <assert.h>
#endif

#ifndef _RSCODE_H
#include "rscode.h"
#endif

// --- BEGIN OF PUBLIC INTERFACE

template<class GF,class EXT> class GBCHCode
  : private RSCode<EXT>, RSCode<GF>
{

public:

  // constructor and destructor

  // GBCHCode(d,n) creates a generalized BCH code of length n
  // with code distance d
  // and information word length k <= n+1 - d.
  // which can correct up to t = (d-1) / 2 symbol errors.
  // GBCHCode(d) is equivalent to GBCHCode(d,EXT::q()-1)

  GBCHCode (int d, int n);
  GBCHCode (int d);
  ~GBCHCode ();

  // code parameters

  int get_t () const; // get max. # of correctable errors
  int get_d () const; // get code distance
  int get_k () const; // get information word length
  int get_n () const; // get code word length
  double get_r () const; // get code rate
  int get_q () const; // get cardinality of underlying GF
  int get_qm () const; // get cardinality of EXT
  int get_deggenpoly () const; // get degree of generator polynomial
  const GF * get_gfgenpoly () const; // get coeffs of generator polynomial
  const EXT * get_extgenpoly () const;

  // encoding and decoding

  // When calling Encode(), dest must point to an array of length n,
  // and src must point to an array of length k.
  // src[] and dest[] must not overlap.

  void Encode (GF * dest, const GF * src) const;

  // Correct() tries to correct cv erasures whose locations are given in Iv[],
  // and up to tau errors where 2*tau + cv < d.
  // dest and src must point to arrays of size n, which may overlap.
  // On success, the number of corrected errors is returned.
  // A return value < 0 indicates that src[] contained more errors
  // than could be corrected.

  int Correct (GF * dest, const GF * src, const int * Iv = NULL, int cv = 0);

  // Decode() works like Correct(), but extracts the information word
  // to dest[].
  // dest must point to an array of size k.
  // src[] and dest[] must not overlap.

  int Decode (GF * dest, const GF * src, const int * Iv = NULL, int cv = 0);

  // support for sequential decoding

  typedef RSState<GF> _State;

  // Allocate buffer memory for an object of class State.
  // The deallocation is done by the destructor of class State.

  void AllocateState (_State &) const;

  void InitialState (_State &, int reverse) const;
  int IsTerminalState (const _State &, int reverse) const;

  double BranchMetric (const _State &, const GF &,
		       const double * const * metric,
		       int reverse) const;

  double KnownSymbolBranchMetric (const _State &, int reverse) const;

  void MoveForward (_State &, const _State &, const GF &,
		    int reverse) const;

  void EncodeConvolutional (GF * dest, const GF * src) const;

// --- END OF PUBLIC INTERFACE

protected:

  // conversion tables between GF and EXT

  EXT * GFtoEXT;
  GF * EXTtoGF;

  // iszero[i] != 0 indicates that (GF::gen)^i is a zero
  // of the generator polynomial
  int * iszero;

  void Initialize (int, int);
  void InitConversionTables ();
  void InitGeneratorPolynomials ();
};

// InitConversionTables() initializes the conversion tables GFtoEXT[]
// and EXTtoGF[].

template<class GF,class EXT> void GBCHCode<GF,EXT>::InitConversionTables ()
{
  int i;
  // allocate memory
  GFtoEXT = new EXT [GF::q()];
  EXTtoGF = new GF [EXT::q()];
  // The conversion of GF into EXT is done via the following mapping:
  // 0 -> 0, a^i -> b^i where a is from GF and b is from EXT.
  // For this mapping in order to be an injective homomorphism,
  // a must be a generator of GF, and the minimal polynomial of b over GF(2)
  // must be equal to the minimal polynomial of a over GF(2).
  // b == EXT::gen()^i can only fulfill this condition if
  // i is divisible by s = (EXT::q()-1)/(GF::q()-1).
  int s = (EXT::q()-1) / (GF::q()-1);
  unsigned int minpoly = GF::gen ().minpoly ();
  for (i = 0; i < EXT::q()-1; i += s) {
    if (EXT::fastexp(i).minpoly () == minpoly)
      break;
  }
  assert (i < EXT::q()-1); // we MUST have found an appropriate b
  EXT b = EXT::fastexp(i);
  GFtoEXT[0] = EXT(0);
  EXTtoGF[0] = GF(0);
  for (i = 1; i < EXT::q()-1; i++)
    EXTtoGF[i] = -1;
  for (i = 0; i < GF::q()-1; i++) {
    GF ai = GF::fastexp(i); // a^i
    EXT bi = b^i;
    assert (!ai.isNaN() && !bi.isNaN());
    GFtoEXT[(int) ai] = bi;
    EXTtoGF[(int) bi] = ai;
  }
}

// InitGeneratorPolynomial() calculates
// genpoly(X) = prod (minpoly (z^i), i = 0..d-2)
// where z = GF::gen,
// and minpoly(a) = lcm (X - a^(expb^j), j = 0..(GF::ldq/b).

template<class GF,class EXT> void
GBCHCode<GF,EXT>::InitGeneratorPolynomials ()
{
  int i,j;
  // First calculate the zeroes of the generator polynomial.
  // For every i=0..d-2, the generator polynomial must have the zeroes
  // z ^ i, z ^ (GF::q() * i), z ^ (GF::q()^2 * i), ...
  iszero = new int [EXT::q()-1];
  for (i = 0; i < EXT::q()-1; i++)
    iszero [i] = 0;
  for (i = 0; i < RSCode<EXT>::d - 1; i++)
    for (j = 0; j < EXT::ldq(); j += GF::ldq())
      iszero [(i * (1<<j)) % (EXT::q()-1)] = 1;
  // now calculate the information word length k = n - deg (genpoly)
  int _n = RSCode<EXT>::n, _k = _n;
  for (i = 0; i < EXT::q() - 1; i++)
    if (iszero[i])
      _k--;
  RSCode<EXT>::k = RSCode<GF>::k = _k;
  EXT * genpoly = RSCode<EXT>::genpoly = new EXT [_n - _k + 1];
  GF * gfgenpoly = RSCode<GF>::genpoly = new GF [_n - _k + 1];
  // set genpoly to 1
  genpoly[0] = 1;
  int deg = 0;
  for (i = 0; i < EXT::q() - 1; i++) {
    if (iszero[i]) {
      // set genpoly = genpoly * (X - EXT::gen()^i)
      EXT zi = EXT::fastexp (i);
      genpoly[deg + 1] = genpoly[deg];
      for (j = deg; j > 0; j--)
	genpoly[j] = genpoly[j-1] - zi * genpoly[j];
      genpoly[0] *= -zi;
      deg++;
    }
  }
  RSCode<GF>::deggenpoly = RSCode<EXT>::deggenpoly = deg;
  // convert genpoly to gfgenpoly
  for (i = 0; i <= deg; i++) {
    gfgenpoly[i] = EXTtoGF[genpoly[i]];
    assert (!gfgenpoly[i].isNaN());
  }
}

template<class GF,class EXT> inline void GBCHCode<GF,EXT>::Initialize
(int _d, int _n)
{
  assert (EXT::ldq() % GF::ldq() == 0);
  RSCode<GF>::Initialize (_d, _n);
  RSCode<EXT>::Initialize (_d, _n);
}

template<class GF,class EXT> inline GBCHCode<GF,EXT>::GBCHCode (int _d, int _n)
{
  assert ((_n > 0) && (_n < EXT::q()) && (_d > 0) && (_d <= _n));
  Initialize (_d, _n);
  InitConversionTables ();
  InitGeneratorPolynomials ();
}

template<class GF,class EXT> inline GBCHCode<GF,EXT>::GBCHCode (int _d)
{
  assert ((_d > 0) && (_d < EXT::q()));
  Initialize (_d, EXT::q()-1);
  InitConversionTables ();
  InitGeneratorPolynomials ();
}

template<class GF,class EXT> GBCHCode<GF,EXT>::~GBCHCode ()
{
  // free all allocated memory
  delete [] EXTtoGF;
  delete [] GFtoEXT;
  delete [] iszero;
}

template<class GF,class EXT> inline int GBCHCode<GF,EXT>::get_t () const
{
  return RSCode<EXT>::get_t();
}

template<class GF,class EXT> inline int GBCHCode<GF,EXT>::get_d () const
{
  return RSCode<EXT>::get_d();
}

template<class GF,class EXT> inline int GBCHCode<GF,EXT>::get_k () const
{
  return RSCode<EXT>::get_k();
}

template<class GF,class EXT> inline int GBCHCode<GF,EXT>::get_n () const
{
  return RSCode<EXT>::get_n();
}

template<class GF,class EXT> inline double GBCHCode<GF,EXT>::get_r () const
{
  return RSCode<EXT>::get_r();
}

template<class GF,class EXT> inline int GBCHCode<GF,EXT>::get_q () const
{
  return GF::q();
}

template<class GF,class EXT> inline int GBCHCode<GF,EXT>::get_qm () const
{
  return EXT::q();
}

template<class GF,class EXT> inline int
GBCHCode<GF,EXT>::get_deggenpoly () const
{
  return RSCode<EXT>::get_deggenpoly();
}


template<class GF,class EXT> inline const GF *
GBCHCode<GF,EXT>::get_gfgenpoly () const
{
  return RSCode<GF>::get_genpoly();
}


template<class GF,class EXT> inline const EXT *
GBCHCode<GF,EXT>::get_extgenpoly () const
{
  return RSCode<EXT>::get_genpoly();
}

// Encode(dest,src) does systematic encoding
// where src[] is copied to dest[deggenpoly..n-1]
// and the redundancy is stored in dest[0..deggenpoly-1]

template<class GF,class EXT> void GBCHCode<GF,EXT>::Encode
(GF * dest, const GF * src) const
{
  RSCode<GF>::Encode (dest, src);
}

template<class GF,class EXT> int GBCHCode<GF,EXT>::Correct
(GF * dest, const GF * src, const int * Iv, int cv)
{
  int i;
  for (i = 0; i < RSCode<EXT>::n; i++) {
    RSCode<GF>::codebuf[i] = src[i];
    RSCode<EXT>::codebuf[i] = GFtoEXT[(int) src[i]];
  }
  int res = RSCode<EXT>::Correct
    (RSCode<EXT>::codebuf, RSCode<EXT>::codebuf, Iv, cv);
  for (i = 0; i < RSCode<EXT>::n; i++) {
    dest[i] = EXTtoGF[RSCode<EXT>::codebuf[i]];
    if (dest[i].isNaN()) { // decoder error, undo all corrections
      int j;
      for (j = 0; j < RSCode<EXT>::n; j++)
	dest[j] = RSCode<GF>::codebuf[j];
      return -10;
    }
  }
  return res;
}

template<class GF,class EXT> inline int GBCHCode<GF,EXT>::Decode
(GF * dest, const GF * src, const int * Iv, int cv)
{
  int i;
  for (i = 0; i < RSCode<EXT>::n; i++)
    RSCode<EXT>::codebuf[i] = GFtoEXT[src[i]];
  int res = RSCode<EXT>::Correct
    (RSCode<EXT>::codebuf, RSCode<EXT>::codebuf, Iv, cv);
  // extract information word
  for (i = 0; i < RSCode<EXT>::k; i++) {
    EXT sym = RSCode<EXT>::codebuf [RSCode<EXT>::deggenpoly + i];
    dest[i] = EXTtoGF[(int) sym];
    if (dest[i].isNaN()) { // decoder error, undo all corrections
      int j;
      for (j = 0; j < RSCode<EXT>::k; j++)
	dest[j] = src[RSCode<EXT>::deggenpoly + j];
      return -10;
    }
  }
  return res;
}

// support for sequential decoding

// encoder state

template<class GF, class EXT> inline void GBCHCode<GF,EXT>::AllocateState
(RSState<GF> & state) const
{
  RSCode<GF>::AllocateState (state);
}

template<class GF, class EXT> inline void GBCHCode<GF,EXT>::InitialState
(RSState<GF> & state, int reverse) const
{
  RSCode<GF>::InitialState (state, reverse);
}

template<class GF, class EXT> inline int GBCHCode<GF,EXT>::IsTerminalState
(const RSState<GF> & state, int reverse) const
{
  return RSCode<GF>::IsTerminalState (state, reverse);
}

template<class GF, class EXT> inline double GBCHCode<GF,EXT>::BranchMetric
(const RSState<GF> & state, const GF & x,
 const double * const * metric, int reverse) const
{
  return RSCode<GF>::BranchMetric (state, x, metric, reverse);
}

template<class GF, class EXT> inline double
GBCHCode<GF,EXT>::KnownSymbolBranchMetric
(const RSState<GF> & state, int reverse) const
{
  return RSCode<GF>::KnownSymbolBranchMetric (state, reverse);
}

template<class GF, class EXT> inline void GBCHCode<GF,EXT>::MoveForward
(RSState<GF> & newstate, const RSState<GF> & oldstate,
 const GF & x, int reverse) const
{
  RSCode<GF>::MoveForward (newstate, oldstate, x, reverse);
}

template<class GF, class EXT> inline void
GBCHCode<GF,EXT>::EncodeConvolutional (GF * dest, const GF * src) const
{
  RSCode<GF>::EncodeConvolutional (dest, src);
}

#endif
