#ifndef _GMDDECODER_H
#define _GMDDECODER_H

// A class for Generalized Minimal Distance (GMD) decoding using the
// improved GMD metric by Dana S. Taipale and Michael B. Pursley
// ("An Improvement to Generalized-Minimum-Distance Decoding",
// IEEE Trans.Inform.Theory vol. 37 no. 5, pp. 167-172, 1991)

// (C) 1997-1998 Frank Paehlke <frank@paehlke.de>
// Created: 03/11/1997
// $Id: gmddecoder.h,v 1.1 2003/10/14 08:49:02 frank Exp $

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// $Log: gmddecoder.h,v $
// Revision 1.1  2003/10/14 08:49:02  frank
// moved from . to src/
//
// Revision 1.4  2002/08/02 14:45:06  paehlke
// added GPL comment
//
// Revision 1.3  1998/05/28 12:57:00  paehlke
// Final version for masters thesis
//
// Revision 1.2  1997/10/18 17:25:42  frank
// arguments changed from UCHAR to GF
//
// Revision 1.1  1997/06/20 18:44:03  frank
// Initial revision
//

#ifndef _ASSERT_H
#include <assert.h>
#endif

#ifndef _QUICKSORT_H
#include "quicksort.h"
#endif

// --- BEGIN OF PUBLIC INTERFACE

// CODE must be an algebraic code over the Galois field GF
// with at least the following member functions:
//   int get_n(), int get_k(), int get_d() to get the code parameters
//   int Correct (GF * dest, const GF * src, const int * Iv, int cv)
//     which does error-and-erasure decoding and returns >=0 if successful

template<class GF, class CODE> class GMDDecoder {
public:

  // constructor and destructor

  GMDDecoder (CODE *);
  ~GMDDecoder ();

  double Reliability (const GF * c, const GF * _src,
		      const double * const * _alpha, const double * _beta);

  // Decode() tries to do GMD decoding of src[] and then extracts the
  // resulting information word.
  // src must point to an array of size n.
  // dest must point to an array of size k.
  // alpha[i][j] must contain the logarithmic likelihood
  // log (P[src[i] == GF(j)]).
  // beta[i] must contain the largest value alpha[i][j] for j != src[i].
  // On success, the number of corrected errors is returned.
  // A return value < 0 indicates that src[] couldn`t be decoded properly.

  int Decode (GF * dest, const GF * src,
	      const double * const * _alpha, const double * _beta);

  int PrimitiveDecode (GF * dest , const GF * src, const double * _beta);

  int CandidateListDecode (GF * dest, const GF * src,
			   const double * const * _alpha,
			   const double * _beta,
			   int listnum, int listlen);

  int GetLoops() const;

  // --- END OF PUBLIC INTERFACE

private:

  typedef int * pInt;

  CODE * code;
  int n,k,d; // code parameters
  // copies of the parameters to Decode()
  const double * const * alpha;
  const double * beta;
  const GF * src;
  GF * codebuf; // Buffer for temporary decoding results
  int * index;  // indices 0..n-1, sorted for ascending beta[index[i]].
  pInt * candlist; // candlist[i] = candidate list for position index[i]
  int * cand;   // cand[i] = current candidate symbol at index[i]
  int loopcount;

  int HammingDistance (const GF *) const;
  double GMD (const GF *, int) const;
  double BetaSum (const GF *, int) const;
};

template<class GF, class CODE> GMDDecoder<GF,CODE>::GMDDecoder (CODE * _code)
{
  int i, j;
  code = _code;
  n = code->get_n();
  k = code->get_k();
  d = code->get_d();
  codebuf = new GF [n];
  index = new int [n];
  // initialize index[] such that index[i] == i
  for (i = 0; i < n; i++)
    index[i] = i;
  candlist = new pInt [n];
  for (i = 0; i < n; i++) {
    candlist[i] = new int [GF::q()];
    for (j = 0; j < GF::q(); j++)
      candlist[i][j] = j;
  }
  cand = new int [n];
}

template<class GF, class CODE> GMDDecoder<GF,CODE>::~GMDDecoder ()
{
  int i;
  delete [] cand;
  for (i = 0; i < n; i++)
    delete [] candlist[i];
  delete [] candlist;
  delete [] index;
  delete [] codebuf;
}

// HammingDistance[] returns the Hamming distance between c[] and src[].

template<class GF, class CODE> int GMDDecoder<GF,CODE>::HammingDistance
(const GF * c) const
{
  int i, hd = 0;
  for (i=0; i<n; i++)
    if (c[i] != src[i])
      hd++;
  return hd;
}

// GMD(hd) returns the Generalized Minimum Distance of c[] and src[],
// provided the Hamming distance is hd,
// and alpha[i][src[i]] >= alpha[i][j] for all i and all j != src[i].

template<class GF, class CODE> double GMDDecoder<GF,CODE>::GMD
(const GF * c, int hd) const
{
  assert (hd >= 0);
  int i,j,ii;
  double res = 0.0;
  i = 0;
  j = hd;
  while (j > 0) {
    do {
      assert (i<n);
      ii = index[i++];
    } while (c[ii] == src[ii]);
    res += alpha[ii][src[ii]] - alpha[ii][c[ii]];
    j--;
  }
  return res;
}

template<class GF, class CODE> double GMDDecoder<GF,CODE>::BetaSum
(const GF * c, int hd) const
{
  assert (hd >= 0);
  int i,j,ii;
  double res = 0.0;
  i = 0;
  j = d - hd;
  while (j > 0) {
    do {
      assert (i < n);
      ii = index[i++];
    } while (c[ii] != src[ii]);
    res += beta[ii];
    j--;
  }
  return res;
}

template<class GF, class CODE> double GMDDecoder<GF,CODE>::Reliability
(const GF * c, const GF * _src,
 const double * const * _alpha, const double * _beta)
{
  src = _src;
  alpha = _alpha;
  beta = _beta;
  QuickSortUp (index, beta, 0, n-1);
  int hd = HammingDistance (c);
  return BetaSum (c, hd) - GMD (c, hd);
}

template<class GF, class CODE> int GMDDecoder<GF,CODE>::Decode
(GF * dest, const GF * _src, const double * const * _alpha,
 const double * _beta)
{
  int erase;
  alpha = _alpha;
  beta = _beta;
  src = _src;
  QuickSortUp (index, beta, 0, n-1);
  loopcount = 0;
  for (erase = d-1; erase >= 0; erase -= 2) {
    loopcount++;
    int nerr = code->Correct (codebuf, src, index, erase);
    if (nerr < 0)
      continue;
    int hd = HammingDistance(codebuf);
    double gmd = GMD(codebuf, hd);
    double bsum = BetaSum(codebuf, hd);
#ifndef NDEBUG
    cerr << '(' << erase << '/' << hd << '/' << gmd << '/' << bsum << ')';
#endif
    if (gmd < bsum) { // we've got it
      code->ExtractInformationWord (dest, codebuf);
      return hd;
    }
  }
  // we couldn't decode it properly
  code->ExtractInformationWord (dest, src);
  return -1;
}

template<class GF, class CODE> int GMDDecoder<GF,CODE>::PrimitiveDecode
(GF * dest, const GF * _src, const double * _beta)
{
  int erase;
  beta = _beta;
  src = _src;
  QuickSortUp (index, beta, 0, n-1);
  loopcount = 0;
  for (erase = d-3; erase >= 0; erase -= 2) {
    loopcount++;
    int res = code->Correct (codebuf, src, index, erase);
    if (res >= 0) { // we've got it
      int hd = HammingDistance(codebuf);
      code->ExtractInformationWord (dest, codebuf);
      return hd;
    }
  }
  // we couldn't decode it properly
  code->ExtractInformationWord (dest, src);
  return -1;
}

template<class GF, class CODE> int GMDDecoder<GF,CODE>::CandidateListDecode
(GF * dest, const GF * _src,
 const double * const * _alpha, const double * _beta,
 int listnum, int listlen)
{
  int i, res;
  alpha = _alpha;
  beta = _beta;
  src = _src;
  QuickSortUp (index, beta, 0, n-1);
  // compute candidate lists
  for (i = 0; i < listnum; i++) {
    QuickSortUp (candlist[i], alpha[index[i]], 0, GF::q()-1);
  }
  // reset candidate codeword
  for (i = 0; i < listnum; i++)
    cand[i] = 0;
  for (i = 0; i < n; i++)
    codebuf[i] = src[i];
  res = 0;
  loopcount = 0;
  while (res == 0) {
    loopcount++;
    // update candidate codeword
    for (i = 0; i < listnum; i++)
      codebuf[index[i]] = candlist[i][GF::q()-1-cand[i]];
    if (code->Decode (dest, codebuf) >= 0) {
      // we've got it
      return 0;
    }
    // increment candidate counter
    for (i = 0; i < listnum; i++) {
      cand[i]++;
      if (cand[i] == listlen)
	cand[i] = 0;
      else
	break;
    }
    if (i == listnum) // generated all possible candidates
      res = -1;
  } // while (res == 0)
  return res;
}

template<class GF, class CODE> int GMDDecoder<GF,CODE>::GetLoops() const
{
  return loopcount;
}

#endif
