#ifndef _HEAP_H
#define _HEAP_H

// Templates for efficiently managing a fixed-size heap structure
// The base type must define an operator <

// (C) 1997-1998 Frank Paehlke <frank@paehlke.de>
// Created: 12/04/1996
// $Id: heap.h,v 1.1 2003/10/14 08:49:02 frank Exp $

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// $Log: heap.h,v $
// Revision 1.1  2003/10/14 08:49:02  frank
// moved from . to src/
//
// Revision 1.4  2003/10/14 08:41:58  frank
// ANSI C++ / gcc 3.x compatibility changes
//
// Revision 1.3  2002/08/02 14:45:06  paehlke
// added GPL comment
//
// Revision 1.2  1998/05/28 12:57:00  paehlke
// Final version for masters thesis
//
// Revision 1.1  1997/06/20 18:16:35  frank
// Initial revision
//

// --- BEGIN OF PUBLIC INTERFACE ---

// class SimpleHeap manages a heap where no elements except for the
// smallest one are ever modified.

template<class T> class SimpleHeap
{
public:

  // The SimpleHeap constructor converts the unordered list data[] into
  // a correct heap structure.
  // maxsize = 0 means that maxsize should be set to the current size.

  SimpleHeap (T * data, int size, int maxsize = 0);
  ~SimpleHeap ();

  int get_size () const;
  int get_maxsize () const;
  int IsEmpty () const;
  int IsFull () const;

  const T & Element (int i) const;
  T & SmallestElement ();
  const T & SmallestElement () const;

  // SmallestElementModified() must be called when the smallest
  // element has been modified

  SimpleHeap & SmallestElementModified ();
  SimpleHeap & InsertElement (T &);
  SimpleHeap & DeleteSmallestElement ();
  SimpleHeap & Empty ();

  // Rebuild() must be called when any elements except for the smallest
  // one have been modified.
  // If you ever need this function, you should consider using
  // class Heap instead of class SimpleHeap because it can handle
  // modified elements much more efficiently.

  SimpleHeap & Rebuild ();

  // the following is an ugly hack which is unfortunately
  // needed by the stack decoder: reduce the size of the heap
  // by throwing away the last num elements

  SimpleHeap & reduce_size (int num);

private:

  typedef T * pT;

  int size, maxsize;
  pT * data;

  SimpleHeap & SinkDown (int pos);
  SimpleHeap & RiseUp (int pos);
};

template<class T> class Heap;

// class HeapElement is used to store the elements of class Heap.

template<class T> class HeapElement
{
  friend class Heap<T>;

public:

  typedef HeapElement<T> * PTR;
  T Data;

private:

  int HeapPos;
};

template<class T> class Heap
{
public:

  // The Heap constructor converts the unordered list data[] into
  // a correct heap structure
  // maxsize = 0 means that maxsize should be set to the current size.

  Heap (HeapElement<T> * data, int size, int maxsize = 0);
  ~Heap ();

  int get_size () const;
  int get_maxsize () const;
  int IsEmpty () const;
  int IsFull () const;

  HeapElement<T> & Element (int i);
  const HeapElement<T> & Element (int i) const;
  HeapElement<T> & SmallestElement ();
  const HeapElement<T> & SmallestElement () const;

  // ElementModified() or SmallestElementModified() must be called
  // when a heap element has been modified

  Heap & ElementModified (HeapElement<T> &);
  Heap & SmallestElementModified ();
  Heap & InsertElement (HeapElement<T> &);
  Heap & DeleteElement (HeapElement<T> &);
  Heap & DeleteSmallestElement ();
  Heap & Empty ();

  // SinkDown(el) can be called instead of ElementModified(el)
  // when el->Data has INCREASED
  Heap & SinkDown (HeapElement<T> & el);

  // RiseUp(el) can be called instead of ElementModified(el)
  // when el->Data has DECREASED
  Heap & RiseUp (HeapElement<T> & el);

  // Rebuild() may be called when many elements have been modified
  // instead of calling ElementModified() for every modified element
  Heap & Rebuild ();

  // --- END OF PUBLIC INTERFACE ---

private:

  int size, maxsize;
  typename HeapElement<T>::PTR * data;
};

// implementation of class SimpleHeap

template<class T> SimpleHeap<T>::SimpleHeap
(T * _data, int _size, int _maxsize)
{
  int i;
  assert ((_size == 0) || ((_size > 0) && (_data != NULL)));
  size = _size;
  if (_maxsize == 0)
    maxsize = size;
  else
    maxsize = _maxsize;
  assert ((maxsize > 0) && (maxsize >= size));
  data = new pT [maxsize];
  for (i = 0; i < size; i++)
    data[i] = &_data[i];
  Rebuild ();
}

template<class T> inline SimpleHeap<T>::~SimpleHeap ()
{
  delete [] data;
}

template<class T> inline int SimpleHeap<T>::get_size () const
{
  return size;
}

template<class T> inline int SimpleHeap<T>::get_maxsize () const
{
  return maxsize;
}

template<class T> inline int SimpleHeap<T>::IsEmpty () const
{
  return (size == 0);
}

template<class T> inline int SimpleHeap<T>::IsFull () const
{
  return (size == maxsize);
}

template<class T> inline const T & SimpleHeap<T>::Element (int i) const
{
  assert ((i >= 0) && (i < size));
  return *data[i];
}

template<class T> inline T & SimpleHeap<T>::SmallestElement ()
{
  return *data[0];
}

template<class T> inline const T & SimpleHeap<T>::SmallestElement () const
{
  return *data[0];
}

template<class T> SimpleHeap<T> & SimpleHeap<T>::SinkDown (int pos)
{
  int i = pos;
  for (;;) {
    int j = 2*i + 1;
    if ((j < size) && (*data[j] < *data[i])) {
      if ((j+1 < size) && (*data[j+1] < *data[j]))
 	j++;
      pT h = data[i];
      data[i] = data[j];
      data[j] = h;
      i = j;
    }
    else if ((++j < size) && (*data[j] < *data[i])) {
      pT h = data[i];
      data[i] = data[j];
      data[j] = h;
      i = j;
    }
    else
      break;
  }
  return *this;
}

template<class T> SimpleHeap<T> & SimpleHeap<T>::RiseUp (int pos)
{
  assert ((pos >= 0) && (pos < size));
  int i = pos;
  while (i > 0) {
    int j = (i-1) / 2;
    if (*data[i] < *data[j]) {
      pT h = data[i];
      data[i] = data[j];
      data[j] = h;
      i = j;
    }
    else
      break;
  }
  return *this;
}

template<class T> inline SimpleHeap<T> &
SimpleHeap<T>::SmallestElementModified ()
{
  return SinkDown (0);
}

template<class T> inline SimpleHeap<T> &
SimpleHeap<T>::InsertElement (T & el)
{
  assert (size < maxsize);
  data[size++] = &el;
  return RiseUp (size - 1);
}

template<class T> inline SimpleHeap<T> &
SimpleHeap<T>::DeleteSmallestElement ()
{
  assert (size > 0);
  size--;
  if (size > 0) {
    data[0] = data[size];
    SinkDown (0);
  }
  return *this;
}

template<class T> inline SimpleHeap<T> & SimpleHeap<T>::Empty ()
{
  size = 0;
  return *this;
}

template<class T> inline SimpleHeap<T> & SimpleHeap<T>::Rebuild ()
{
  int i;
  for (i = size/2 - 1; i >= 0; i--)
    (void) SinkDown (i);
  return *this;
}

template<class T> inline SimpleHeap<T> & SimpleHeap<T>::reduce_size
(int num)
{
  assert ((num >= 0) && (num <= size));
  size -= num;
  return *this;
}

// implementation of class Heap

template<class T> Heap<T>::Heap
(HeapElement<T> * _data, int _size, int _maxsize)
{
  int i;
  assert ((_size == 0) || ((_size > 0) && (_data != NULL)));
  size = _size;
  if (_maxsize == 0)
    maxsize = size;
  else
    maxsize = _maxsize;
  assert ((maxsize > 0) && (maxsize >= size));
  data = new typename HeapElement<T>::PTR [maxsize];
  for (i=0; i<size; i++) {
    data[i] = &_data[i];
    data[i]->HeapPos = i;
  }
  for (i = size; i < maxsize; i++)
    data[i] = NULL;
  Rebuild ();
}

template<class T> inline Heap<T>::~Heap ()
{
  delete [] data;
}

template<class T> inline int Heap<T>::get_size () const
{
  return size;
}

template<class T> inline int Heap<T>::get_maxsize () const
{
  return maxsize;
}

template<class T> inline int Heap<T>::IsEmpty () const
{
  return (size == 0);
}

template<class T> inline int Heap<T>::IsFull () const
{
  return (size == maxsize);
}

template<class T> inline HeapElement<T> & Heap<T>::Element (int i)
{
  assert ((i >= 0) && (i < size));
  return *data[i];
}

template<class T> inline const HeapElement<T> &
Heap<T>::Element (int i) const
{
  assert ((i >= 0) && (i < size));
  return *data[i];
}

template<class T> inline HeapElement<T> & Heap<T>::SmallestElement ()
{
  return *data[0];
}

template<class T> inline const HeapElement<T> &
Heap<T>::SmallestElement () const
{
  return *data[0];
}

template<class T> Heap<T> & Heap<T>::SinkDown (HeapElement<T> & el)
{
  int i = el.HeapPos;
  for (;;) {
    int j = 2*i + 1;
    if ((j < size) && (data[j]->Data < data[i]->Data)) {
      if ((j+1 < size) && (data[j+1]->Data < data[j]->Data)) j++;
      HeapElement<T> * h = data[i];
      data[i] = data[j];
      data[i]->HeapPos = i;
      data[j] = h;
      data[j]->HeapPos = j;
      i = j;
    }
    else if ((++j < size) && (data[j]->Data) < (data[i]->Data)) {
      HeapElement<T> * h = data[i];
      data[i] = data[j];
      data[i]->HeapPos = i;
      data[j] = h;
      data[j]->HeapPos = j;
      i = j;
    }
    else
      break;
  }
  return *this;
}

template<class T> Heap<T> & Heap<T>::RiseUp (HeapElement<T> & el)
{
  int i = el.HeapPos;
  while (i > 0) {
    int j = (i-1) / 2;
    if (data[i]->Data < data[j]->Data) {
      HeapElement<T> * h = data[i];
      data[i] = data[j];
      data[i]->HeapPos = i;
      data[j] = h;
      data[j]->HeapPos = j;
      i = j;
    }
    else
      break;
  }
  return *this;
}

template<class T> inline Heap<T> & Heap<T>::ElementModified
(HeapElement<T> & el)
{
  assert ((el.HeapPos >= 0) && (el.HeapPos < size)
	  && (data[el.HeapPos] = &el));
  SinkDown (el);
  return RiseUp (el);
}

template<class T> inline Heap<T> & Heap<T>::SmallestElementModified ()
{
  return SinkDown (*data[0]);
}

template<class T> inline Heap<T> & Heap<T>::InsertElement
(HeapElement<T> & el)
{
  assert (size < maxsize);
  el.HeapPos = size;
  data[size++] = &el;
  return RiseUp (el);
}

template<class T> inline Heap<T> & Heap<T>::DeleteElement
(HeapElement<T> & el)
{
  assert ((el.HeapPos >= 0) && (el.HeapPos < size)
	  && (data[el.HeapPos] = &el));
  size--;
  if (el.HeapPos < size) {
    data[el.HeapPos] = data[size];
    data[el.HeapPos]->HeapPos = el.HeapPos;
    SinkDown (*data[el.HeapPos]);
    RiseUp (*data[el.HeapPos]);
  }
  return *this;
}

template<class T> inline Heap<T> & Heap<T>::DeleteSmallestElement ()
{
  assert (size > 0);
  size--;
  if (size > 0) {
    data[0] = data[size];
    data[0]->HeapPos = 0;
    SinkDown (*data[0]);
  }
  return *this;
}

template<class T> inline Heap<T> & Heap<T>::Empty ()
{
  size = 0;
  return *this;
}

template<class T> inline Heap<T> & Heap<T>::Rebuild ()
{
  int i;
  for (i = size/2 - 1; i >= 0; i--)
    SinkDown (*data[i]);
  return *this;
}

#endif
