#ifndef _RSCODE_H
#define _RSCODE_H

// A class for Reed-Solomon encoding and decoding
// where the code symbols come from a Galois field GF = GF(q)
// and the code word length is n <= q - 1

// (C) 1997-1998 Frank Paehlke <frank@paehlke.de>
// Created: 01/27/1997
// $Id: rscode.h,v 1.1 2003/10/14 08:49:02 frank Exp $

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// $Log: rscode.h,v $
// Revision 1.1  2003/10/14 08:49:02  frank
// moved from . to src/
//
// Revision 1.5  2002/08/02 14:45:06  paehlke
// added GPL comment
//
// Revision 1.4  1998/05/28 12:57:04  paehlke
// Final version for masters thesis
//
// Revision 1.3  1997/10/19 16:27:01  frank
// added support for shortened RS codes
//
// Revision 1.2  1997/10/17 21:14:28  frank
// general cleanup and simplification
// error-and-erasure decoding included into Berlekamp-Massey algorithm
// arguments of Encode() and Decode(): (GF *) instead of (UBYTE *)
//
// Revision 1.1  1997/06/20 18:30:26  frank
// Initial revision
//

#ifndef _ASSERT_H
#include <assert.h>
#endif

template<class GF> class RSState;

// --- BEGIN OF PUBLIC INTERFACE

template<class GF> class RSCode {

public:

  // constructor and destructor

  // RSCode(d,n) creates a Reed-Solomon code of length n
  // with code distance d
  // and information word length k = n+1 - d
  // which can correct up to t = (d-1) / 2 symbol errors.
  // RSCode(d) is equivalent to RSCode(d,GF::q()-1)

  RSCode (int,int);
  RSCode (int);
  ~RSCode ();

  // code parameters

  int get_t () const; // get max. # of correctable errors
  int get_d () const; // get code distance
  int get_k () const; // get information word length
  int get_n () const; // get code word length
  double get_r () const; // get code rate
  int get_q () const; // get cardinality of underlying GF
  int get_deggenpoly () const; // get degree of generator polynomial
  const GF * get_genpoly () const; // get coeffs of generator polynomial

  // encoding and decoding

  int ExtractInformationWord (GF * dest, const GF * src);

  // When calling Encode(), dest must point to an array of length n,
  // and src must point to an array of length k.
  // src[] and dest[] must not overlap.

  void Encode (GF * dest, const GF * src) const;

  // Correct() tries to correct cv erasures whose locations are given in Iv[],
  // and up to tau errors where 2*tau + cv < d.
  // dest and src must point to arrays of size n, which may overlap.
  // On success, the number of corrected errors is returned.
  // A return value < 0 indicates that src[] contained more errors
  // than could be corrected.

  int Correct (GF * dest, const GF * src, const int * Iv = NULL, int cv = 0);

  // Decode() works like Correct(), but extracts the information word
  // to dest[].
  // dest must point to an array of size k.
  // src[] and dest[] must not overlap.

  int Decode (GF * dest, const GF * src, const int * Iv = NULL, int cv = 0);

  // support for sequential decoding

  typedef RSState<GF> State;

  // Allocate buffer memory for an object of class State.
  // The deallocation is done by the destructor of class State.

  void AllocateState (State &) const;

  void InitialState (State &, int reverse) const;
  int IsTerminalState (const State &, int reverse) const;

  double BranchMetric (const State &, const GF &,
		       const double * const * metric,
		       int reverse) const;

  double KnownSymbolBranchMetric (const State &, int reverse) const;

  double TunnelMetric (const State &, const State &,
		       const double * const * metric, int reverse) const;

  double MaximumRemainingMetric (const State & state, int reverse) const;

  void MoveForward (State &, const State &, const GF &,
		    int reverse) const;

  void EncodeConvolutional (GF * dest, const GF * src) const;

  // --- END OF PUBLIC INTERFACE

protected:

  int n, k, d; // code parameters
  GF * genpoly; // generator polynomial
  int deggenpoly; // degree of genpoly(X)
  GF * S; // syndrome vector
  GF * C; // error [and erasure] locator polynomial
  int degC; // degree of C(X)
  GF * lastC; // needed for Berlekamp-Massey algorithm
  int deglastC;
  GF * H; // temporary polynomial
  int degH;
  GF * T; // error value polynomial
  int degT;
  int * I; // error location vector
  GF * errval; // errval[j] gives the error value at I[j]
  int Ilen; // length of I [and of errval]
  GF * codebuf; // buffer for one code word

  RSCode ();

  // some auxiliary functions

  void Initialize (int, int);
  void InitGeneratorPolynomial ();

  // parts of decoder

  void CalculateELP (const int *, int);
  void BerlekampMassey ();
  void CalculateT ();
  void ChienSearch ();
  void Forney () const;
};

// InitGeneratorPolynomial() calculates
// genpoly(X) = prod (X - z^i, i = 0..2*t-1)
// where z = GF::gen.

template<class GF> void RSCode<GF>::InitGeneratorPolynomial ()
{
  int i,j;
  genpoly = new GF [d];
  // set genpoly to 1
  genpoly[0] = 1;
  deggenpoly = 0;
  for (i = 0; i < d-1; i++) {
    // set genpoly = genpoly * (X-z^i)
    GF zi = GF::fastexp (i);
    genpoly[deggenpoly+1] = genpoly[deggenpoly];
    for (j=deggenpoly; j>0; j--)
      genpoly[j] = genpoly[j-1] - zi * genpoly[j];
    genpoly[0] *= -zi;
    deggenpoly++;
  }
}

template<class GF> void RSCode<GF>::Initialize(int _d, int _n)
{
  // set code parameters
  n = _n;
  d = _d;
  k = n - d + 1;
  // allocate memory for polynomials and vectors
  S = new GF [d-1];
  C = new GF [d];
  lastC = new GF [d];
  H = new GF [d];
  T = new GF [d-1];
  I = new int [d-1];
  errval = new GF [d-1];
  codebuf = new GF [n];
}

template<class GF> inline RSCode<GF>::RSCode (int _d, int _n)
{
  assert ((_n > 0) && (_n < GF::q()) && (_d > 0) && (_d <= _n));
  Initialize (_d, _n);
  InitGeneratorPolynomial ();
}

template<class GF> inline RSCode<GF>::RSCode (int _d)
{
  assert ((_d > 0) && (_d < GF::q()));
  Initialize (_d, GF::q()-1);
  InitGeneratorPolynomial ();
}

template<class GF> RSCode<GF>::RSCode ()
{}

template<class GF> RSCode<GF>::~RSCode ()
{
  // free all allocated memory
  delete [] genpoly;
  delete [] codebuf;
  delete [] errval;
  delete [] I;
  delete [] T;
  delete [] H;
  delete [] lastC;
  delete [] C;
  delete [] S;
}

template<class GF> inline int RSCode<GF>::get_t () const
{
  return (d-1) >> 1;
}

template<class GF> inline int RSCode<GF>::get_d () const
{
  return d;
}

template<class GF> inline int RSCode<GF>::get_k () const
{
  return k;
}

template<class GF> inline int RSCode<GF>::get_n () const
{
  return n;
}

template<class GF> inline double RSCode<GF>::get_r () const
{
  return (double) k / (double) n;
}

template<class GF> inline int RSCode<GF>::get_q () const
{
  return GF::q();
}

template<class GF> inline int RSCode<GF>::get_deggenpoly () const
{
  return deggenpoly;
}

template<class GF> inline const GF * RSCode<GF>::get_genpoly () const
{
  return genpoly;
}

template<class GF> inline int RSCode<GF>::ExtractInformationWord
(GF * dest, const GF * src)
{
  int i;
  for (i = 0; i < k; i++)
    dest[i] = src[i];
  return 0; // always successful because the code is systematic
}

// Encode(dest,src) does systematic encoding
// where src[] is copied to dest[deggenpoly..n-1]
// and the redundancy is stored in dest[0..deggenpoly-1]

template<class GF> void RSCode<GF>::Encode (GF * dest, const GF * src) const
{
  int i, j;
  // initialize dest(X) with 0
  for (j = 0; j < n; j++)
    dest[j] = GF(0);
  for (i = 0; i < k; i++) {
    // invariant: dest(X) == 0 mod genpoly(X) && dest[0..i-1] == src[0..i-1]
    GF h = (src[i] - dest[i]) / genpoly[0];
    // set dest'(X) = dest(X) + X^cpos * h * genpoly(X)
    // to ensure that dest'(X) == 0 mod genpoly(X) and dest'[i] == src[i]
    for (j = 0; j <= deggenpoly; j++)
      dest[i + j] += h * genpoly[j];
    assert (dest[i] == src[i]);
  }
}

// BerlekampMassey() takes the erasure locator polynomial in C(X)
// and extends it to an error-and-erasure locator polynomial using
// the syndrome vector S[] as calculated by CalculateSyndrome().
// If there are no erasures, C(X) must be set to 1 before calling
// BerlekampMassey().

template<class GF> inline void RSCode<GF>::BerlekampMassey ()
{
  int i,j,k,len,degC0;
  GF delta, lastdelta;
  if (degC == 0) {
    // search for the first j with S[j] != 0
    for (j = 0; j < d-1; j++) {
      if (S[j] != GF(0)) {
	delta = S[j];
	break;
      }
    }
  }
  else { // degC > 0
    // search for the first j with delta != 0
    for (j = degC; j < d-1; j++) {
      delta = S[j];
      for (i=1; i<=degC; i++)
	delta += C[i] * S[j-i];
      if (delta != GF(0))
	break;
    }
  }
  if (j == d-1) {
    // no error has to be corrected (but maybe some erasures)
    return;
  }
  // set starting values for Berlekamp-Massey algorithm
  for (i=0; i<=degC; i++)
    lastC[i] = C[i];
  deglastC = degC0 = degC;
  lastdelta = delta;
  len = j+1; // shift register length
  k = 1;
  for (j++; j < d-1; j++, k++) {
    // calculate delta
    delta = S[j];
    for (i=1; i<=degC; i++)
      delta += C[i] * S[j-i];
    if (delta != GF(0)) {
      GF delta1 = delta / lastdelta;
      if (2*len > degC0 + j) {
	// set C(X) = C(X) - (delta/lastdelta) * x^k * lastC(X)
	for (i = k+deglastC; i > degC; i--)
	  C[i] = 0;
	for (i = 0; i <= deglastC; i++)
	  C[k+i] -= delta1 * lastC[i];
	// recalculate the degree of C(X)
	if (k + deglastC > degC)
	  degC = k + deglastC;
      }
      else { // 2*len <= degC0 + j
        // set H(X) = C(X)
        for (i=0; i<=degC; i++)
  	  H[i] = C[i];
        degH = degC;
	// set C(X) = C(X) - (delta/lastdelta) * x^k * lastC(X)
	for (i = k+deglastC; i > degC; i--)
	  C[i] = 0;
	for (i=0; i<=deglastC; i++)
	  C[k+i] -= delta1 * lastC[i];
	// recalculate the degree of C(X)
	if (k + deglastC > degC)
	  degC = k + deglastC;
	// set lastC(X) = H(X), reset k
        for (i=0; i<=degH; i++)
          lastC[i] = H[i];
	deglastC = degH;
	lastdelta = delta;
        k = 0;
	// set the new shift register length
	len = degC0 + j + 1 - len;
      } // endif (2*len > j)
      // recalculate the degree of C(X)
      while (C[degC] == GF(0))
	degC--;
    } // endif (delta != GF(0))
  } // endfor (j)
}

// CalculateT() calculates the error value polynomial T(X)
// using the syndrome vector S[] as calculated by CalculateSyndrome()
// and the error locator polynomial C(X) as calculated by BerlekampMassey()

template<class GF> inline void RSCode<GF>::CalculateT()
{
  assert (degC > 0);
  int i;
  for (degT=0; degT<degC; degT++) {
    T[degT] = -S[degT];
    for (i=1; i<=degT; i++)
      T[degT] -= C[i]*S[degT-i];
  }
  do
    degT--;
  while (degT > 0 && T[degT] == GF(0));
}

// ChienSearch() searches the zeroes of the error locator polynomial C(X),
// as calculated by BerlekampMassey().
// Those values of i for which C(GF::gen^(-i)) == 0 are stored into the
// array I in ascending order, and Ilen is set according to their number.

template<class GF> inline void RSCode<GF>::ChienSearch ()
{
  assert (degC >= 0);
  int i;
  Ilen = 0;
  for (i=0; i<n; i++)
    if (GF::EvalPolynomial (C, degC, GF::fastinvexp (i)) == GF(0)) {
      I[Ilen++] = i;
      if (Ilen == degC)
	return;
    }
}

// Forney() computes errval[j] = error value at I[j] (j = 0..Ilen-1)
// using the error locator polynomial C(X) as calculated by BerlekampMassey(),
// the error value polynomial T(X) as calculated by CalculateT()
// and the error location vector I[] as generated by ChienSearch()

template<class GF> inline void RSCode<GF>::Forney () const
{
  assert (degC > 0 && degT >= 0 && Ilen >= 0);
  int j;
  for (j=0; j<Ilen; j++) {
    GF zmi = GF::fastinvexp (I[j]);
    errval[j] = - (GF::EvalPolynomial (T, degT, zmi)
		   / GF::EvalPolynomialD (C, degC, zmi)) / zmi;
  }
}

// CalculateELP (Iv, cv) calculates the erasure locator polynomial C(X)
// from a given erasure location vector Iv[] of length cv

template<class GF> inline void RSCode<GF>::CalculateELP
(const int * Iv, int cv)
{
  int i;
  // set C(X) to 1
  C[0] = 1;
  degC = 0;
  for (; degC < cv; degC++) {
    // set C(X) = C(X) * (1 - X * GF64::gen^Iv[degC])
    GF zi = GF::fastexp (Iv[degC]);
    C[degC+1] = -zi * C[degC];
    for (i=degC; i>0; i--)
      C[i] -= zi * C[i-1];
  }
}

// Correct(dest,src,Iv,cv) corrects cv erasures whose places are given
// in cv[], and up to tau errors where 2*tau + cv < d,
// while copying src[] to dest[].

template<class GF> int RSCode<GF>::Correct (GF * dest, const GF * src,
					    const int * Iv, int cv)
{
  assert (cv >= 0);
  if (cv >= d)
    return -5;
  int i;
  // copy src[] to dest[]
  for (i=0; i<n; i++)
    dest[i] = src[i];
  // calculate syndrome vector
  int noerrors = 1;
  for (i=0; i<d-1; i++) {
    S[i] = GF::EvalPolynomial (src, n-1, GF::fastexp (i));
    if (S[i] != GF(0))
      noerrors = 0;
  }
  if (noerrors) {
    // no errors or erasures need to be corrected
    return 0;
  }
  CalculateELP (Iv, cv);
  BerlekampMassey ();
  if (2*degC - cv >= d) {
    // more errors occured than could be corrected
    return -1;
  }
  else if (degC > 0) {
    ChienSearch ();
    if (Ilen < degC) {
      // C(X) doesn't split into linear factors
      return -2;
    }
    CalculateT ();
    Forney ();
    // apply the corrections to dest[]
    for (i=0; i<Ilen; i++)
      dest[I[i]] += errval[i];
  }
  else // degC <= 0
    Ilen = 0;
  // check for decoder malfunction
  for (i=0; i<d-1; i++) {
    if (GF::EvalPolynomial (dest, n-1, GF::fastexp (i)) != GF(0)) {
      // Decoder malfunction! Undo all corrections.
      int j;
      for (j=0; j<Ilen; j++)
	dest[I[j]] -= errval[j];
      return -3;
    }
  }
  // correction successful, return # of corrected errors and erasures
  return degC;
}

template<class GF> inline int RSCode<GF>::Decode
(GF * dest, const GF * src, const int * Iv, int cv)
{
  int res = Correct (codebuf, src, Iv, cv);
  (void) ExtractInformationWord (dest, codebuf);
  return res;
}

// support for sequential decoding

// encoder state

template<class GF> class RSState
{
public:

  int ipos;
  GF * buf;

  RSState() : buf(NULL) {}
  ~RSState() { if (buf != NULL) delete [] buf; }
}; // size: 8 bytes + sizeof(buf)

template<class GF> void RSCode<GF>::AllocateState
(RSState<GF> & state) const
{
  assert (state.buf == NULL);
  state.buf = new GF [deggenpoly];
}

template<class GF> void RSCode<GF>::InitialState
(RSState<GF> & state, int reverse) const
{
  int i;
  if (reverse)
    state.ipos = k - 1;
  else
    state.ipos = 0;
  for (i = 0; i < deggenpoly; i++)
    state.buf[i] = GF(0);
}

template<class GF> inline int RSCode<GF>::IsTerminalState
(const RSState<GF> & state, int reverse) const
{
  if (reverse)
    return (state.ipos == -1);
  else
    return (state.ipos == k);
}

template<class GF> double RSCode<GF>::BranchMetric
(const RSState<GF> & state, const GF & x, const double * const * metric,
 int reverse) const
{
  double bm;
  int i;
  if (reverse) {
    if (state.ipos == 0) { // next node is terminal
      bm = metric[deggenpoly]
	[(int) (state.buf[deggenpoly-1] + x * genpoly[deggenpoly])];
      for (i = deggenpoly-1; i > 0; i--) {
	bm += metric[i][(int) (state.buf[i-1] + x * genpoly[i])];
      }
      bm += metric[0][(int) (x * genpoly[0])];
    }
    else { // next node isn't terminal
      bm = metric[state.ipos + deggenpoly]
	[(int) (state.buf[deggenpoly-1] + x * genpoly[deggenpoly])];
    }
  }
  else { // !reverse
    if (state.ipos == k-1) { // next node is terminal
      bm = metric[state.ipos][(int) (state.buf[0] + x * genpoly[0])];
      for (i = 1; i < deggenpoly; i++) {
	bm += metric[state.ipos + i][(int) (state.buf[i] + x * genpoly[i])];
      }
      bm += metric[state.ipos + deggenpoly][(int) (x * genpoly[deggenpoly])];
    }
    else // next node isn't terminal
      bm = metric[state.ipos][(int) (state.buf[0] + x * genpoly[0])];
  }
  return bm;
}

template<class GF> double RSCode<GF>::KnownSymbolBranchMetric
(const RSState<GF> & state, int reverse) const
{
  double bm = (1.0 - get_r()) * GF::ldq();
  if ((reverse && (state.ipos == 0)) || (!reverse && (state.ipos == k-1))) {
    // next node is terminal
    bm *= deggenpoly+1;
  }
  return bm;
}

template<class GF> double RSCode<GF>::TunnelMetric
(const RSState<GF> & state1, const RSState<GF> & state2,
 const double * const * metric, int reverse) const
{
  int ipos, i;
  assert (state1.ipos == state2.prev->ipos);
  if (reverse)
    ipos = state2.ipos;
  else
    ipos = state1.ipos;
  double bm = metric[ipos][(int) (state1.buf[0] + state2.buf[0])];
  for (i = 1; i < deggenpoly; i++)
    bm += metric[ipos + i][(int) (state1.buf[i] + state2.buf[i])];
  return bm;
}

template<class GF> double RSCode<GF>::MaximumRemainingMetric
(const RSState<GF> & state, int reverse) const
{
  double bm = (1.0 - get_r()) * GF::ldq();
  if (reverse) {
    bm *= (deggenpoly + state.ipos);
  }
  else { // !reverse
    bm *= (n-1 - state.ipos);
  }
  return bm;
}

template<class GF> void RSCode<GF>::MoveForward
(RSState<GF> & newstate, const RSState<GF> & oldstate, const GF & x,
 int reverse) const
{
  int i;
  assert ((oldstate.ipos >= 0) && (oldstate.ipos < k));
  if (reverse) {
    newstate.ipos = oldstate.ipos - 1;
    for (i = deggenpoly-1; i > 0; i--)
      newstate.buf[i] = oldstate.buf[i-1] + x * genpoly[i];
    newstate.buf[0] = x * genpoly[0];
  }
  else { // !reverse
    newstate.ipos = oldstate.ipos + 1;
    for (i = 1; i < deggenpoly; i++)
      newstate.buf[i-1] = oldstate.buf[i] + x * genpoly[i];
  }
}

template<class GF> inline void RSCode<GF>::EncodeConvolutional
(GF * dest, const GF * src) const
{
  int i, j;
  // initialize dest[] with 0
  for (i = 0; i < n; i++)
    dest[i] = 0;
  for (i = 0; i < k; i++) {
    GF srci = src[i];
    for (j = 0; j <= deggenpoly; j++)
      dest[i+j] += srci * genpoly[j];
    i++;
  }
}

#endif
