#ifndef _SXRSCODE_H
#define _SXRSCODE_H

// A class for encoding and decoding of systematic expurgated RS codes
// where the code symbols are from the Galois field GF = GF(q)
// and the code word length is n <= q - 1

// (C) 1997-1998 Frank Paehlke <frank@paehlke.de>
// Created: 1997/09/22
// $Id: sxrscode.h,v 1.1 2003/10/14 08:49:02 frank Exp $

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// $Log: sxrscode.h,v $
// Revision 1.1  2003/10/14 08:49:02  frank
// moved from . to src/
//
// Revision 1.4  2002/08/02 14:45:06  paehlke
// added GPL comment
//
// Revision 1.3  1998/05/28 12:57:06  paehlke
// Final version for masters thesis
//
// Revision 1.2  1998/01/26 14:34:02  paehlke
// Minor corrections
//
// Revision 1.1  1997/11/27 15:06:00  paehlke
// Initial revision
//

#ifndef _ASSERT_H
#include <assert.h>
#endif

#ifndef _XRSCODE_H
#include "xrscode.h"
#endif

// --- BEGIN OF PUBLIC INTERFACE

template<class GF> class SXRSCode : public XRSCode<GF> {

public:

  // constructor and destructor

  // SXRSCode(d,xd,n) creates a systematic expurgated RS code of length n
  // with code distance d and expurgation distance xd
  // which can correct up to t = (d-1) / 2 symbol errors.
  // SXRSCode(d,xd) is equivalent to SXRSCode(d,xd,GF::q()-1)

  SXRSCode (int, int, int);
  SXRSCode (int, int);
  ~SXRSCode ();

  // encoding and decoding

  int ExtractInformationWord (GF * dest, const GF * src);

  // When calling Encode(), dest must point to an array of length n,
  // and src must point to an array of length k.
  // src[] and dest[] must not overlap.

  void Encode (GF * dest, const GF * src) const;

  // When calling Decode(), dest must point to an array of length k,
  // and src must point to an array of length n.
  // src[] and dest[] must not overlap.

  int Decode (GF * dest, const GF * src, const int * Iv = NULL, int cv = 0);

  // --- END OF PUBLIC INTERFACE

protected:

  SXRSCode ();
};

template<class GF> inline SXRSCode<GF>::SXRSCode (int _d, int _xd, int _n)
  : XRSCode<GF> (_d, _xd, _n)
{}

template<class GF> inline SXRSCode<GF>::SXRSCode (int _d, int _xd)
  : XRSCode<GF> (_d, _xd)
{}

template<class GF> SXRSCode<GF>::SXRSCode ()
{}

template<class GF> SXRSCode<GF>::~SXRSCode ()
{}

template<class GF> inline int SXRSCode<GF>::ExtractInformationWord
(GF * dest, const GF * src)
{
  int ipos, cpos, xcount;
  ipos = cpos = 0;
  xcount = (n - deggenpoly) % xd; // expurgation counter
  while (ipos < k) {
    if (xcount == 0) {
      // the code symbol at cpos is redundant
      cpos++;
      xcount = xd - 1;
    }
    assert (cpos + deggenpoly < n);
    dest[ipos++] = src[cpos++];
    xcount--;
  }
  return 0; // always successful because the code is systematic
}

// Encode(dest,src) encodes the information word src[]
// and writes the resulting code word to dest[]

template<class GF> void SXRSCode<GF>::Encode (GF * dest, const GF * src) const
{
  int ipos, cpos, xcount, i;
  // initialize dest(X) with 0
  for (i = 0; i < n; i++)
    dest[i] = GF(0);
  ipos = cpos = 0;
  xcount = (n - deggenpoly) % xd; // expurgation counter
  while (ipos < k) {
    // invariant: dest(X) == 0 mod genpoly(X)
    if (xcount == 0) {
      // the code symbol at cpos is redundant
      cpos++;
      xcount = xd - 1;
    }
    assert (cpos + deggenpoly < n);
    GF h = (src[ipos] - dest[cpos]) / genpoly[0];
    // set dest'(X) = dest(X) + X^cpos * h * genpoly(X)
    // to ensure that dest'(X) == 0 mod genpoly(X)
    // and dest'[cpos] == src[ipos]
    for (i = 0; i <= deggenpoly; i++)
      dest[cpos + i] += h * genpoly[i];
    assert (dest[cpos] == src[ipos]);
    ipos++; cpos++;
    xcount--;
  }
  assert ((cpos + deggenpoly == n) && (xcount == 0));
}

template<class GF> inline int SXRSCode<GF>::Decode
(GF * dest, const GF * src, const int * Iv, int cv)
{
  int res = Correct (codebuf, src, Iv, cv);
  (void) ExtractInformationWord (dest, codebuf);
  return res;
}

#endif
