#ifndef _TREEDECODER_H
#define _TREEDECODER_H

// A class for several variants of tree (sequential) decoding

// (C) 1997-1998 Frank Paehlke <frank@paehlke.de>
// $Id: treedecoder.h,v 1.1 2003/10/14 08:49:02 frank Exp $

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// $Log: treedecoder.h,v $
// Revision 1.1  2003/10/14 08:49:02  frank
// moved from . to src/
//
// Revision 1.4  2002/08/02 14:45:06  paehlke
// added GPL comment
//
// Revision 1.3  1998/05/28 12:57:06  paehlke
// Final version for masters thesis
//
// Revision 1.2  1998/03/31 11:31:57  paehlke
// added bidirectional stack algorithm
// changed stack organization: heap of heaps instead of single heap
// adapted to gcc 2.8.0
//
// Revision 1.1  1998/01/27 21:28:29  paehlke
// Initial revision
//

#ifndef _ASSERT_H
#include <assert.h>
#endif

#ifndef _HEAP_H
#include "heap.h"
#endif

#ifndef _QUICKSORT_H
#include "quicksort.h"
#endif

// --- BEGIN OF PUBLIC INTERFACE

// CODE must be an algebraic code over the Galois field GF
// with at least the following member functions:
//   int get_n(), int get_k(), int get_d() to get the code parameters
//   int Correct (GF * dest, const GF * src, const int * Iv, int cv)
//     which does error-and-erasure decoding and returns >=0 if successful

template<class GF> class RSCode;
template<class GF, class CODE> class FanoNode;
template<class GF, class CODE> class StackNode;
template<class GF, class CODE> class Stack;

template<class GF, class CODE> class TreeDecoder {
public:

  TreeDecoder (CODE &);
  ~TreeDecoder ();

  // FanoDecode() tries to do sequential soft-decision decoding using
  // the Fano algorithm with quantization parameter delta.
  // alpha[i][j] must contain the logarithmic likelihood
  // log (P[src[i] == GF(j)]) for i = 0..n-1 and j = 0..GF::q()-1.
  // maxnodes is the maximum number of (not necessarily different)
  // nodes to be visited.

  // isknown must either be NULL or point to an array of size k.
  // In the latter case, if isknown[i] is != 0, dest[i] is assumed
  // to have already been set correctly.

  int FanoCorrect (GF * dest, const GF * src, const double * const * alpha,
		   double delta, int maxnodes, int reverse = 0,
		   const int * isknown = NULL);

  inline int FanoDecode (GF * dest, const GF * src,
			 const double * const * alpha, double delta,
			 int maxnodes, int reverse = 0,
			 const int * isknown = NULL);

  // StackDecode() tries to do sequential soft-decision decoding using
  // the Stack algorithm.
  // alpha[i][j] must contain the logarithmic likelihood
  // log (P[src[i] == GF(j)]) for i = 0..n-1 and j = 0..GF::q()-1.

  // Before StackDecode() is called for the first time,
  // the stack must allocated by AllocateStack(maxnodes), where
  // maxnodes is the maximum number of bytes to be visited.
  // DeleteStack() frees the memory allocated by InitStack()
  // CAUTION: the stack occupies 50-100 bytes per visited node!

  // isknown must either be NULL or point to an array of size k.
  // In the latter case, if isknown[i] is != 0, dest[i] is assumed
  // to have already been set correctly.

  void AllocateStack (int maxnodes);
  void DeleteStack ();

  int StackCorrect (GF * dest, const GF * src,
		    const double * const * alpha,
		    int reverse = 0, const int * isknown = NULL);

  inline int StackDecode (GF * dest, const GF * src,
			  const double * const * alpha,
			  int reverse = 0,
			  const int * isknown = NULL);

  inline int BidirStackCorrect (GF * dest, const GF * src,
				const double * const * alpha,
				int maxtermnodes = 1,
				const int * isknown = NULL);

  inline int BidirStackDecode (GF * dest, const GF * src,
			       const double * const * alpha,
			       int maxtermnodes = 1,
			       const int * isknown = NULL);

  // GetNodes() returns the number of nodes visited during the last
  // call to FanoDecode() or StackDecode()

  int GetNodes () const;

  // --- END OF PUBLIC INTERFACE

private:

  typedef double * pDouble;
  typedef int * pInt;
  typedef RSState<GF> State;

  // general declarations for all algorithms

  CODE * code;
  int k, n; // CODE parameters
  State state; // encoder/decoder state
  int nodecount; // visited nodes counter
  pDouble * metric; // Fano metric table
  const int * isknown; // isknown[i] != 0 means that dest[i] is already known
  GF * infobuf; // buffer for one information word
  GF * codebuf; // buffer for one code word

  void SetupMetricTable (const double * const * alpha);

  // special declarations for Fano algorithm

  typedef FanoNode<GF,CODE> FN;

  int maxnodes; // max. # of visited nodes for Fano decoding
  FN * fanonode; // fanonode[i], i=0..k
  double thresh, delta;

  int FanoAlgorithm (const FN &, double bbm, int reverse);

  // auxiliary declarations for stack algorithm

  typedef class StackNode<GF,CODE> SN;
  typedef class Stack<GF,CODE> ST;

  int stackmaxnodes; // max. # of visited nodes for stack decoding
  Heap<ST> * stackheap;
  Heap<ST> * revstackheap;
  HeapElement<ST> * stack; // stack[i], i=0..k-1
  HeapElement<ST> * revstack;
  HeapElement<ST> stack0;  // containing only stacknode[0]
  HeapElement<ST> revstack0;
  int stacksize;        // size of each stack[i] and revstack[i]
  SN * stacknode;
  int stacknodes;    // size of stacknode[]
  int nextstacknode; // next unused index in stacknode[]

  void DumpStackPath (GF * dest, const SN & sn);
  const SN * TakeSmallestStackNode (int reverse);
  SN * NewStackNode (const SN & prev, const GF & sym,
			double bm, int reverse);
  int InsertStackNode (const SN & sn, int reverse);

  // auxiliary declarations for bidirectional stack algorithm

  const SN * ConcatenatePaths (const SN &, const SN &, int reverse);
};

template<class GF,class CODE> class FanoNode : public RSCode<GF>::State
{
public:

  int * sym; // elements of GF, ordered by descending bm[sym[i]]
  double * bm; // bm[i] = branch metric of GF(i)
  double m;  // path metric
};

template<class GF, class CODE> TreeDecoder<GF,CODE>::TreeDecoder
(CODE & _code)
{
  int i;
  code = &_code;
  k = code->get_k();
  n = code->get_n();
  infobuf = new GF [k];
  codebuf = new GF [n];
  code->AllocateState (state);
  fanonode = new FN [k+1];
  for (i = 0; i <= k; i++) {
    code->AllocateState (fanonode[i]);
    fanonode[i].sym = new int [GF::q()];
    fanonode[i].bm = new double [GF::q()];
  }
  metric = new pDouble [n];
  for (i = 0; i < n; i++)
    metric[i] = new double [GF::q()];
  stack = NULL;
}

template<class GF, class CODE> TreeDecoder<GF,CODE>::~TreeDecoder ()
{
  int i;
  // free all allocated memory
  if (stack != NULL)
    DeleteStack ();
  for (i = 0; i < n; i++)
    delete [] metric[i];
  delete [] metric;
  for (i = 0; i <= k; i++) {
    delete [] fanonode[i].bm;
    delete [] fanonode[i].sym;
  }
  delete [] fanonode;
  delete [] codebuf;
  delete [] infobuf;
}

template<class GF, class CODE> void TreeDecoder<GF,CODE>::SetupMetricTable
(const double * const * alpha)
{
  int i,j;
  double corr = (1.0 - code->get_r()) * GF::ldq(); // correction term
  for (i = 0; i < n; i++) {
    const double * ai = alpha[i];
    double * mi = metric[i];
    for (j = 0; j < GF::q(); j++)
      mi[j] = ai[j] + corr;
  }
}

template<class GF, class CODE> inline int
TreeDecoder<GF,CODE>::GetNodes() const
{
  return nodecount;
}

// Sequential decoding: Fano algorithm

template<class GF, class CODE> int TreeDecoder<GF,CODE>::FanoAlgorithm
(const FanoNode<GF,CODE> & fn, double bbm, int reverse)
{
  int i, res;
  if (isknown != NULL && isknown[fn.ipos]) {
    // infobuf[ipos] is already known
    double bm = code->KnownSymbolBranchMetric (state, reverse);
    FN & fn1 = fanonode[fn.ipos + 1];
    code->MoveForward (fn1, fn, infobuf[fn.ipos], reverse);
    fn1.m = fn.m + bm;
    if (fn.m - thresh < delta) {
      // tighten threshold
      thresh += delta * floor ((fn1.m - thresh) / delta);
    }
    if (code->IsTerminalState (fn1, reverse)) // terminal node reached
      res = 0;
    else
      res = FanoAlgorithm (fn1, bm, reverse);
  }
  else {
    // infobuf[ipos] is not yet known
    for (i = 0; i < GF::q(); i++) {
      fn.sym[i] = i;
      fn.bm[i] = code->BranchMetric (fn, GF(i), (const double **) metric,
				     reverse);
    }
    QuickSortDown (fn.sym, fn.bm, 0, GF::q()-1);
    res = -1;
    i = 0;
    while (i < GF::q()) {
      if (++nodecount > maxnodes) { // too many nodes
	res = -2;
	break;
      }
      GF sym = GF (fn.sym[i]);
      double bm = fn.bm[(int) sym];
#ifdef VERBOSE
      cerr << "ipos=" << fn.ipos << ", sym[" << i << "]=" << (int) sym
	   << ", bm=" << bm << ", m=" << fn.m << ", thresh=" << thresh << endl;
#endif
      if (fn.m + bm < thresh) {
	if (fn.m - bbm < thresh) {
	  // lower threshold and look forward to best node again
	  thresh -= delta;
	  i = 0;
	}
	else { // move back
	  res = -1;
	  break;
	}
      }
      else {
	// move forward
	infobuf[fn.ipos] = sym;
	FN & fn1 = fanonode[fn.ipos + 1];
	code->MoveForward (fn1, fn, sym, reverse);
	fn1.m = fn.m + bm;
	if (fn.m - thresh < delta) {
	  // tighten threshold
	  thresh += delta * floor ((fn1.m - thresh) / delta);
	}
	if (code->IsTerminalState (fn1, reverse)) {
	  // terminal node reached
	  res = 0;
	  break;
	}
	else {
	  res = FanoAlgorithm (fn1, bm, reverse);
	  if ((res == 0) || (res == -2)) {
	    // optimal path found or maximum node count exceeded
	    break;
	  }
	  else
	    i++; // look forward to next best node
	}
      } // if (fn.m + bm < thresh)
    } // while (i < GF::q())
  } // if (isknown)
  return res;
}

template<class GF, class CODE> int
TreeDecoder<GF,CODE>::FanoCorrect (GF * dest, const GF * src,
				   const double * const * alpha,
				   double _delta, int _maxnodes,
				   int reverse, const int * _isknown)
{
  int i, res;
  SetupMetricTable (alpha);
  delta = _delta;
  maxnodes = _maxnodes;
  isknown = _isknown;
  if (isknown != NULL) {
    // copy known information symbols to infobuf[]
    for (i=0; i<k; i++)
      infobuf[i] = dest[i];
  }
  // Initialize root node
  FN & fn = fanonode[0];
  code->InitialState (fn, reverse);
  fn.m = 0.0;
  thresh = 0.0;
  nodecount = 0;
  // The following expression raises an "internal compiler error"
  // when compiled with g++ 2.8.1:
  // res = FanoAlgorithm (fn, HUGE_VAL, reverse);
  res = FanoAlgorithm (fn, 1e300, reverse);
#ifdef VERBOSE
  cerr << "res=" << res << ", nodecount=" << nodecount << endl;
#endif
  if (res >= 0)
    code->EncodeConvolutional (dest, infobuf);
  else { // decoder failure
    for (i=0; i<n; i++)
      dest[i] = src[i];
  }
  return res;
}

template<class GF, class CODE> int
TreeDecoder<GF,CODE>::FanoDecode (GF * dest, const GF * src,
				  const double * const * alpha,
				  double _delta, int _maxnodes,
				  int reverse, const int * _isknown)
{
  int res = FanoCorrect (codebuf, src, alpha, _delta, _maxnodes,
			 reverse, _isknown);
  // in case of decoder failure, src[] has been copied to codebuf[]
  (void) code->ExtractInformationWord (dest, codebuf);
  return res;
}

// Sequential decoding: stack algorithm

template<class GF, class CODE> class StackNode : public RSCode<GF>::State
{
public:

  GF sym;
  double metric;
  const StackNode<GF,CODE> * prev; // previous node in code tree
}; // size: 24 bytes + sizeof(buf)

template<class GF, class CODE> int operator <
(const StackNode<GF,CODE> & sn1, const StackNode<GF,CODE> & sn2)
{
  return (sn1.metric > sn2.metric);
}

template<class GF, class CODE> class Stack
{
public:

  typedef StackNode<GF,CODE> SN;
  SimpleHeap<const SN> * heap;

  Stack () { heap = NULL; }
};

template<class GF, class CODE> int operator <
(const Stack<GF,CODE> & st1, const Stack<GF,CODE> & st2)
{
  assert ((st1.heap != NULL) && (st1.heap->get_size() > 0)
	  && (st2.heap != NULL) && (st2.heap->get_size() > 0));
  return (st1.heap->SmallestElement() < st2.heap->SmallestElement());
}

template<class GF, class CODE> void TreeDecoder<GF,CODE>::AllocateStack
(int _maxnodes)
{
  assert ((stack == NULL) && (_maxnodes > 0));
  int i;
  stackmaxnodes = _maxnodes;
  // at most GF::q() elements are added for each node visited
  stacknodes = GF::q() * _maxnodes;
  stacknode = new SN [stacknodes];
  for (i = 0; i < stacknodes; i++)
    code->AllocateState (stacknode[i]);
  stacksize = _maxnodes;
  stack = (new HeapElement<ST> [k]);
  for (i = 0; i < k; i++)
    stack[i].Data.heap = new SimpleHeap<const SN> (NULL, 0, stacksize);
  stack0.Data.heap = new SimpleHeap<const SN> (NULL, 0, 1);
  stackheap = new Heap<ST> (NULL, 0, k+1);
  revstack = (new HeapElement<ST> [k]);
  for (i = 0; i < k; i++)
    revstack[i].Data.heap = new SimpleHeap<const SN> (NULL, 0, stacksize);
  revstack0.Data.heap = new SimpleHeap<const SN> (NULL, 0, 1);
  revstackheap = new Heap<ST> (NULL, 0, k+1);
}

template<class GF, class CODE> void TreeDecoder<GF,CODE>::DeleteStack ()
{
  int i;
  assert (stackheap != NULL);
  delete revstackheap; revstackheap = NULL;
  delete revstack0.Data.heap; revstack0.Data.heap = NULL;
  for (i = 0; i < k; i++)
    delete revstack[i].Data.heap;
  delete [] revstack; revstack = NULL;
  delete stackheap; stackheap = NULL;
  delete stack0.Data.heap; stack0.Data.heap = NULL;
  for (i = 0; i < k; i++)
    delete stack[i].Data.heap;
  delete [] stack; stack = NULL;
  delete [] stacknode; stacknode = NULL;
}

template<class GF, class CODE> void TreeDecoder<GF,CODE>::DumpStackPath
(GF * dest, const StackNode<GF,CODE> & _sn)
{
  const SN * sn = &_sn;
  while (sn->prev != NULL) {
    const SN * psn = sn->prev;
    assert ((psn->ipos >= 0) && (psn->ipos < k));
    dest[psn->ipos] = sn->sym;
    sn = psn;
  }
}

template<class GF, class CODE> inline const StackNode<GF,CODE> *
TreeDecoder<GF,CODE>::TakeSmallestStackNode (int reverse)
{
  const SN * sn;
  if (reverse) {
    if (revstackheap->IsEmpty())
      sn = NULL;
    else {
      sn = &revstackheap->SmallestElement().Data.heap->SmallestElement();
#ifdef VERBOSE
      cerr << "=> revtop = (ipos=" << sn->ipos
	   << ", sym=" << (int) sn->sym
	   << ", metric=" << sn->metric << ")\n";
#endif
      if (sn->prev == NULL) { // initial element
	assert ((sn == &revstack0.Data.heap->SmallestElement())
		&& (&revstack0 == &revstackheap->SmallestElement()));
	revstack0.Data.heap->DeleteSmallestElement();
	revstackheap->DeleteElement (revstack0);
      }
      else { // sn->prev != NULL
	int ipos = sn->prev->ipos;
	assert ((sn == &revstack[ipos].Data.heap->SmallestElement())
		&& (&revstack[ipos] == &revstackheap->SmallestElement()));
	revstack[ipos].Data.heap->DeleteSmallestElement ();
	if (revstack[ipos].Data.heap->IsEmpty())
	  revstackheap->DeleteElement (revstack[ipos]);
	else
	  revstackheap->ElementModified (revstack[ipos]);
      } // if (sn->prev == NULL)
    } // if (revstackheap->IsEmpty())
  }
  else { // !reverse
    if (stackheap->IsEmpty())
      sn = NULL;
    else {
      sn = &stackheap->SmallestElement().Data.heap->SmallestElement();
#ifdef VERBOSE
      cerr << "=> top = (ipos=" << sn->ipos
	   << ", sym=" << (int) sn->sym
	   << ", metric=" << sn->metric << ")\n";
#endif
      if (sn->prev == NULL) { // initial element
	assert ((sn == &stack0.Data.heap->SmallestElement())
		&& (&stack0 == &stackheap->SmallestElement()));
	stack0.Data.heap->DeleteSmallestElement();
	stackheap->DeleteElement (stack0);
      }
      else { // sn->prev != NULL
	int ipos = sn->prev->ipos;
	assert ((sn == &stack[ipos].Data.heap->SmallestElement())
		&& (&stack[ipos] == &stackheap->SmallestElement()));
	stack[ipos].Data.heap->DeleteSmallestElement ();
	if (stack[ipos].Data.heap->IsEmpty())
	  stackheap->DeleteElement (stack[ipos]);
	else
	  stackheap->ElementModified (stack[ipos]);
      } // if (sn->prev == NULL)
    } // if (stackheap->IsEmpty())
  } // if (reverse)
  return sn;
}

template<class GF, class CODE> inline StackNode<GF,CODE> *
TreeDecoder<GF,CODE>::NewStackNode
(const StackNode<GF,CODE> & prev, const GF & sym, double bm, int reverse)
{
  if (nextstacknode >= stacknodes)
    return NULL;
  else {
    SN & sn = stacknode[nextstacknode];
    code->MoveForward (sn, prev, sym, reverse);
    sn.sym = sym;
    sn.metric = prev.metric + bm;
    sn.prev = &prev;
#ifdef VERBOSE
    cerr << "==> sn[" << nextstacknode << "] = (ipos=" << sn.ipos
	 << ", sym=" << (int) sn.sym << ", bm=" << bm
	 << ", metric=" << sn.metric << ")\n";
#endif
    nextstacknode++;
    return &sn;
  }
}

template<class GF, class CODE> inline int
TreeDecoder<GF,CODE>::InsertStackNode
(const StackNode<GF,CODE> & sn, int reverse)
{
  assert (sn.prev != NULL);
  int ipos = sn.prev->ipos;
  if (reverse) {
    if (revstack[ipos].Data.heap->IsFull())
      revstack[ipos].Data.heap->reduce_size (1); // ugly but working...
    int wasempty = revstack[ipos].Data.heap->IsEmpty();
    revstack[ipos].Data.heap->InsertElement (sn);
    if (wasempty)
      revstackheap->InsertElement (revstack[ipos]);
    else
      revstackheap->ElementModified (revstack[ipos]);
    return 0;
  }
  else { // !reverse
    if (stack[ipos].Data.heap->IsFull())
      stack[ipos].Data.heap->reduce_size (1); // see above
    int wasempty = stack[ipos].Data.heap->IsEmpty();
    stack[ipos].Data.heap->InsertElement (sn);
    if (wasempty)
      stackheap->InsertElement (stack[ipos]);
    else
      stackheap->ElementModified (stack[ipos]);
    return 0;
  }
}

template<class GF, class CODE> int
TreeDecoder<GF,CODE>::StackCorrect (GF * dest, const GF * src,
				    const double * const * alpha,
				    int reverse, const int * _isknown)
{
  int i, res;
  SetupMetricTable (alpha);
  isknown = _isknown;
  // clear all stacks
  for (i = 0; i < k; i++)
    stack[i].Data.heap->Empty();
  stackheap->Empty();
  // set initial state
  SN & sn0 = stacknode[0];
  nextstacknode = 1;
  code->InitialState (sn0, reverse);
  sn0.metric = 0.0;
  sn0.prev = NULL;
  stack0.Data.heap->Empty().InsertElement (sn0);
  stackheap->InsertElement (stack0);
  nodecount = 0;
  for (;;) {
    if (stackheap->IsEmpty()) {
      // stack emptied, no terminal node found
      res = -1;
      break;
    }
    const SN & sn = *TakeSmallestStackNode (0);
    if (code->IsTerminalState (sn, reverse)) {
      // terminal node reached
      DumpStackPath (infobuf, sn);
      code->EncodeConvolutional (dest, infobuf);
#ifdef VERBOSE
      cerr << "nodecount=" << nodecount << endl;
#endif
      res = 0;
      break;
    }
    else if (++nodecount > stackmaxnodes) {
      // too many nodes visited
      res = -2;
      break;
    }
    else if (nextstacknode >= stacknodes) {
      // stack full
      res = -3;
      break;
    }
    else {
      assert ((sn.ipos >= 0) && (sn.ipos < k));
      // extend the path with the largest metric
      if (isknown != NULL && isknown[sn.ipos]) {
	// dest[sn.ipos] is already known
 	double bm = code->KnownSymbolBranchMetric (sn, reverse);
	SN * nsn = NewStackNode (sn, dest[sn.ipos], bm, reverse);
	assert (nsn != NULL);
	InsertStackNode (*nsn, 0);
      }
      else {
	// dest[sn.ipos] is not yet known, examine all successors
	for (i = 0; i < GF::q(); i++) {
	  double bm = code->BranchMetric (sn, GF(i),
					  (const double **)metric, reverse);
	  if (sn.metric + bm + code->MaximumRemainingMetric (sn, reverse)
	      >= 0) {
	    SN * nsn = NewStackNode (sn, GF(i), bm, reverse);
	    if (nsn == NULL) // stack full
	      break;
	    else
	      InsertStackNode (*nsn, 0);
	  }
	} // for (i)
      } // if (isknown)
    } // if (terminal state)
  } // for (;;)
  if (res < 0) {
#ifdef VERBOSE
    cerr << "decoder failure!\n";
#endif
    for (i = 0; i < n; i++)
      dest[i] = src[i];
  }
  return res;
}

template<class GF, class CODE> int
TreeDecoder<GF,CODE>::StackDecode (GF * dest, const GF * src,
				   const double * const * alpha,
				   int reverse, const int * _isknown)
{
  int res = StackCorrect (codebuf, src, alpha, reverse, _isknown);
  // in case of decoder failure, src[] has been copied to codebuf[]
  (void) code->ExtractInformationWord (dest, codebuf);
  return res;
}

// bidirectional stack algorithm

template<class GF, class CODE> const StackNode<GF,CODE> *
TreeDecoder<GF,CODE>::ConcatenatePaths
(const StackNode<GF,CODE> & _sn1, const StackNode<GF,CODE> & _sn2,
 int reverse)
{
  const SN * sn1, * sn2;
  sn1 = &_sn1;
  sn2 = &_sn2;
  while (sn2->prev != NULL) {
    assert (sn1->ipos == sn2->prev->ipos);
    double bm = code->BranchMetric (*sn1, sn2->sym,
				    (const double **) metric, reverse);
    sn1 = NewStackNode (*sn1, sn2->sym, bm, reverse);
    if (sn1 == NULL)
      break;
    sn2 = sn2->prev;
  }
  return sn1;
}

template<class GF, class CODE> int
TreeDecoder<GF,CODE>::BidirStackCorrect (GF * dest, const GF * src,
					 const double * const * alpha,
					 int maxtermnodes,
					 const int * _isknown)
{
  int reverse, res, i;
  int terminalnodesfound = 0; // number of terminal nodes found so far
  const SN * terminal = NULL; // best terminal node found so far
  double terminalmetric = 0.0;
  SetupMetricTable (alpha);
  isknown = _isknown;
  // initialize forward stack
  for (i = 0; i < k-1; i++)
    stack[i].Data.heap->Empty();
  stackheap->Empty();
  SN & sn0 = stacknode[0];
  code->InitialState (sn0, 0);
  sn0.metric = 0.0;
  sn0.prev = NULL;
  stack0.Data.heap->Empty().InsertElement (sn0);
  stackheap->InsertElement (stack0);
  // initialize backward stack
  for (i = 1; i < k; i++)
    revstack[i].Data.heap->Empty();
  revstackheap->Empty();
  SN & sn1 = stacknode[1];
  code->InitialState (sn1, 1);
  sn1.metric = 0.0;
  sn1.prev = NULL;
  revstack0.Data.heap->Empty().InsertElement (sn1);
  revstackheap->InsertElement (revstack0);
  nextstacknode = 2;
  nodecount = 0;
  reverse = 0;
  res = -10;
  while (terminalnodesfound < maxtermnodes) {
    if (stackheap->IsEmpty() && revstackheap->IsEmpty()) {
      // both stacks emptied
      res = -1;
      break;
    }
    // take the node with the greatest metric from the active stack
    const SN * sn = TakeSmallestStackNode (reverse);
    if (sn == NULL) { // the active stack was empty
      reverse = !reverse;
      continue;
    }
    else if ((sn->metric + code->KnownSymbolBranchMetric (*sn, reverse)
	      + code->MaximumRemainingMetric (*sn, reverse))
	     < terminalmetric) {
      // empty the active stack
      if (reverse) {
	for (i = 1; i < k; i++)
	  revstack[i].Data.heap->Empty();
	revstackheap->Empty();
      }
      else {
	for (i = 0; i < k-1; i++)
	  stack[i].Data.heap->Empty();
	stackheap->Empty();
      }
      reverse = !reverse;
      continue;
    }
    else if (++nodecount > stackmaxnodes) {
      // too many nodes visited
      res = -2;
      break;
    }
    else if (nextstacknode >= stacknodes) {
      // stack full
      res = -3;
      break;
    }
    else {
      // extend the path with the greatest metric from the active stack
      assert ((sn->ipos >= 0) && (sn->ipos < k));
      if (isknown != NULL && isknown[sn->ipos]) {
	// dest[sn->ipos] is already known
	double bm = code->KnownSymbolBranchMetric (*sn, reverse);
	SN * nsn = NewStackNode (*sn, dest[sn->ipos], bm, reverse);
	assert (nsn != NULL);
	if (code->IsTerminalState (*nsn, reverse)) {
	  if (nsn->metric >= terminalmetric) {
	    terminalnodesfound++;
	    terminal = nsn;
	    terminalmetric = nsn->metric;
	  }
	}
	else
	  InsertStackNode (*nsn, reverse);
      }
      else {
	// dest[sn.ipos] is not yet known, first try to find a tunnel
	// determine the complementary stack
	ST * compstack;
	if (reverse)
	  compstack = &stack[sn->ipos].Data;
	else
	  compstack = &revstack[sn->ipos].Data;
	int stackfull = 0;
	for (i = 0; i < compstack->heap->get_size(); i++) {
	  const SN * csn = &compstack->heap->Element(i);
	  double tm = code->TunnelMetric (*sn, *csn,
					  (const double **) metric, reverse);
	  if (sn->metric + csn->metric + tm >= terminalmetric) {
	    const SN * nsn = ConcatenatePaths (*sn, *csn, reverse);
	    if (nsn == NULL) {
	      stackfull = 1;
	      break;
	    }
	    else {
	      assert (fabs (sn->metric + csn->metric + tm - nsn->metric)
		      < 1e-5);
	      if (code->IsTerminalState (*nsn, reverse)) {
		assert (nsn->metric >= terminalmetric);
		terminalnodesfound++;
		terminal = nsn;
		terminalmetric = nsn->metric;
	      }
	      else
		InsertStackNode (*nsn, reverse);
	    }
	  }
	} // for (i)
	if (!stackfull) {
	  // examine all successors
	  for (i = 0; i < GF::q(); i++) {
	    double bm = code->BranchMetric (*sn, GF(i),
					    (const double **)metric, reverse);
	    if ((sn->metric + bm + code->MaximumRemainingMetric
		 (*sn, reverse)) >= terminalmetric) {
	      SN * nsn = NewStackNode (*sn, GF(i), bm, reverse);
	      if (nsn == NULL)
		break;
	      else if (code->IsTerminalState (*nsn, reverse)) {
		assert (nsn->metric >= terminalmetric);
		terminalnodesfound++;
		terminal = nsn;
		terminalmetric = nsn->metric;
	      }
	      else
		InsertStackNode (*nsn, reverse);
	    }
	  } // for (i)
	} // if (!stackfull)
      } // if (isknown[sn->ipos])
    } // if
    reverse = !reverse;
  } // for (;;)
  if (terminal != NULL) {
    // dump the path of the best terminal node found
    DumpStackPath (infobuf, *terminal);
    code->EncodeConvolutional (dest, infobuf);
#ifdef VERBOSE
    cerr << "nodecount=" << nodecount << endl;
#endif
    return 0;
  }
  else {
#ifdef VERBOSE
    cerr << "decoder failure!\n";
#endif
    for (i = 0; i < n; i++)
      dest[i] = src[i];
    return res;
  }
}

template<class GF, class CODE> int
TreeDecoder<GF,CODE>::BidirStackDecode (GF * dest, const GF * src,
					const double * const * alpha,
					int maxtermnodes,
					const int * _isknown)
{
  int res = BidirStackCorrect (codebuf, src, alpha, maxtermnodes, _isknown);
  // in case of decoder failure, src[] has been copied to codebuf[]
  (void) code->ExtractInformationWord (dest, codebuf);
  return res;
}

#endif
