#ifndef _VECTOR_H
#define _VECTOR_H

// A simple vector class which acts mainly as syntactic sugar for
// double[].
// (optimized for speed, not for elegance and stability...

// (C) 1997-1998 Frank Paehlke <frank@paehlke.de>
// Created: 12/17/1996
// $Id: vector.h,v 1.1 2003/10/14 08:49:02 frank Exp $

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// $Log: vector.h,v $
// Revision 1.1  2003/10/14 08:49:02  frank
// moved from . to src/
//
// Revision 1.4  2003/10/14 08:41:58  frank
// ANSI C++ / gcc 3.x compatibility changes
//
// Revision 1.3  2002/08/02 14:45:06  paehlke
// added GPL comment
//
// Revision 1.2  1998/05/28 12:57:07  paehlke
// Final version for masters thesis
//
// Revision 1.1  1997/06/20 12:55:01  frank
// Initial revision
//

#ifndef _ASSERT_H
#include <assert.h>
#endif

#ifndef _CPP_IOSTREAM
#include <iostream>
#endif

#ifndef _MATH_H
#include <math.h>
#endif

// --- BEGIN OF PUBLIC INTERFACE ---

class Vector
{
public:

  Vector (int n, double *); // pointer copy
  Vector (Vector &); // pointer copy

  Vector & allocate (); // allocate memory
  Vector & allocate (int);

  Vector & deallocate ();

  double & operator [] (int n);
  const double & operator [] (int n) const;
  int get_n () const;

  int isnull () const;	// tests if pointer is initialized

  Vector & operator = (Vector &);	// pointer copy
  const Vector & operator = (const Vector &);

  Vector & copy (const Vector &);	// elementwise copy
  Vector & copy (const double *);

  Vector & operator += (const Vector &);
  Vector & operator -= (const Vector &);
  Vector & operator *= (double);
  Vector & operator /= (double);

  friend double operator * (const Vector &, const Vector &); // scalar product

  friend double sqrabs (const Vector &);
  friend double abs (const Vector &);

  friend double sqrdist (const Vector &, const Vector &);
  friend double dist (const Vector &, const Vector &);

  friend std::ostream & operator << (std::ostream &, const Vector &);

// --- END OF PUBLIC INTERFACE --

protected:

  double * c;	// the coefficients
  int n;
};

inline int Vector::isnull () const
{
  return c == NULL;
}

inline Vector::Vector (int _n = 0, double * _c = NULL)
{
  assert (_n >= 0);
  c = _c;
  n = _n;
}

inline Vector::Vector (Vector & v)
{
  c = v.c;
  n = v.n;
}

inline Vector & Vector::allocate ()
{
  if (n > 0 && isnull ())
    c = new double [n];
  return *this;
}

inline Vector & Vector::allocate (int _n)
{
  if (isnull ()) {
    n = _n;
    c = new double [n];
  }
  else if (n != _n) {
    delete [] c;
    n = _n;
    c = new double [n];
  }
  return *this;
}

inline Vector & Vector::deallocate ()
{
  delete [] c;
  c = NULL;
  return * this;
}

inline double & Vector::operator [] (int i)
{
  assert (!isnull() && i >= 0 && i < n);
  return c[i];
}

inline const double & Vector::operator [] (int i) const
{
  assert (!isnull() && i >= 0 && i < n);
  return c[i];
}

inline int Vector::get_n () const
{
  return n;
}

inline Vector & Vector::operator = (Vector & v)
{
  if (!isnull ())
    delete [] c;
  c = v.c;
  n = v.n;
  return *this;
}

inline const Vector & Vector::operator = (const Vector & v)
{
  if (!isnull ())
    delete [] c;
  c = v.c;
  n = v.n;
  return *this;
}

inline Vector & Vector::copy (const Vector & v)
{
  int i;
  if (isnull ()) {
    n = v.n;
    c = new double [n];
  }
  else if (n != v.n) {
    delete [] c;
    n = v.n;
    c = new double [n];
  }
  for (i=0; i<n; i++)
    c[i] = v.c[i];
  return *this;
}

inline Vector & Vector::operator += (const Vector & v)
{
  assert (n==v.n && (n==0 || (!isnull() && !v.isnull())));
  int i;
  for (i=0; i<n; i++)
    c[i] += v.c[i];
  return *this;
}

inline Vector & Vector::operator -= (const Vector & v)
{
  assert (n==v.n && !isnull() && !v.isnull());
  int i;
  for (i=0; i<n; i++)
    c[i] -= v.c[i];
  return *this;
}

inline Vector & Vector::operator *= (double d)
{
  assert (n==0 || !isnull());
  int i;
  for (i=0; i<n; i++)
    c[i] *= d;
  return *this;
}

inline Vector & Vector::operator /= (double d)
{
  assert (n==0 || !isnull());
  int i;
  for (i=0; i<n; i++)
    c[i] /= d;
  return *this;
}

inline double operator * (const Vector & v1, const Vector & v2)
{
  assert (v1.n==v2.n && (v1.n==0 || (!v1.isnull() && !v2.isnull())));
  int i;
  double d = 0.0;
  for (i=0; i<v1.n; i++)
    d += v1[i] * v2[i];
  return d;
}

inline double sqrabs (const Vector & v)
{
  assert (v.n==0 || !v.isnull());
  int i;
  double d = 0.0;
  for (i=0; i<v.n; i++)
    d += v.c[i] * v.c[i];
  return d;
}

inline double abs (const Vector & v)
{
  return sqrt (sqrabs (v));
}

inline double sqrdist (const Vector & v1, const Vector & v2)
{
  assert (v1.n==v2.n && (v1.n==0 || (!v1.isnull() && !v2.isnull())));
  int i;
  double d = 0.0;
  for (i=0; i<v1.n; i++) {
    double di = v2[i] - v1[i];
    d += di * di;
  }
  return d;
}

inline double dist (const Vector & v1, const Vector & v2)
{
  return sqrt (sqrdist (v1, v2));
}

inline std::ostream & operator << (std::ostream & os, const Vector & v)
{
  int i;
  assert (v.n==0 || !v.isnull());
  os << '[';
  if (v.n > 0) {
    os << v.c[0];
    for (i=1; i<v.n; i++)
      os << ',' << v.c[i];
  }
  return os << ']';
}

#endif
