#ifndef _XRSCODE_H
#define _XRSCODE_H

// A class for encoding and decoding of expurgated Reed-Solomon codes
// where the code symbols are from the Galois field GF = GF(q)
// and the code word length is n <= q - 1

// (C) 1997-1998 Frank Paehlke <frank@paehlke.de>
// Created: 1997/09/22
// $Id: xrscode.h,v 1.2 2003/10/14 09:32:23 frank Exp $

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// $Log: xrscode.h,v $
// Revision 1.2  2003/10/14 09:32:23  frank
// fixes to avoid GCC 3.3 "implicit typename" warnings
//
// Revision 1.5  2002/08/02 14:45:06  paehlke
// added GPL comment
//
// Revision 1.4  2002/08/02 12:32:27  paehlke
// Imported from RCS to CVS
// On 2000-08-10: added "typedef RSCode<GF>::State State"
//
// Revision 1.3  1998/05/28 13:03:10  paehlke
// Final version for masters thesis
//
// Revision 1.2  1997/11/27 15:05:32  paehlke
// PRSCode made subclass of RSCode, Stack decoding added
//
// Revision 1.1  1997/10/19 12:31:36  frank
// Initial revision
//

#ifndef _ASSERT_H
#include <assert.h>
#endif

#ifndef _RSCODE_H
#include "rscode.h"
#endif

// --- BEGIN OF PUBLIC INTERFACE

template<class GF> class XRSCode : public RSCode<GF> {

public:

  // constructor and destructor

  // XRSCode(d,xd,n) creates a expurgated Reed-Solomon code of length n
  // with code distance d and expurgation distance xd
  // which can correct up to t = (d-1) / 2 symbol errors.
  // XRSCode(d,xd) is equivalent to XRSCode(d,xd,GF::q()-1)

  XRSCode (int, int, int);
  XRSCode (int, int);
  ~XRSCode ();

  // code parameters

  int get_xd () const; // get expurgation distance

  // encoding and decoding

  int ExtractInformationWord (GF * dest, const GF * src);

  // When calling Encode(), dest must point to an array of length n,
  // and src must point to an array of length k.
  // src[] and dest[] must not overlap.

  void Encode (GF * dest, const GF * src) const;

  // When calling Decode(), dest must point to an array of length k,
  // and src must point to an array of length n.
  // src[] and dest[] must not overlap.

  int Decode (GF * dest, const GF * src, const int * Iv = NULL, int cv = 0);

  // support for sequential decoding

  typedef typename RSCode<GF>::State State;

  double BranchMetric (const State &, const GF &,
		       const double * const * metric,
		       int reverse) const;

  double KnownSymbolBranchMetric (const State &, int reverse) const;

  double TunnelMetric (const State &, const State &,
		       const double * const * metric, int reverse) const;

  double MaximumRemainingMetric (const State & state, int reverse) const;

  void MoveForward (State &, const State &, const GF &,
		    int reverse) const;

  void EncodeConvolutional (GF * dest, const GF * src) const;

  // --- END OF PUBLIC INTERFACE

protected:

  typedef double * pDouble;
  typedef int * pInt;

  int xd; // expurgation distance
  GF * statebuf; // encoder state buffer
  GF * infobuf; // buffer for one information word

  XRSCode ();
  void Initialize (int);
};

template<class GF> void XRSCode<GF>::Initialize (int _xd)
{
  assert (_xd > 0);
  xd = _xd;
  k -= k / xd; // every xd-th row of the generator matrix is deleted
  infobuf = new GF [k];
  statebuf = new GF [deggenpoly];
}

template<class GF> inline XRSCode<GF>::XRSCode (int _d, int _xd, int _n)
  : RSCode<GF> (_d, _n)
{
  Initialize (_xd);
}

template<class GF> inline XRSCode<GF>::XRSCode (int _d, int _xd)
  : RSCode<GF> (_d)
{
  Initialize (_xd);
}

template<class GF> XRSCode<GF>::XRSCode ()
{}

template<class GF> XRSCode<GF>::~XRSCode ()
{
  delete [] statebuf;
  delete [] infobuf;
}

template<class GF> inline int XRSCode<GF>::get_xd () const
{
  return xd;
}

template<class GF> inline int XRSCode<GF>::ExtractInformationWord
(GF * dest, const GF * src)
{
  int i, res;
  int ipos, cpos, xcount;
  for (i = 0; i < deggenpoly; i++)
    statebuf[i] = src[i];
  ipos = cpos = 0;
  xcount = (n-deggenpoly) % xd; // expurgation counter
  res = 0;
  while (cpos < n - deggenpoly) {
    GF nextsym = statebuf[0] / genpoly[0];
    for (i = 1; i < deggenpoly; i++)
      statebuf[i-1] = statebuf[i] - nextsym * genpoly[i];
    statebuf[deggenpoly-1]
      = src[cpos + deggenpoly] - nextsym * genpoly[deggenpoly];
    if (xcount != 0)
      dest[ipos++] = nextsym;
    else { // xcount == 0
      // the code symbol at cpos is redundant, so nextsym should be 0
      if (nextsym != 0)
	res = -1;
      xcount = xd;
    }
    cpos++; xcount--;
  }
  return res;
}

// Encode(dest,src) encodes the information word src[]
// and writes the resulting code word to dest[]

template<class GF> void XRSCode<GF>::Encode (GF * dest, const GF * src) const
{
  int ipos, cpos, xcount, i;
  // initialize dest(X) with 0
  for (i = 0; i < n; i++)
    dest[i] = 0;
  ipos = cpos = 0;
  xcount = (n-deggenpoly) % xd; // expurgation counter
  while (ipos < k) {
    // invariant: dest(X) == 0 mod genpoly(X)
    if (xcount == 0) {
      // the code symbol at cpos is redundant
      cpos++;
      xcount = xd - 1;
    }
    assert (cpos + deggenpoly < n);
    GF nextsym = src[ipos];
    for (i = 0; i <= deggenpoly; i++)
      dest[cpos + i] += nextsym * genpoly[i];
    ipos++; cpos++;
    xcount--;
  }
  assert ((cpos + deggenpoly == n) && (xcount == 0));
}

template<class GF> inline int XRSCode<GF>::Decode
(GF * dest, const GF * src, const int * Iv, int cv)
{
  int res = Correct (codebuf, src, Iv, cv);
  if (ExtractInformationWord (dest, codebuf) < 0)
    res = -10;
  return res;
}

// support for sequential decoding

// encoder state

template<class GF> double XRSCode<GF>::BranchMetric
(const typename RSCode<GF>::State & state,
 const GF & x, const double * const * metric,
 int reverse) const
{
  double bm;
  int ipos, cpos, i;
  ipos = state.ipos;
  if (reverse) {
    cpos = ipos + (n-k) - (k-1-ipos) / (xd-1);
    if (ipos == 0) {
      // next node is terminal
      assert ((cpos == deggenpoly) || (cpos == deggenpoly+1));
      bm = metric[cpos]
	[(int) (state.buf[deggenpoly-1] + x * genpoly[deggenpoly])];
      for (i = 1; i <= deggenpoly-1; i++) {
	bm += metric[cpos - i][(int) (state.buf[deggenpoly-1 - i]
				      + x * genpoly[deggenpoly - i])];
      }
      bm += metric[cpos - deggenpoly][(int) (x * genpoly[0])];
    }
    else if ((k-ipos) % (xd-1) == 0) {
      // the code symbol before cpos is redundant
      bm = (metric[cpos] [(int) (state.buf[deggenpoly - 1]
				 + x * genpoly[deggenpoly])]
	    + metric[cpos - 1][(int) (state.buf[deggenpoly - 2]
				      + x * genpoly[deggenpoly - 1])]);
    }
    else
      bm = metric[cpos]
	[(int) (state.buf[deggenpoly-1] + x * genpoly[deggenpoly])];
  }
  else { // !reverse
    cpos = ipos + (n-k-deggenpoly) - (k-1-ipos) / (xd-1);
    if (ipos == k-1) {
      // next node is terminal
      assert (cpos + deggenpoly == n-1);
      bm = metric[cpos][(int) (state.buf[0] + x * genpoly[0])];
      for (i = 1; i < deggenpoly; i++) {
	bm += metric[cpos + i][(int) (state.buf[i] + x * genpoly[i])];
      }
      bm += metric[cpos + deggenpoly][(int) (x * genpoly[deggenpoly])];
    }
    else if ((k-ipos) % (xd-1) == 1) {
      // the code symbol after cpos is redundant
      bm = (metric[cpos][(int) (state.buf[0] + x * genpoly[0])]
	    + metric[cpos + 1][(int) (state.buf[1] + x * genpoly[1])]);
    }
    else
      bm = metric[cpos][(int) (state.buf[0] + x * genpoly[0])];
  }
  return bm;
}

template<class GF> double XRSCode<GF>::KnownSymbolBranchMetric
(const typename RSCode<GF>::State & state, int reverse) const
{
  double bm = (1.0 - get_r()) * GF::ldq();
  if (reverse) {
    if (state.ipos == 0) // next node is terminal
      bm *= deggenpoly+1;
    else if ((k-state.ipos) % (xd-1) == 0)
      bm *= 2;
  } else { // !reverse
    if (state.ipos == k-1) // next node is terminal
      bm *= deggenpoly+1;
    else if ((k-state.ipos) % (xd-1) == 1)
      bm *= 2;
  }
  return bm;
}

template<class GF> double XRSCode<GF>::TunnelMetric
(const typename RSCode<GF>::State & state1,
 const typename RSCode<GF>::State & state2,
 const double * const * metric, int reverse) const
{
  int ipos, cpos, i;
  double bm;
  if (reverse) {
    assert (state2.ipos == state1.ipos + 1);
    ipos = state2.ipos;
  }
  else {
    assert (state1.ipos == state2.ipos + 1);
    ipos = state1.ipos;
  }
  cpos = ipos + (n-k-deggenpoly) - (k-1-ipos) / (xd-1);
  if ((k-ipos) % (xd-1) == 0) {
    // the code symbol before cpos is redundant
    if (reverse) {
      bm = metric[cpos][(int) (state2.buf[0] + state1.buf[1])];
      for (i = 1; i < deggenpoly-1; i++)
	bm += metric[cpos + i][(int) (state2.buf[i] + state1.buf[i+1])];
    }
    else {
      bm = metric[cpos][(int) (state1.buf[0] + state2.buf[1])];
      for (i = 1; i < deggenpoly-1; i++)
	bm += metric[cpos + i][(int) (state1.buf[i] + state2.buf[i+1])];
    }
  }
  else {
    bm = metric[cpos][(int) (state1.buf[0] + state2.buf[0])];
    for (i = 1; i < deggenpoly; i++)
      bm += metric[cpos + i][(int) (state1.buf[i] + state2.buf[i])];
  }
  return bm;
}

template<class GF> double XRSCode<GF>::MaximumRemainingMetric
(const typename RSCode<GF>::State & state, int reverse) const
{
  int ipos, cpos;
  double bm;
  if (reverse) {
    if (state.ipos == 0) // next node is terminal
      bm = 0;
    else {
      bm = (1.0 - get_r()) * GF::ldq();
      ipos = state.ipos - 1;
      cpos = ipos + (n-k) - (k-1-ipos) / (xd-1);
      if ((n - deggenpoly) % xd == 0)
	bm *= cpos;
      else
	bm *= (cpos + 1);
    }
  }
  else { // !reverse
    if (state.ipos == k-1) // next node is terminal
      bm = 0;
    else {
      bm = (1.0 - get_r()) * GF::ldq();
      ipos = state.ipos + 1;
      cpos = ipos + (n-k-deggenpoly) - (k-1-ipos) / (xd-1);
      bm *= (n - cpos);
    }
  }
  return bm;
}

template<class GF> void XRSCode<GF>::MoveForward
(typename RSCode<GF>::State & newstate,
 const typename RSCode<GF>::State & oldstate,
 const GF & x, int reverse) const
{
  int i;
  assert ((oldstate.ipos >= 0) && (oldstate.ipos < k));
  if (reverse) {
    newstate.ipos = oldstate.ipos - 1;
    if ((k-oldstate.ipos) % (xd-1) == 0) {
      for (i = deggenpoly-1; i > 1; i--)
	newstate.buf[i] = oldstate.buf[i-2] + x * genpoly[i-1];
      newstate.buf[1] = x * genpoly[0];
      newstate.buf[0] = GF(0);
    }
    else {
      for (i = deggenpoly-1; i > 0; i--)
	newstate.buf[i] = oldstate.buf[i-1] + x * genpoly[i];
      newstate.buf[0] = x * genpoly[0];
    }
  }
  else { // !reverse
    newstate.ipos = oldstate.ipos + 1;
    if ((k-newstate.ipos) % (xd-1) == 0) {
      for (i = 2; i < deggenpoly; i++)
	newstate.buf[i-2] = oldstate.buf[i] + x * genpoly[i];
      newstate.buf[deggenpoly-2] = x * genpoly[deggenpoly];
      newstate.buf[deggenpoly-1] = GF(0);
    }
    else {
      for (i = 1; i < deggenpoly; i++)
	newstate.buf[i-1] = oldstate.buf[i] + x * genpoly[i];
      newstate.buf[deggenpoly-1] = x * genpoly[deggenpoly];
    }
  }
}

template<class GF> inline void XRSCode<GF>::EncodeConvolutional
(GF * dest, const GF * src) const
{
  Encode (dest, src);
}

#endif
