// Create logarithm/exponential/multiplication/division tables
// for an arbitrary Galois Field GF(2^k)

// (C) 1997-1998 Frank Paehlke <frank@paehlke.de>
// Created: 02/11/1997
// $Id: create-gftables.cc,v 1.1 2003/10/14 08:56:29 frank Exp $

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// $Log: create-gftables.cc,v $
// Revision 1.1  2003/10/14 08:56:29  frank
// moved from . to libgf/
//
// Revision 1.5  2003/10/14 08:41:58  frank
// ANSI C++ / gcc 3.x compatibility changes
//
// Revision 1.4  2002/08/02 14:45:06  paehlke
// added GPL comment
//
// Revision 1.3  1998/05/28 12:56:09  paehlke
// Final version for masters thesis
//
// Revision 1.2  1997/10/19 16:47:58  frank
// representation of GF elements changed from unsigned char to int
//
// Revision 1.1  1997/06/20 12:44:36  frank
// Initial revision
//

#include <fstream>
#include <iostream>
#include <stdio.h>
#include <ctype.h>

using namespace std;

int ldq; // dimension of Galois field
int pp;  // primitive polynomial
int q;   // # of elements of Galois field
int qm1; // q - 1

int muldivtables; // !=0 if multtable/divtable are to be created

typedef int * pInt;

pInt logtable;
pInt exptable;
pInt invtable;
pInt * multtable;
pInt * divtable;

int gf_mul (int a, int b)
{
  int i;
  int a2i = a; // a2i = a * 2^i
  int res = 0; // result
  for (i=0; i<ldq; i++) {
    if (b & 1)
      res ^= a2i;
    b >>= 1;
    if (b == 0)
      break;
    a2i <<= 1;
    if (a2i & q)
      a2i = (a2i & qm1) ^ pp;
  }
  return res;
}

void print_preamble (ostream & out)
{
  int i;
  out << "// THIS FILE HAS BEEN GENERATED AUTOMATICALLY - DO NOT EDIT !\n\n"
    "// Lookup tables for arithmetic in the Galois field GF(2^"
      << ldq << ") = GF(" << q << ").\n"
    "// Every element of GF(" << q
      << ") is coded as an integer in the range\n"
    "// [0.." << qm1
      << "] where the i-th bit corresponds to the coefficient of X^i\n"
    "// in the polynomial representation modulo the primitive polynomial\n"
    "// X^" << ldq;
  for (i = ldq-1; i >= 0; i--)
    if (pp & (1<<i))
      out << "+X^" << i;
  out << " over GF(2).\n\n";
}

void print_declarations (ostream & out)
{
  out << "class GF" << q << "Tables\n"
    "{\n"
    "public:\n\n"
    "  static int q ();\n"
    "  static int ldq ();\n\n"
    "  static int gen ();\n\n"
    "  static int logtable [" << q << "];\n"
    "  // logtable[0] is undefined and should not be used\n\n"
    "  static int exptable [" << q << "];\n"
    "  static int invtable [" << q << "];\n\n"
    "  static int multtable (int, int);\n"
    "  static int divtable (int, int);\n";
  if (muldivtables)
    out << "\nprivate:\n\n"
      "  // Each of the following two tables uses " << q*q
	<< " bytes of memory.\n\n"
      "  static unsigned char _multtable [" << q << "][" << q << "];\n"
      "  static unsigned char _divtable [" << q << "][" << q << "];\n";
  out << "};\n\n"
    "inline int GF" << q << "Tables::q ()\n"
    "{\n  return " << q << ";\n}\n\n"
    "inline int GF" << q << "Tables::ldq ()\n"
    "{\n  return " << ldq << ";\n}\n\n"
    "inline int GF" << q << "Tables::gen ()\n"
    "{\n  return 2;\n}\n\n";
  if (muldivtables)
    out << "inline int GF" << q << "Tables::multtable\n"
      "(int a, int b)\n"
      "{\n"
      "  return (int) _multtable[a][b];\n"
      "}\n\n"
      "inline int GF" << q << "Tables::divtable\n"
      "(int a, int b)\n"
      "{\n"
      "  return (int) _divtable[a][b];\n"
      "}\n";
  else
    out << "inline int GF" << q << "Tables::multtable\n"
      "(int a, int b)\n"
      "{\n"
      "  if (a == 0 || b == 0)\n"
      "    return 0;\n"
      "  else {\n"
      "    int l = logtable[a] + logtable[b];\n"
      "    if (l >= " << qm1 << ")\n"
      "      l -= " << qm1 << ";\n"
      "    return exptable[l];\n"
      "  }\n"
      "}\n\n"
      "inline int GF" << q << "Tables::divtable\n"
      "(int a, int b)\n"
      "{\n"
      "  if (a == 0)\n"
      "    return 0;\n"
      "  else {\n"
      "    int l = logtable[a] - logtable[b];\n"
      "    if (l < 0)\n"
      "      l += " << qm1 << ";\n"
      "    return exptable[l];\n"
      "  }\n"
      "}\n";
}

void create_explogtables ()
{
  exptable = new int [q];
  logtable = new int [q];
  int i;
  int zi = 1;
  for (i=0; i<qm1; i++) {
    exptable[i] = zi;
    logtable[zi] = i;
    zi = gf_mul (zi, 2);
  }
  exptable[i] = zi;
  logtable[0] = -1;
}

void print_logtable (ostream & out)
{
  int i;
  out << "#include \"gf" << q << "tables.h\"\n\n"
    "// logtable[i] contains l such that gen^l = GF" << q << "(i)\n\n"
    "int GF" << q << "Tables::logtable [" << q << "] =\n{\n  ";
  for (i=0; i<qm1; i++) {
    out << logtable[i];
    if ((i&15) == 15)
      out << ",\n  ";
    else
      out << ", ";
  }
  out << logtable[i] << "\n};\n";
}

void print_exptable (ostream & out)
{
  int i;
  out << "// exptable[i] contains the integer representation of gen^i\n\n"
    "int GF" << q << "Tables::exptable [" << q << "] =\n{\n  ";
  for (i=0; i<qm1; i++) {
    out << exptable[i];
    if ((i&15) == 15)
      out << ",\n  ";
    else
      out << ", ";
  }
  out << exptable[i] << "\n};\n";
}

void create_invtable ()
{
  int i;
  invtable = new int [q];
  invtable[0] = -1;
  for (i=1; i<q; i++)
    invtable[i] = exptable[qm1-logtable[i]];
}

void print_invtable (ostream & out)
{
  int i;
  out << "// invtable[i] contains the integer representation of GF"
      << q << "(i)^(-1)\n\n"
    "int GF" << q << "Tables::invtable [" << q << "] =\n{\n  ";
  for (i=0; i<qm1; i++) {
    out << invtable[i];
    if ((i&15) == 15)
      out << ",\n  ";
    else
      out << ", ";
  }
  out << invtable[i] << "\n};\n";
}

void create_multtable ()
{
  int i, j;
  multtable = new pInt [q];
  for (i=0; i<q; i++)
    multtable[i] = new int [q];
  for (j=0; j<q; j++)
    multtable[0][j] = 0;
  for (i=1; i<q; i++) {
    multtable[i][0] = 0;
    for (j=1; j<q; j++) {
      int l = logtable[i] + logtable[j];
      if (l >= qm1)
	l -= qm1;
      multtable[i][j] = exptable[l];
    }
  }
}

void create_divtable ()
{
  int i, j;
  divtable = new pInt [q];
  for (i=0; i<q; i++)
    divtable[i] = new int [q];
  divtable[0][0] = 255;
  for (j=1; j<q; j++)
    divtable[0][j] = 0;
  for (i=1; i<q; i++) {
    divtable[i][0] = 255;
    for (j=1; j<q; j++) {
      int l = logtable[i] - logtable[j];
      if (l < 0)
	l += qm1;
      divtable[i][j] = exptable[l];
    }
  }
}

void print_multtable (ostream & out)
{
  int i, j;
  out << "// _multtable[i][j] contains the integer representation of"
    " GF" << q << "(i) * GF" << q << "(j)\n\n"
    "unsigned char GF" << q << "Tables::_multtable ["
      << q << "][" << q << "] =\n{\n  ";
  for (i=0; i<q; i++) {
    out << "{\n    ";
    for (j=0; j<qm1; j++) {
      out << multtable[i][j];
      if ((j&15) == 15)
	out << ",\n    ";
      else
	out << ", ";
    }
    if (i == qm1)
      out << multtable[i][j] << "\n  }\n};\n";
    else
      out << multtable[i][j] << "\n  }, ";
  }
}

void print_divtable (ostream & out)
{
  int i, j;
  out << "// _divtable[i][j] contains the integer representation of"
    " GF" << q << "(i) / GF" << q << "(j)\n\n"
    "unsigned char GF" << q << "Tables::_divtable ["
      << q << "][" << q << "] =\n{\n  ";
  for (i=0; i<q; i++) {
    out << "{\n    ";
    for (j=0; j<qm1; j++) {
      out << divtable[i][j];
      if ((j&15) == 15)
	out << ",\n    ";
      else
	out << ", ";
    }
    if (i == qm1)
      out << divtable[i][j] << "\n  }\n};\n";
    else
      out << divtable[i][j] << "\n  }, ";
  }
}

int main (int argc, char ** argv)
{
  int i,nmon;
  if (argc > 1)
    cerr << argv[0] << ": ignoring argument(s)\n";
  cout << "Enter degree of Galois field: ";
  cin >> ldq;
  q = 1 << ldq;
  qm1 = q - 1;
  cout << "Enter # of monomials for primitive polynomial: ";
  cin >> nmon;
  pp = 0;
  cout << "1-st monomial: X^" << ldq << endl;
  for (i=2; i<=nmon; i++) {
    int mon;
    do {
      cout << i << "-th monomial: X^";
      cin >> mon;
    } while (mon < 0 || mon >= ldq);
    pp |= 1<<mon;
  }
  if (ldq <= 8) {
    cout << "Create multiplication/division tables (y/n)? ";
    char resp;
    do {
      cin >> resp;
      resp = tolower (resp);
    } while (resp != 'y' && resp != 'n');
    muldivtables = (resp == 'y');
  }
  else // ldq > 8
    muldivtables = 0;
  char fname [21];
  sprintf (fname, "gf%dtables.h", q);
  cout << "Writing to file " << fname << endl;
  ofstream hfile (fname);
  if (hfile.fail ()) {
    cerr << argv[0] << ": error opening output file " << fname << endl;
    return 1;
  }
  print_preamble (hfile);
  print_declarations (hfile);
  hfile.close ();
  sprintf (fname, "gf%dtables.cc", q);
  cout << "Writing to file " << fname << endl;
  ofstream ccfile (fname);
  if (ccfile.fail ()) {
    cerr << argv[0] << ": error opening output file " << fname << endl;
    return 1;
  }
  print_preamble (ccfile);
  create_explogtables ();
  print_logtable (ccfile);
  ccfile << endl;
  print_exptable (ccfile);
  ccfile << endl;
  create_invtable ();
  print_invtable (ccfile);
  if (muldivtables) {
    ccfile << endl;
    create_multtable ();
    create_divtable ();
    print_multtable(ccfile);
    ccfile << endl;
    print_divtable(ccfile);
  }
  ccfile.close();
}
