#ifndef _GF_H
#define _GF_H

// A class for arithmetic in any Galois field GF(2^n) = GF(q) using
// lookup tables for exponentiation, logarithm, multiplicative inverse,
// and optionally for multiplication and division.
// The lookup tables must be defined in class TABLES.
// Every element of GF(q) is coded as an integer in the
// range [0..q-1] where the i-th bit corresponds to the coefficient
// of X^i in the equivalent polynomial representation over GF(2).

// (C) 1997-1998 Frank Paehlke <frank@paehlke.de>
// Created: 01/26/1997
// $Id: gf.h,v 1.1 2003/10/14 08:56:29 frank Exp $

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// $Log: gf.h,v $
// Revision 1.1  2003/10/14 08:56:29  frank
// moved from . to libgf/
//
// Revision 1.4  2002/08/02 14:45:06  paehlke
// added GPL comment
//
// Revision 1.3  1998/05/28 12:56:44  paehlke
// Final version for masters thesis
//
// Revision 1.2  1997/11/06 09:59:56  paehlke
// added EvalPolynomial(), EvalPolynomialD() and minpoly()
//
// Revision 1.1  1997/06/20 12:50:53 frank
// Initial revision
//

#ifndef _ASSERT_H
#include <assert.h>
#endif

// --- BEGIN OF PUBLIC INTERFACE

template<class TABLES> class GF {

public:

  // constructors

  GF (int i = 0) : rep(i) {}
  GF (const GF<TABLES> & g) : rep(g.rep) {}

  // number of elements

  static int q ();
  static int ldq ();

  // type casting operators

  operator int () const;

  // comparison and assignment operators

  int isNaN () const; // true if integer representation is < 0 or >= q
  int operator == (const GF<TABLES> &) const;
  int operator != (const GF<TABLES> &) const;

  GF<TABLES> & operator = (const GF<TABLES> &);

  // unary plus and minus

  GF<TABLES> operator + () const;
  GF<TABLES> operator - () const;

  // basic binary operators

  GF<TABLES> operator + (const GF<TABLES> &) const;
  GF<TABLES> operator - (const GF<TABLES> &) const;
  GF<TABLES> operator * (const GF<TABLES> &) const;
  GF<TABLES> operator / (const GF<TABLES> &) const;
  GF<TABLES> inv () const; // multiplicative inverse
  GF<TABLES> operator ^ (int) const; // power, not XOR

  GF<TABLES> & operator += (const GF<TABLES> &);
  GF<TABLES> & operator -= (const GF<TABLES> &);
  GF<TABLES> & operator *= (const GF<TABLES> &);
  GF<TABLES> & operator /= (const GF<TABLES> &);
  GF<TABLES> & operator ^= (int); // power, not XOR

  // logarithm and exponentiation with base gen().
  // gen() is a generator of the multiplicative group of GF(q).
  // invexp(i) calculates gen()^(-i).

  static GF<TABLES> gen ();
  static int log (const GF<TABLES> &);
  int log () const; // for convenience
  static GF<TABLES> exp (int);
  static GF<TABLES> invexp (int);


  // fast[inv]exp is faster than [inv]exp, but requires 0 <= i < q

  static GF<TABLES> fastexp (int);
  static GF<TABLES> fastinvexp (int);

  // EvalPolynomial(poly,deg,g) evaluates the polynomial
  // p(X) = poly[0] * X^0 + poly[1] * X^1 + ... + poly[deg] * X^deg
  // at the point X=g using the Horner scheme.
  // poly must point to a GF array with (deg+1) elements.
  // EvalPolynomialD(poly,deg,g) is similar to EvalPolynomial(poly,deg,x),
  // but evaluates the derivative p'(X) = diff (p(X), X) instead.

  static GF<TABLES> EvalPolynomial (const GF<TABLES> *,
				    int, const GF<TABLES> &);
  static GF<TABLES> EvalPolynomialD (const GF<TABLES> *,
				     int, const GF<TABLES> &);

  // g.minpoly() computes the minimal polynomial of g over GF(2), that is,
  // the polynomial p(X) of minimal degree with coefficients from GF(2)
  // for which p(g) == 0.
  // The result is returned as a bit sequence where 2^i represents the
  // coefficient of X^i.

  unsigned int minpoly () const;

  // --- END OF PUBLIC INTERFACE

protected:

  int rep; // integer representation
};

template<class TABLES> inline int GF<TABLES>::q ()
{
  return TABLES::q ();
}

template<class TABLES> inline int GF<TABLES>::ldq ()
{
  return TABLES::ldq ();
}

template<class TABLES> inline GF<TABLES>::operator int () const
{
  return rep;
}

template<class TABLES>
inline int GF<TABLES>::isNaN () const
{
  return rep < 0 || rep >= TABLES::q();
}

template<class TABLES>
inline int GF<TABLES>::operator == (const GF<TABLES> & g) const
{
  return rep == g.rep;
}

template<class TABLES>
inline int GF<TABLES>::operator != (const GF<TABLES> & g) const
{
  return rep != g.rep;
}

template<class TABLES>
inline GF<TABLES> & GF<TABLES>::operator = (const GF<TABLES> & g)
{
  rep = g.rep;
  return *this;
}

template<class TABLES> inline GF<TABLES> GF<TABLES>::operator + () const
{
  return rep;
}

template<class TABLES> inline GF<TABLES> GF<TABLES>::operator - () const
{
  return rep; // we're in GF(2^n), so -a==a for all a
}

template<class TABLES>
inline GF<TABLES> GF<TABLES>::operator + (const GF<TABLES> & g) const
{
  return GF<TABLES> (rep ^ g.rep);
}

template<class TABLES>
inline GF<TABLES> GF<TABLES>::operator - (const GF<TABLES> & g) const
{
  // we're in GF(2^n), so a+b==a-b for all a,b
  return GF<TABLES> (rep ^ g.rep);
}

template<class TABLES>
inline GF<TABLES> GF<TABLES>::operator * (const GF<TABLES> & g) const
{
  assert (!isNaN() && !g.isNaN());
  return GF<TABLES> (TABLES::multtable (rep, g.rep));
}

template<class TABLES>
inline GF<TABLES> GF<TABLES>::operator / (const GF<TABLES> & g) const
{
  assert (!isNaN() && !g.isNaN() && g.rep != 0);
  return GF<TABLES> (TABLES::divtable (rep, g.rep));
}

template<class TABLES> inline GF<TABLES> GF<TABLES>::inv () const
{
  assert (!isNaN() && rep != 0);
  return GF<TABLES> (TABLES::invtable [rep]);
}

template<class TABLES> GF<TABLES> GF<TABLES>::operator ^ (int e) const
{
  assert (!isNaN());
  int _rep = rep;
  if (_rep == 0) {
    assert (e > 0);
    // 0^e == 0 for all e > 0
    return GF<TABLES> (0);
  }
  else {
    if (e < 0) {
      // g^e == (g^(-1))^(-e)
      _rep = TABLES::invtable[_rep];
      e = -e;
    }
    // logarithmic calculation
    int h = TABLES::logtable[_rep] * e;
    if (h < TABLES::q())
      return TABLES::exptable[h];
    else if (h < (TABLES::q()-1) * TABLES::q()) {
      // fast reduction modulo q-1 for 0 <= h < (q-1)*q
      h = (h & (TABLES::q()-1)) + (h >> TABLES::ldq());
      // now 0 <= h < 2*(q()-1)
      if (h >= TABLES::q()-1)
	h -= TABLES::q()-1;
      return TABLES::exptable[h];
    }
    else // h >= (q-1)*q
      return TABLES::exptable[h % (TABLES::q()-1)];
  }
}

template<class TABLES>
inline GF<TABLES> & GF<TABLES>::operator += (const GF<TABLES> & g)
{
  rep ^= g.rep;
  return *this;
}

template<class TABLES>
inline GF<TABLES> & GF<TABLES>::operator -= (const GF<TABLES> & g)
{
  rep ^= g.rep;
  return *this;
}

template<class TABLES>
inline GF<TABLES> & GF<TABLES>::operator *= (const GF<TABLES> & g)
{
  assert (!isNaN() && !g.isNaN());
  rep = TABLES::multtable (rep, g.rep);
  return *this;
}

template<class TABLES>
inline GF<TABLES> & GF<TABLES>::operator /= (const GF<TABLES> & g)
{
  assert (!isNaN() && !g.isNaN() && g.rep != 0);
  rep = TABLES::divtable (rep, g.rep);
  return *this;
}

template<class TABLES> GF<TABLES> & GF<TABLES>::operator ^= (int e)
{
  assert (!isNaN());
  if (rep == 0) {
    assert (e > 0);
    // 0^e == 0 for all e > 0
    // rep is already 0, so we needn't change anything
  }
  else { // rep != 0
    if (e < 0) {
      // g^e == (g^(-1))^(-e)
      rep = TABLES::invtable[rep];
      e = -e;
    }
    // logarithmic calculation
    int h = TABLES::logtable[rep] * e;
    if (h < q())
      rep = TABLES::exptable[h];
    else if (h < (q()-1)*q()) {
      // fast reduction modulo q-1 for 0 <= h < (q-1)*q
      h = (h & (q()-1)) + (h >> TABLES::ldq());
      // now 0 <= h < 2*(q()-1)
      if (h >= q()-1)
	h -= q()-1;
      rep = TABLES::exptable[h];
    }
    else // h >= (q-1)*q
      rep = TABLES::exptable[h % (q()-1)];
  }
  return *this;
}

template<class TABLES> inline GF<TABLES> GF<TABLES>::gen ()
{
  return GF(TABLES::gen ());
}

template<class TABLES> inline int GF<TABLES>::log () const
{
  assert (!isNaN() && rep != 0);
  return TABLES::logtable[rep];
}

template<class TABLES> GF<TABLES> GF<TABLES>::exp (int i)
{
  if (i < 0) {
    if (i > -q())
      return TABLES::exptable[i+q()-1];
    else if (i >= -(q()-1)*q()) {
      // fast reduction modulo q-1 for -(q-1)*q <= i < 0
      i = (i & (q()-1)) + (i >> TABLES::ldq());
      // now -(q-1) <= i < q-1
      if (i < 0)
	i += q()-1;
      return TABLES::exptable[i];
    }
    else // i < -(q-1)*q
      return TABLES::exptable[(q()-1)-(-i)%(q()-1)];
  }
  else { // i >= 0
    if (i < q())
      return TABLES::exptable[i];
    else if (i < (q()-1)*q()) {
      // fast reduction modulo q-1 for 0 <= i < (q-1)*q
      i = (i & (q()-1)) + (i >> TABLES::ldq());
      // now 0 <= i < 2*(q-1)
      if (i >= q()-1)
	i -= q()-1;
      return TABLES::exptable[i];
    }
    else // i >= (q-1)*q
      return TABLES::exptable[i % (q()-1)];
  }
}

template<class TABLES> inline GF<TABLES> GF<TABLES>::invexp (int i)
{
  return exp (-i);
}

template<class TABLES> inline GF<TABLES> GF<TABLES>::fastexp (int i)
{
  assert (i >= 0 && i < q());
  return TABLES::exptable[i];
}

template<class TABLES> inline GF<TABLES> GF<TABLES>::fastinvexp (int i)
{
  assert (i >= 0 && i < q());
  return TABLES::exptable[TABLES::q()-1-i];
}

template<class TABLES> inline GF<TABLES> GF<TABLES>::EvalPolynomial
(const GF<TABLES> * poly, int deg, const GF<TABLES> & x)
{
  assert (!x.isNaN() && deg >= 0);
  int i;
  GF<TABLES> res = poly[deg];
  for (i = deg-1; i >= 0; i--)
    res = res * x + poly[i];
  return res;
}

template<class TABLES> inline GF<TABLES> GF<TABLES>::EvalPolynomialD
(const GF<TABLES> * poly, int deg, const GF<TABLES> & x)
{
  assert (!x.isNaN() && deg > 0);
  int i;
  // This algorithm exploits the fact that poly is a polynomial
  // over a field F with char(F)==2 and thus
  // poly'[i-1] == (i mod 2) * poly[i] == 0 * poly[i] == 0
  // whenever i is even, so we only need to use poly[i] where i is odd.
  // First we set i to the biggest odd number <= deg:
  i = (deg&1)? deg: deg-1;
  GF<TABLES> res = poly[i], x2 = x*x;
  // modified horner scheme:
  for (i-=2; i>=0; i-=2)
    res = res * x2 + poly[i];
  return res;
}

template<class TABLES> unsigned int
GF<TABLES>::minpoly () const
{
  assert (!isNaN());
  int i,j;
  GF<TABLES> p [8];
  if (rep == 0)
    return 1;
  // compute p[i] = (*this)^(2^i)
  p[0] = *this;
  int ldt2i = TABLES::logtable[rep];
  for (i=1; i<ldq(); i++) {
    // set ldt2i = (2*ldt2i) % (q()-1)
    ldt2i += ldt2i;
    if (ldt2i >= q()-1)
      ldt2i -= q()-1;
    p[i].rep = TABLES::exptable[ldt2i]; // == (*this)^(2^i)
    // search first j with p[j] == p[i]
    for (j=0;; j++)
      if (p[j].rep == p[i].rep)
	break;
    if (j < i) // p[i] equals a p[j] with j<i
      break;
  }
  int deg = i;
  // compute (X-p[0])*(X-p[1])*...*(X-p[deg-1]);
  p[deg-1] = -p[deg-1];
  for (i = deg-2; i >= 0; i--) {
    // set p[i..deg] = p[i+1..deg] * (X-p[i])
    GF pi = p[i];
    p[i] *= -p[i+1];
    for (j = i+1; j < deg-1; j++)
      p[j] -= pi * p[j+1];
    p[deg-1] -= pi;
  }
  // convert the result to a bit sequence
  unsigned int res=0, bi=1;
  for (i=0; i<deg; i++) {
    // bi == 2^i
    assert (p[i].rep == 0 || p[i].rep == 1);
    if (p[i].rep != 0)
      res |= bi;
    bi += bi;
  }
  return res | bi;
}

#endif
