// Funktionen zur numerischen Berechnung von Kanalkapazitaet und
// Fehlerexponent verschiedener mathematischer Kanalmodelle

// Autor: Frank Paehlke <Frank.Paehlke@stud.uni-karlsruhe.de>
// $Id: kanalkap.cc,v 1.1 2003/10/14 08:59:11 frank Exp $

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// $Log: kanalkap.cc,v $
// Revision 1.1  2003/10/14 08:59:11  frank
// moved from . to src/
//
// Revision 1.2  2002/08/02 14:45:06  paehlke
// added GPL comment
//
// Revision 1.1  1998/05/28 12:56:31  paehlke
// Final version for masters thesis
//

#include "kanalkap.h"

#ifndef infnan
static double infnan (int error)
{
  if (error == ERANGE) {
    errno = ERANGE;
    return HUGE_VAL;
  }
  else if (error == -ERANGE) {
    errno = ERANGE;
    return -HUGE_VAL;
  }
  else {
    errno = error;
    return HUGE_VAL;
  }
}
#endif

// solve(f, y, x0, x1, epsilon) liefert eine Loesung von f(x)=y
// im Intervall x0 <= x <= x1 mit Genauigkeit +/- epsilon/2.
// Bedingung: f(x0)-y und f(x1)-y muessen unterschiedliche Vorzeichen haben.

static double _solve (double (* f) (double), double y, double x0, double fx0,
		      double x1, double fx1, double epsilon)
{
  double x = (x0 + x1) / 2;
  if (x1 - x0 < epsilon)
    return x;
  else {
    double fx = (*f) (x);
    if (fx == y)
      return x;
    else if (fx < y) {
      if (fx0 < y)
	return _solve (f, y, x, fx, x1, fx1, epsilon);
      else
	return _solve (f, y, x0, fx0, x, fx, epsilon);
    }
    else if (fx > y) {
      if (fx0 < y)
	return _solve (f, y, x0, fx0, x, fx, epsilon);
      else
	return _solve (f, y, x, fx, x1, fx1, epsilon);
    }
  }
}

double solve (double (* f) (double), double y, double x0, double x1,
	      double epsilon)
{
  double fx0 = (*f) (x0), fx1 = (*f) (x1);
  if (x0 > x1) {
    double h = x0; x0 = x1; x1 = h;
  }
  if ((fx0 < y && fx1 < y) || (fx0 > y && fx1 > y))
    return infnan (EDOM);
  else if (fx0 == y)
    return x0;
  else if (fx1 == y)
    return x1;
  else
    return _solve (f, y, x0, fx0, x1, fx1, epsilon);
}

// solve2(f, para1, y, x0, x1, epsilon) liefert eine Loesung von
// f(para1,x)=y im Intervall x0 <= x <= x1 mit Genauigkeit +/- epsilon/2.
// Bedingung: f(para1,x0)-y und f(para1,x1)-y muessen unterschiedliche
// Vorzeichen haben.

static double _solve2 (double (* f) (int, double), int para1,
		       double y, double x0, double fx0,
		       double x1, double fx1, double epsilon)
{
  double x = (x0 + x1) / 2;
  if (x1 - x0 < epsilon)
    return x;
  else {
    double fx = (*f) (para1, x);
    if (fx == y)
      return x;
    else if (fx < y) {
      if (fx0 < y)
	return _solve2 (f, para1, y, x, fx, x1, fx1, epsilon);
      else
	return _solve2 (f, para1, y, x0, fx0, x, fx, epsilon);
    }
    else if (fx > y) {
      if (fx0 < y)
	return _solve2 (f, para1, y, x0, fx0, x, fx, epsilon);
      else
	return _solve2 (f, para1, y, x, fx, x1, fx1, epsilon);
    }
  }
}

double solve2 (double (* f) (int, double), int para1,
	      double y, double x0, double x1, double epsilon)
{
  double fx0 = (*f) (para1, x0), fx1 = (*f) (para1, x1);
  if (x0 > x1) {
    double h = x0; x0 = x1; x1 = h;
  }
  if ((fx0 < y && fx1 < y) || (fx0 > y && fx1 > y))
    return infnan (EDOM);
  else if (fx0 == y)
    return x0;
  else if (fx1 == y)
    return x1;
  else
    return _solve2 (f, para1, y, x0, fx0, x1, fx1, epsilon);
}

// Simpson: 1/3 [1 4 1]
// besser:  1/45 [7 32 12 32 7]

// integral(f, x0, x1, epsilon) berechnet das Integral int(f(x), x=x0..x1).
// Dabei wird versucht, eine Genauigkeit von +/- epsilon einzuhalten.

static double _integral (double (* f) (double), double x0, double fx0,
			 double x1, double fx1,
			 double x2, double fx2, double epsilon)
{
  double m = (x2 - x0) / 12;
  double x01 = (x0 + x1) / 2;
  double x12 = (x1 + x2) / 2;
  double fx01 = (*f)(x01), fx12 = (*f)(x12);
  double s1 = m * 2 * (fx0 + fx2 + 4 * fx1);
  double s2 = m * (fx0 + fx2 + 2 * fx1 + 4 * (fx01 + fx12));
  if (fabs (s2 - s1) < epsilon)
    return s2;
  else
    return (_integral (f, x0, fx0, x01, fx01, x1, fx1, 0.5*epsilon)
	    + _integral (f, x1, fx1, x12, fx12, x2, fx2, 0.5*epsilon));
}

double integral (double (* f) (double), double _x0, double _x1,
		 int n0, double epsilon)
{
  int i, n02 = 2*n0;
  double res = 0;
  double step = (_x1 - _x0) / n02;
  double x0, x1, fx0, fx1;
  x0 = _x0; x1 = x0 + step;
  fx0 = (*f)(x0); fx1 = (*f)(x1);
  for (i = 2; i <= n02; i += 2) {
    double x2 = _x0 + i * step;
    double fx2 = (*f)(x2);
    res += _integral (f, x0, fx0, x1, fx1, x2, fx2, epsilon/n0);
    x0 = x2; x1 = x0 + step;
    fx0 = fx2; fx1 = (*f)(x1);
  }
  return res;
}

// grundlegende informationstheoretische Funktionen

// komplementaere Verteilungsfunktion der Standard-Normalverteilung

double Q (double x)
{
  return 0.5 * erfc (M_SQRT1_2 * x);
}

// binaere und q-naere Entropiefunktion

double H2 (double x)
{
  if (x < 0.0 || x > 1.0)
    return infnan (EDOM);
  else if (x == 0.0 || x == 1.0)
    return 0.0;
  else
    return -M_1_LN2 * (x * log(x) + (1-x) * log1p (-x));
}

double Hq (int q, double x)
{
  if (q < 2 || x < 0.0 || x > 1.0)
    return infnan (EDOM);
  else if (x == 0.0)
    return 0.0;
  else if (x == 1.0)
    return M_1_LN2 * log (q-1.0);
  else
    return M_1_LN2 * (x * (log(q-1.0) - log(x))
		      - (1-x) * log1p (-x));
}

// gemischte Entropiefunktion

double Tp (double p, double x)
{
  if (p <= 0.0 || p >= 1.0 || x < 0.0 || x > 1.0)
    return infnan(EDOM);
  return -M_1_LN2 * (x * log(p) + (1-x) * log1p (-p));
}

// binrer und q-nrer Bhattacharyya-Multiplikator

double U2 (double p)
{
  if (p < 0.0 || p > 1.0)
    return infnan(EDOM);
  else if (p == 0.0 || p == 1.0)
    return infnan(ERANGE);
  else
    return -M_1_LN2 * log(2 * sqrt (p * (1-p)));
}

double Uq (int q, double p)
{
  if (q < 2 || p < 0.0 || p > 1.0)
    return infnan(EDOM);
  else if (p == 0.0 || (q==2 && p == 1.0))
    return infnan(ERANGE);
  else if (p == 0.5)
    return 0.0;
  else if (p == 1.0)
    return -M_1_LN2 * log((q - 2.0)/(q - 1.0));
  else {
    double qm1 = q - 1.0;
    double h = (q-2) / qm1 * p + 2 * sqrt (p * (1-p) / qm1);
    return -M_1_LN2 * log (h);
  }
}

// expU2(p)    == 2^(-U2(p))
// qm1expUq(p) == (q-1) * 2^(-Uq(p))

double expU2 (double p)
{
  if (p < 0.0 || p > 1.0)
    return infnan(EDOM);
  else
    return 2 * sqrt (p * (1-p));
}

double Qm1expUq (int q, double p)
{
  if (q < 2 || p < 0.0 || p > 1.0)
    return infnan(EDOM);
  else
    return (q-2) * p + 2 * sqrt (p * (1-p) * (q-1));
}

// Gilbert-Varshamov-Schranke

double gv2_r_delta (double delta)
{
  if (delta < 0.0 || delta > 0.5)
    return infnan(EDOM);
  else if (delta == 0.0)
    return 1.0;
  else if (delta == 0.5)
    return 0.0;
  else
    return 1.0 - H2(delta);
}

double gv2_delta_r (double r)
{
  if (r < 0.0 || r > 1.0)
    return infnan(EDOM);
  else if (r == 0.0)
    return 0.5;
  else if (r == 1.0)
    return 0.0;
  else
    return solve(gv2_r_delta, r, 0.0, 0.5);
}

double gvq_r_delta (int q, double delta)
{
  if (q < 2 || delta < 0.0 || q * delta > q-1)
    return infnan(EDOM);
  else if (delta == 0.0)
    return M_1_LN2 * log(q);
  else
    return M_1_LN2 * (log(q) + delta * (log(delta) - log(q-1.0))
		      + (1-delta) * log1p (-delta));
}

double gvq_delta_r (int q, double r)
{
  if (q < 2)
    return infnan(EDOM);
  double ldq = M_1_LN2 * log(q);
  if (r < 0.0 || r > ldq)
    return infnan(EDOM);
  else if (r == 0.0)
    return (q-1.0) / q;
  else if (r == ldq)
    return 0.0;
  else
    return solve2 (gvq_r_delta, q, r, 0.0, (q-1.0)/q);
}

// Kanalkapazitaet eines USC

double usc_rc_p (int q, double p)
{
  if (q < 2 || p < 0.0 || p > 1.0)
    return infnan (EDOM);
  else
    return M_1_LN2 * log((double)q) - Hq (q,p);
}

double usc_p_rc (int q, double rc)
{
  if (q < 2)
    return infnan(EDOM);
  double ldq = M_1_LN2 * log((double)q);
  if (rc < 0.0 || rc > ldq)
    return infnan (EDOM);
  else if (rc == 0.0)
    return (q-1.0) / q;
  else if (rc == ldq)
    return 0.0;
  else
    return solve2 (usc_rc_p, q, rc, 0.0, (q-1.0)/q, 1e-6);
}

// Cutoff Rate eines USC

double usc_r0_p (int q, double p)
{
  if (q < 2 || p < 0.0 || p > 1.0)
    return infnan (EDOM);
  else if (p == 0.0)
    return M_1_LN2 * log((double)q);
  else
    return M_1_LN2 * (log((double)q)
		      - log1p ((q-2)*p + 2*sqrt ((q-1)*p*(1-p))));
}

double usc_p_r0 (int q, double r0)
{
  if (q < 2)
    return infnan(EDOM);
  double ldq = M_1_LN2 * log((double)q);
  if (r0 < 0.0 || r0 > ldq)
    return infnan (EDOM);
  else if (r0 == 0.0)
    return (q-1.0) / q;
  else {
    double h = expm1 (M_LN2 * (ldq - r0));
    double h2 = (h*(q-2) + 2*(q-1)) / q;
    return (h2 - sqrt (h2*h2 - h*h)) / q;
  }
}

// Grenzraten eines USC

double usc_rcrit_p (int q, double p)
{
  if (p < 0.0 || p > 1.0)
    return infnan(EDOM);
  else if (p == 0.0)
    return M_1_LN2 * log((double)q);
  else if (q * p == q - 1.0)
    return 0.0;
  else {
    double a = sqrt((q-1.0) * p), b = sqrt(1.0 - p);
    return M_1_LN2 * (log((double)q) - (a * (log(q-1.0) + log1p(b/a))
					+ b * log1p(a/b)) / (a+b));
  }
}

// Fehlerexponent eines USC (Unter-/Oberschranke)

double usc_rex_p (int q, double p)
{
  if (q < 2 || p < 0.0 || p > 1.0)
    return infnan (EDOM);
  else if (p == 0.0)
    return M_1_LN2 * log((double)q);
  else if (q * p == q - 1.0)
    return 0.0;
  else {
    double uq = Uq(q,p), quq = Qm1expUq(q,p);
    return usc_r0_p(q,p) - uq * quq / (1.0+quq);
  }
}

double usc_eelow_p_r (int q, double p, double r)
{
  if (q < 2 || p < 0.0 || p > 1.0)
    return infnan(EDOM);
  else if (p == 0.0)
    return infnan(ERANGE);
  else if (q * p == q - 1.0)
    return 0.0;
  double ldq = M_1_LN2 * log((double)q);
  if (r < 0.0 || r > ldq)
    return infnan(EDOM);
  else if (r == 0.0)
    return Uq(q,p) * (q-1) / q;
  else if (r < usc_rex_p(q,p))
    return Uq(q,p) * gvq_delta_r(q,r);
  else if (r <= usc_rcrit_p(q,p))
    return usc_r0_p(q,p) - r;
  else if (r < usc_rc_p(q,p)) {
    double d = gvq_delta_r(q,r);
    return Tp(p,d) - H2(d);
  }
  else
    return 0.0;
}

double usc_eehigh_p_r (int q, double p, double r)
{
  if (q < 2 || p < 0.0 || p > 1.0)
    return infnan(EDOM);
  else if (p == 0.0)
    return infnan(ERANGE);
  else if (q * p == q - 1.0)
    return 0.0;
  double ldq = M_1_LN2 * log((double)q);
  if (r < 0.0 || r > ldq)
    return infnan(EDOM);
  else if (r == 0.0) {
    double d = (q-1.0)/q;
    return Tp(p,d) - H2(d);
  }
  else if (r < usc_rc_p(q,p)) {
    double d = gvq_delta_r(q,r);
    return Tp(p,d) - H2(d);
  }
  else
    return 0.0;
}

// Kanalkapazitaet eines BSC

double bsc_rc_p (double p)
{
  if (p < 0.0 || p > 1.0)
    return infnan (EDOM);
  else
    return 1.0 - H2(p);
}

double bsc_p_rc (double rc)
{
  if (rc < 0.0 || rc > 1.0)
    return infnan (EDOM);
  else if (rc == 0.0)
    return 0.5;
  else if (rc == 1.0)
    return 0.0;
  else
    return solve (bsc_rc_p, rc, 0.0, 0.5, 1e-6);
}

// Cutoff Rate eines BSC

double bsc_r0_p (double p)
{
  if (p < 0.0 || p > 1.0)
    return infnan (EDOM);
  else if (p == 0.0 || p == 1.0)
    return 1.0;
  else
    return 1 - M_1_LN2 * log1p (2*sqrt (p*(1-p)));
}

double bsc_p_r0 (double r0)
{
  if (r0 < 0.0 || r0 > 1.0)
    return infnan (EDOM);
  else if (r0 == 0.0)
    return 0.5;
  else {
    double h = expm1 (M_LN2 * (1-r0));
    h = 0.25 * (1 - h*h);
    return 0.5 - sqrt (h);
  }
}

// Grenzraten eines BSC

double bsc_rcrit_p (double p)
{
  if (p < 0.0 || p > 1.0)
    return infnan(EDOM);
  else if (p == 0.0 || p == 1.0)
    return 1.0;
  else if (p == 0.5)
    return 0.0;
  else {
    double sp = sqrt(p), s1p = sqrt(1.0 - p);
    return 1.0 - M_1_LN2 * ((sp * log1p(s1p/sp) + s1p * log1p(sp/s1p))
			    / (sp+s1p));
  }
}

double bsc_rex_p (double p)
{
  if (p < 0.0 || p > 1.0)
    return infnan(EDOM);
  else if (p == 0.0 || p == 1.0)
    return 1.0;
  else if (p == 0.5)
    return 0.0;
  else {
    double u = U2(p), eu = expU2(p);
    return bsc_r0_p(p) - u * eu / (1.0+eu);
  }
}

// Fehlerexponent eines BSC (Unter-/Oberschranke)

double bsc_eelow_p_r (double p, double r)
{
  if (p < 0.0 || p > 1.0 || r < 0.0 || r > 1.0)
    return infnan(EDOM);
  else if (p == 0.0 || p == 1.0)
    return infnan(ERANGE);
  else if (p == 0.5)
    return 0.0;
  else if (r == 0.0)
    return 0.5 * U2(p);
  else if (r < bsc_rex_p(p))
    return U2(p) * gv2_delta_r(r);
  else if (r <= bsc_rcrit_p(p))
    return bsc_r0_p(p) - r;
  else if (r < bsc_rc_p(p)) {
    double d = gv2_delta_r(r);
    return Tp(p,d) - H2(d);
  }
  else
    return 0.0;
}

// Kanalkapazitaet eines AWGN-Kanals (Ein-/Ausgabe reell)

double bsc_eehigh_p_r (double p, double r)
{
  if (p < 0.0 || p > 1.0 || r < 0.0 || r > 1.0)
    return infnan(EDOM);
  else if (p == 0.0 || p == 1.0)
    return infnan(ERANGE);
  else if (p == 0.5)
    return 0.0;
  else if (r == 0.0)
    return U2(p);
  else if (r < bsc_rc_p(p)) {
    double d = gv2_delta_r(r);
    return Tp(p,d) - H2(d);
  }
  else
    return 0.0;
}

// Kanalkapazitaet eines AWGN-Kanals (Ein-/Ausgabe reell)

double awgn_rc_snr (double snr)
{
  if (snr < 0)
    return infnan(EDOM);
  else
    return 0.5 * M_1_LN2 * log1p (snr);
}

double awgn_rc_snrdb (double snrdb)
{
  return awgn_rc_snr (from_db (snrdb));
}

double awgn_snr_rc (double rc)
{
  if (rc < 0)
    return infnan(EDOM);
  else
    return expm1 (2.0 * M_LN2 * rc);
}

double awgn_snrdb_rc (double rc)
{
  return to_db (awgn_snr_rc (rc));
}

double awgn_ebn0_rc (double rc)
{
  if (rc == 0)
    return M_LN2;
  else
    return awgn_snr_rc(rc) / (2.0 * rc);
}

double awgn_ebn0db_rc (double rc)
{
  return to_db (awgn_ebn0_rc (rc));
}

double awgn_rc_ebn0 (double ebn0)
{
  if (ebn0 < 0)
    return infnan(EDOM);
  else if (ebn0 <= M_LN2)
    return 0;
  else
    return solve (awgn_ebn0_rc, ebn0, 0.0, 500.0);
}

double awgn_rc_ebn0db (double ebn0db)
{
  return awgn_rc_ebn0 (from_db (ebn0db));
}

// Cutoff Rate eines AWGN-Kanals (Ein-/Ausgabe reell)

double awgn_r0_snr (double snr)
{
  if (snr < 0.0)
    return infnan(EDOM);
  else if (snr == 0.0)
    return 0.0;
  else {
    double nu = 0.5 * snr, s1 = sqrt(1.0 + nu*nu) - 1.0;
    return 0.5 * M_1_LN2 * (nu - s1 + log1p(0.5 * s1));
  }
}

double awgn_r0_snrdb (double snrdb)
{
  return awgn_r0_snr (from_db (snrdb));
}

double awgn_r0_ebn0 (double ebn0)
{
  if (ebn0 < 0.0)
    return infnan(EDOM);
  else if (ebn0 <= 2.0 * M_LN2)
    return 0.0;
  else
    return solve (awgn_ebn0_r0, ebn0, 0.0, 100.0);
}

double awgn_r0_ebn0db (double ebn0db)
{
  return awgn_r0_ebn0 (from_db (ebn0db));
}

double awgn_snr_r0 (double r0)
{
  if (r0 < 0.0)
    return infnan(EDOM);
  else if (r0 == 0.0)
    return 0.0;
  else
    return solve(awgn_r0_snr, r0, 0, 1e100);
}

double awgn_snrdb_r0 (double r0)
{
  return to_db(awgn_snr_r0(r0));
}

double awgn_ebn0_r0 (double r0)
{
  if (r0 < 0.0)
    return infnan(EDOM);
  else if (r0 == 0)
    return 2.0 * M_LN2;
  else
    return awgn_snr_r0(r0) / (2.0 * r0);
}

double awgn_ebn0db_r0 (double r0)
{
  return to_db(awgn_ebn0_r0(r0));
}

// Grenzraten eines AWGN-Kanals (Ein-/Ausgabe reell)

extern double awgn_rex_snr (double snr)
{
  if (snr < 0.0)
    return infnan(EDOM);
  else if (snr == 0.0)
    return 0.0;
  else
    return 0.5 * M_1_LN2 * log1p(0.5 * (sqrt(1.0 + 0.25*snr*snr) - 1.0));
}

extern double awgn_rex_snrdb (double snrdb)
{
  return awgn_rex_snr(from_db(snrdb));
}

extern double awgn_rcrit_snr (double snr)
{
  if (snr < 0.0)
    return infnan(EDOM);
  else if (snr == 0.0)
    return 0.0;
  else {
    double nu = 0.5 * snr;
    return 0.5 * M_1_LN2 * log1p(0.5 * (nu + (sqrt(1.0 + nu*nu) - 1.0)));
  }
}

extern double awgn_rcrit_snrdb (double snrdb)
{
  return awgn_rcrit_snr(from_db(snrdb));
}

// Fehlerexponent eines AWGN-Kanals (Ein-/Ausgabe reell)
// (Unter-/Oberschranke)

double awgn_eelow_snr_r (double snr, double r)
{
  if ((snr < 0.0) || (r < 0.0))
    return infnan(EDOM);
  else if (snr == 0.0)
    return 0.0;
  else if (r == 0.0)
    return 0.25 * M_1_LN2 * snr;
  else if (r < awgn_rex_snr(snr))
    return 0.25 * M_1_LN2 * snr * (1.0 - sqrt(1.0 - pow(2.0, -2.0*r)));
  else if (r <= awgn_rcrit_snr(snr))
    return awgn_r0_snr(snr) - r;
  else if (r < awgn_rc_snr(snr)) {
    double nu = 0.5 * snr, p4mr = pow(4.0,-r);
    double p4mr1 = -expm1 (-2.0 * M_LN2 * r); // 1-4^(-r)
    return r + 0.5 * M_1_LN2
      * (nu * (p4mr + 1.0)
	 - sqrt((nu * p4mr1) * (nu*p4mr1 + 2.0))
	 + log1p(nu*p4mr1 - sqrt((nu*p4mr1 * (nu*p4mr1 + 2.0)))));
  }
  else
    return 0.0;
}

double awgn_eelow_snrdb_r (double snrdb, double r)
{
  return awgn_eelow_snr_r(from_db(snrdb), r);
}

double awgn_eehigh_snr_r (double snr, double r)
{
  if ((snr < 0.0) || (r < 0.0))
    return infnan(EDOM);
  else if (snr == 0.0)
    return 0.0;
  else if (r == 0.0)
    return 0.5 * M_1_LN2 * snr;
  else if (r < awgn_rc_snr(snr)) {
    double nu = 0.5 * snr, p4mr = pow(4.0,-r);
    double p4mr1 = -expm1 (-2.0 * M_LN2 * r); // 1-4^(-r)
    double h = nu * p4mr1;
    return r + 0.5 * M_1_LN2 *
      (nu * (p4mr + 1.0) - sqrt(h * (h + 2.0))
       + log1p(h - sqrt((h * (h + 2.0)))));
  }
  else
    return 0.0;
}

double awgn_eehigh_snrdb_r (double snrdb, double r)
{
  return awgn_eehigh_snr_r (from_db (snrdb), r);
}

double awgn_eehigh_ebn0_r (double ebn0, double r)
{
  return awgn_eehigh_snr_r (2.0*r*ebn0, r);
}

double awgn_eehigh_ebn0db_r (double ebn0db, double r)
{
  return awgn_eehigh_ebn0_r (from_db (ebn0db), r);
}

// Kanalkapazitaet eines AWGN-Kanals (Eingabe binaer, Ausgabe reell)
// 1mrc bedeutet dabei 1 - rc

static double nu;

static double _func1 (double x)
{
  double xmnu = x - nu;
  double mxmnu = -x - nu;
  return ((exp (-0.5*xmnu*xmnu) * (M_LN2 - log1p (exp (-2*nu*x)))
	   + exp (-0.5*mxmnu*mxmnu) * (M_LN2 - log1p (exp (2*nu*x))))
	  / (M_LN2 * sqrt(2*M_PI)));
}

static double _func2 (double x)
{
  double xmnu = x-nu;
  double mxmnu = -x-nu;
  return ((exp (-0.5*xmnu*xmnu) * log1p (exp (-2*nu*x))
	   + exp (-0.5*mxmnu*mxmnu) * log1p (exp (2*nu*x)))
	  / (M_LN2 * sqrt(2*M_PI)));
}

double awgnbin_rc_snr (double snr)
{
  if (snr < 0.0)
    return infnan(EDOM);
  if (snr == 0.0)
    return 0.0;
  else {
    nu = sqrt(snr);
    if (snr < 1.0)
      return integral (_func1, 0.0, nu+10.0, 5, 1e-6);
    else
      return 1.0 - integral (_func2, 0.0, nu+10.0, 5, 1e-6);
  }
}

double awgnbin_rc_snrdb (double snrdb)
{
  return awgnbin_rc_snr (from_db (snrdb));
}

double awgnbin_rc_ebn0 (double ebn0)
{
  if (ebn0 < 0.0)
    return infnan(EDOM);
  else if (ebn0 <= M_LN2)
    return 0.0;
  else
    return solve (awgnbin_ebn0_rc, ebn0, 0.0, 0.999999);
}

double awgnbin_rc_ebn0db (double ebn0db)
{
  return awgnbin_rc_ebn0 (from_db (ebn0db));
}

double awgnbin_1mrc_snr (double snr)
{
  if (snr < 0.0)
    return infnan(EDOM);
  else if (snr == 0.0)
    return 1.0;
  else {
    nu = sqrt (snr);
    if (snr < 1.0)
      return 1.0 - integral (_func1, 0.0, nu+10.0, 5, 1e-6);
    else
      return integral (_func2, 0.0, nu+10.0, 5, 1e-6);
  }
}

double awgnbin_1mrc_snrdb (double snrdb)
{
  return awgnbin_1mrc_snr (from_db (snrdb));
}

double awgnbin_1mrc_ebn0 (double ebn0)
{
  if (ebn0 < 0.0)
    return infnan(EDOM);
  else if (ebn0 <= M_LN2)
    return 0.0;
  else
    return solve (awgnbin_ebn0_1mrc, ebn0, 1e-6, 1.0);
}

double awgnbin_1mrc_ebn0db (double ebn0db)
{
  return awgnbin_1mrc_ebn0 (from_db (ebn0db));
}

double awgnbin_snr_rc (double rc)
{
  if (rc < 0 || rc >= 1)
    return infnan (EDOM);
  else if (rc < 0.5)
    return solve (awgnbin_rc_snr, rc, 0.0, 1.1);
  else
    return solve (awgnbin_1mrc_snr, 1-rc, 1.0, 100.0);
}

double awgnbin_snrdb_rc (double rc)
{
  return to_db (awgnbin_snr_rc (rc));
}

double awgnbin_ebn0_rc (double rc)
{
  return awgnbin_snr_rc(rc) / (2.0 * rc);
}

double awgnbin_ebn0db_rc (double rc)
{
  return to_db (awgnbin_ebn0_rc (rc));
}

double awgnbin_snr_1mrc (double r)
{
  if (r < 0.0 || r > 1.0)
    return infnan (EDOM);
  else if (r == 0.0)
    return infnan (ERANGE);
  else if (r < 0.5)
    return solve (awgnbin_1mrc_snr, r, 1.0, 100.0);
  else
    return solve (awgnbin_rc_snr, 1-r, 0.0, 1.1);
}

double awgnbin_snrdb_1mrc (double r)
{
  return to_db(awgnbin_snr_1mrc(r));
}

double awgnbin_ebn0_1mrc (double r)
{
  if (r == 0.0)
    return M_LN2;
  else
    return awgnbin_snr_1mrc(r) / (2.0 * r);
}

double awgnbin_ebn0db_1mrc (double r)
{
  return to_db(awgnbin_ebn0_1mrc(r));
}

// Cutoff Rate eines AWGN-Kanals (Eingabe binaer, Ausgabe reell)
// 1mr0 bedeutet dabei 1 - r0

double awgnbin_r0_snr (double snr)
{
  if (snr < 0.0)
    return infnan(EDOM);
  else
    return 1 - M_1_LN2 * log1p(exp(-0.5 * snr));
}

double awgnbin_r0_snrdb (double snrdb)
{
  return awgnbin_r0_snr(from_db(snrdb));
}

double awgnbin_r0_ebn0 (double ebn0)
{
  if (ebn0 < 0.0)
    return infnan(EDOM);
  else if (ebn0 <= 2.0 * M_LN2)
    return 0;
  else
    return solve (awgnbin_ebn0_r0, ebn0, 0.0, 0.999999);
}

double awgnbin_r0_ebn0db (double ebn0db)
{
  return awgnbin_r0_ebn0(from_db(ebn0db));
}

double awgnbin_1mr0_snr (double snr)
{
  if (snr < 0.0)
    return infnan(EDOM);
  else
    return M_1_LN2 * log1p(exp(-0.5 * snr));
}

double awgnbin_1mr0_snrdb (double snrdb)
{
  return awgnbin_1mr0_snr(from_db(snrdb));
}

double awgnbin_1mr0_ebn0 (double ebn0)
{
  if (ebn0 < 0.0)
    return infnan(EDOM);
  else if (ebn0 <= 2.0 * M_LN2)
    return 0.0;
  else
    return solve (awgnbin_ebn0_1mr0, ebn0, 1e-6, 1.0);
}

double awgnbin_1mr0_ebn0db (double ebn0db)
{
  return awgnbin_1mr0_ebn0(from_db(ebn0db));
}

double awgnbin_snr_r0 (double r0)
{
  if (r0 < 0.0 || r0 > 1.0)
    return infnan(EDOM);
  else if (r0 == 1.0)
    return infnan(ERANGE);
  else
    return -2 * log(expm1(M_LN2 * (1-r0)));
}

double awgnbin_snrdb_r0 (double r0)
{
  return to_db(awgnbin_snr_r0(r0));
}

double awgnbin_ebn0_r0 (double r0)
{
  if (r0 == 0.0)
    return 2.0 * M_LN2;
  else
    return awgnbin_snr_r0(r0) / (2*r0);
}

double awgnbin_ebn0db_r0 (double r0)
{
  return to_db(awgnbin_ebn0_r0(r0));
}

double awgnbin_snr_1mr0 (double r)
{
  if (r < 0.0 || r > 1.0)
    return infnan(EDOM);
  else if (r == 0.0)
    return infnan(ERANGE);
  else
    return -2 * log(expm1(M_LN2 * r));
}

double awgnbin_snrdb_1mr0 (double r)
{
  return to_db(awgnbin_snr_1mr0(r));
}

double awgnbin_ebn0_1mr0 (double r)
{
  if (r == 0.0)
    return 2.0 * M_LN2;
  else
    return awgnbin_snr_1mr0(r) / (2.0 * r);
}

double awgnbin_ebn0db_1mr0 (double r)
{
  return to_db(awgnbin_ebn0_1mr0(r));
}
