#ifndef _KANALKAP_H
#define _KANALKAP_H

// Funktionen zur numerischen Berechnung von Kanalkapazitaet und
// Fehlerexponent verschiedener mathematischer Kanalmodelle

// Autor: Frank Paehlke <Frank.Paehlke@stud.uni-karlsruhe.de>
// $Id: kanalkap.h,v 1.1 2003/10/14 08:59:11 frank Exp $

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// $Log: kanalkap.h,v $
// Revision 1.1  2003/10/14 08:59:11  frank
// moved from . to src/
//
// Revision 1.2  2002/08/02 14:45:06  paehlke
// added GPL comment
//
// Revision 1.1  1998/05/28 12:57:01  paehlke
// Final version for masters thesis
//

#ifndef _MATH_H
#include <math.h>
#endif

#ifndef _ERRNO_H
#include <errno.h>
#endif

const double M_1_LN2 = 1.0 / M_LN2;

// solve(f, y, x0, x1, epsilon) liefert eine Loesung von f(x)=y
// im Intervall x0 <= x <= x1 mit Genauigkeit +/- epsilon/2.
// Bedingung: f(x0)-y und f(x1)-y muessen unterschiedliche Vorzeichen haben.

extern double solve (double (* f) (double), double y,
		     double x0, double x1, double epsilon = 1e-5);

// solve2(f, para1, y, x0, x1, epsilon) liefert eine Loesung von
// f(para1,x)=y im Intervall x0 <= x <= x1 mit Genauigkeit +/- epsilon/2.
// Bedingung: f(para1,x0)-y und f(para1,x1)-y muessen unterschiedliche
// Vorzeichen haben.

extern double solve2 (double (* f) (int,double), int para1, double y,
		      double x0, double x1, double epsilon = 1e-5);

// integral(f, x0, x1, epsilon) berechnet das Integral int(f(x), x=x0..x1).
// Dabei wird versucht, eine Genauigkeit von +/- epsilon einzuhalten.

extern double integral (double (* f) (double), double x0, double x1,
			int n0, double epsilon);

// Umrechnung von/in Dezibel

inline double to_db (double x)
{
  return 10 * log10 (x);
}

inline double from_db (double x)
{
  return pow (10.0, x / 10.0);
}

// grundlegende informationstheoretische Funktionen

// komplementaere Verteilungsfunktion der Standard-Normalverteilung

extern double Q (double x);

// binaere und q-naere Entropiefunktion

extern double H2 (double x);
extern double Hq (int q, double x);

// gemischte Entropiefunktion

extern double Tp (double p, double x);

// binrer und q-nrer Bhattacharyya-Multiplikator

extern double U2 (double p);
extern double Uq (int q, double p);

// expU2(p)    == 2^(-U2(p))
// qm1expUq(p) == (q-1) * 2^(-Uq(p))

extern double expU2 (double p);
extern double qm1expUq (int q, double p);

// Gilbert-Varshamov-Schranke

extern double gv2_r_delta (double delta);
extern double gv2_delta_r (double r);

extern double gvq_r_delta (int q, double delta);
extern double gvq_delta_r (int q, double r);

// Kanalkapazitaet eines USC

extern double usc_rc_p (int q, double p);
extern double usc_p_rc (int q, double rc);

// Cutoff Rate eines USC

extern double usc_r0_p (int q, double p);
extern double usc_p_r0 (int q, double r0);

// Grenzraten eines USC

extern double usc_rcrit_p (int q, double p);
extern double usc_rex_p (int q, double p);

// Fehlerexponent eines USC (Unter-/Oberschranke)

extern double usc_eelow_p_r (int q, double p, double r);
extern double usc_eehigh_p_r (int q, double p, double r);

// Kanalkapazitaet eines BSC

extern double bsc_rc_p (double p);
extern double bsc_p_rc (double rc);

// Cutoff Rate eines BSC

extern double bsc_r0_p (double p);
extern double bsc_p_r0 (double r0);

// Grenzraten eines BSC

extern double bsc_rcrit_p (double p);
extern double bsc_rex_p (double p);

// Fehlerexponent eines BSC (Unter-/Oberschranke)

extern double bsc_eelow_p_r (double p, double r);
extern double bsc_eehigh_p_r (double p, double r);

// Kanalkapazitaet eines AWGN-Kanals (Ein-/Ausgabe reell)

extern double awgn_rc_snr (double snr);
extern double awgn_rc_snrdb (double snrdb);
extern double awgn_rc_ebn0 (double ebn0);
extern double awgn_rc_ebn0db (double ebn0db);
extern double awgn_snr_rc (double rc);
extern double awgn_snrdb_rc (double rc);
extern double awgn_ebn0_rc (double rc);
extern double awgn_ebn0db_rc (double rc);

// Cutoff Rate eines AWGN-Kanals (Ein-/Ausgabe reell)

extern double awgn_r0_snr (double snr);
extern double awgn_r0_snrdb (double snrdb);
extern double awgn_r0_ebn0 (double ebn0);
extern double awgn_r0_ebn0db (double ebn0db);
extern double awgn_snr_r0 (double r0);
extern double awgn_snrdb_r0 (double r0);
extern double awgn_ebn0_r0 (double r0);
extern double awgn_ebn0db_r0 (double r0);

// Grenzraten eines AWGN-Kanals (Ein-/Ausgabe reell)

extern double awgn_rex_snr (double snr);
extern double awgn_rex_snrdb (double snrdb);
extern double awgn_rcrit_snr (double snr);
extern double awgn_rcrit_snrdb (double snrdb);

// Fehlerexponent eines AWGN-Kanals (Ein-/Ausgabe reell)
// (Unter-/Oberschranke)

extern double awgn_eelow_snr_r (double snr, double r);
extern double awgn_eelow_snrdb_r (double snrdb, double r);
extern double awgn_eelow_ebn0_r (double ebn0, double r);
extern double awgn_eelow_ebn0db_r (double ebn0db, double r);
extern double awgn_eehigh_snr_r (double snr, double r);
extern double awgn_eehigh_snrdb_r (double snrdb, double r);
extern double awgn_eehigh_ebn0_r (double ebn0, double r);
extern double awgn_eehigh_ebn0db_r (double ebn0db, double r);

// Kanalkapazitaet eines AWGN-Kanals (Eingabe binaer, Ausgabe reell)
// 1mrc bedeutet dabei 1 - rc

extern double awgnbin_rc_snr (double snr);
extern double awgnbin_rc_snrdb (double snrdb);
extern double awgnbin_rc_ebn0 (double ebn0);
extern double awgnbin_rc_ebn0db (double ebn0db);
extern double awgnbin_1mrc_snr (double snr);
extern double awgnbin_1mrc_snrdb (double snrdb);
extern double awgnbin_1mrc_ebn0 (double ebn0);
extern double awgnbin_1mrc_ebn0db (double ebn0db);
extern double awgnbin_snr_rc (double rc);
extern double awgnbin_snrdb_rc (double rc);
extern double awgnbin_ebn0_rc (double rc);
extern double awgnbin_ebn0db_rc (double rc);
extern double awgnbin_snr_1mrc (double r);
extern double awgnbin_snrdb_1mrc (double r);
extern double awgnbin_ebn0_1mrc (double r);
extern double awgnbin_ebn0db_1mrc (double r);

// Cutoff Rate eines AWGN-Kanals (Eingabe binaer, Ausgabe reell)
// 1mr0 bedeutet dabei 1 - r0

extern double awgnbin_r0_snr (double snr);
extern double awgnbin_r0_snrdb (double snrdb);
extern double awgnbin_r0_ebn0 (double ebn0);
extern double awgnbin_r0_ebn0db (double ebn0db);
extern double awgnbin_1mr0_snr (double snr);
extern double awgnbin_1mr0_snrdb (double snrdb);
extern double awgnbin_1mr0_ebn0 (double ebn0);
extern double awgnbin_1mr0_ebn0db (double ebn0db);
extern double awgnbin_snr_r0 (double r0);
extern double awgnbin_snrdb_r0 (double r0);
extern double awgnbin_ebn0_r0 (double r0);
extern double awgnbin_ebn0db_r0 (double r0);
extern double awgnbin_snr_1mr0 (double r);
extern double awgnbin_snrdb_1mr0 (double r);
extern double awgnbin_ebn0_1mr0 (double r);
extern double awgnbin_ebn0db_1mr0 (double r);

#endif
