// A Multiplicative Linear Congruential Generator
// for generating uniformly distributed pseudorandom numbers

// The following is mainly taken from the GNU C++ library,
// but I've modified it to use templates rather than overloaded virtual
// functions for speed reasons

// (C) 1997-1998 Frank Paehlke <frank@paehlke.de>
// Created: 12/15/1996
// $Id: mlcg.cc,v 1.1 2003/10/14 08:59:11 frank Exp $

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// $Log: mlcg.cc,v $
// Revision 1.1  2003/10/14 08:59:11  frank
// moved from . to src/
//
// Revision 1.3  2002/08/02 14:45:06  paehlke
// added GPL comment
//
// Revision 1.2  1998/05/28 12:56:31  paehlke
// Final version for masters thesis
//
// Revision 1.1  1997/06/20 18:04:44  frank
// Initial revision
//

#include "mlcg.h"

// SEED_TABLE_SIZE must be a power of 2

#define SEED_TABLE_SIZE 32

static _G_int32_t seedTable[SEED_TABLE_SIZE] = {
  0xbdcc47e5, 0x54aea45d, 0xec0df859, 0xda84637b,
  0xc8c6cb4f, 0x35574b01, 0x28260b7d, 0x0d07fdbf,
  0x9faaeeb0, 0x613dd169, 0x5ce2d818, 0x85b9e706,
  0xab2469db, 0xda02b0dc, 0x45c60d6e, 0xffe49d10,
  0x7224fea3, 0xf9684fc9, 0xfc7ee074, 0x326ce92a,
  0x366d13b5, 0x17aaa731, 0xeb83a675, 0x7781cb32,
  0x4ec7c92d, 0x7f187521, 0x2cf346b4, 0xad13310f,
  0xb89cff2b, 0x12164de1, 0xa865168d, 0x32b56cdf
};

void MLCG::reset()
{
  _G_int32_t seed1 = initialSeedOne;
  _G_int32_t seed2 = initialSeedTwo;

  // Most people pick stupid seed numbers that do not have enough
  // bits. In this case, if they pick a small seed number, we
  // map that to a specific seed.

  if (seed1 < 0) {
    seed1 = (seed1 + 2147483561);
    seed1 = (seed1 < 0) ? -seed1 : seed1;
  }

  if (seed2 < 0) {
    seed2 = (seed2 + 2147483561);
    seed2 = (seed2 < 0) ? -seed2 : seed2;
  }

  if (seed1 > -1 && seed1 < SEED_TABLE_SIZE) {
    seedOne = seedTable[seed1];
  } else {
    seedOne = seed1 ^ seedTable[seed1 & (SEED_TABLE_SIZE-1)];
  }

  if (seed2 > -1 && seed2 < SEED_TABLE_SIZE) {
    seedTwo = seedTable[seed2];
  } else {
    seedTwo = seed2 ^ seedTable[ seed2 & (SEED_TABLE_SIZE-1) ];
  }
  seedOne = (seedOne % 2147483561) + 1;
  seedTwo = (seedTwo % 2147483397) + 1;
}

_G_uint32_t MLCG::asLong()
{
  _G_int32_t k = seedOne % 53668;

  seedOne = 40014 * (seedOne-k * 53668) - k * 12211;
  if (seedOne < 0) {
    seedOne += 2147483563;
  }

  k = seedTwo % 52774;
  seedTwo = 40692 * (seedTwo - k * 52774) - k * 3791;
  if (seedTwo < 0) {
    seedTwo += 2147483399;
  }

  _G_int32_t z = seedOne - seedTwo;
  if (z < 1) {
    z += 2147483562;
  }
  return( (unsigned long) z);
}
