#ifndef _NORMAL_H
#define _NORMAL_H

// Templates for generating normal distributed pseudorandom numbers

// The following is mainly taken from the GNU C++ library,
// but I've modified it to use templates rather than overloaded
// virtual functions for speed reasons.

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// (C) 1997-1998 Frank Paehlke <frank@paehlke.de>
// Created: 12/15/1996
// $Id: normal.h,v 1.1 2003/10/14 08:59:11 frank Exp $

// $Log: normal.h,v $
// Revision 1.1  2003/10/14 08:59:11  frank
// moved from . to src/
//
// Revision 1.3  2002/08/02 14:45:06  paehlke
// added GPL comment
//
// Revision 1.2  1998/05/28 12:57:02  paehlke
// Final version for masters thesis
//
// Revision 1.1  1997/06/20 18:15:09  frank
// Initial revision
//

// --- BEGIN OF PUBLIC INTERFACE ---

template<class RNG> class Normal
{
public:

  Normal (RNG *gen, double xmean=0.0, double xvariance=1.0);
  double mean();
  double mean(double x);	// modify mean + return previous value
  double variance();
  double variance(double x);	// modify variance + return previous value
  virtual double operator()();	// the actual generator function

// --- END OF PUBLIC INTERFACE ---

private:

  RNG * pGenerator;
  double pMean;
  double pVariance;
  double pStdDev;
  char haveCachedNormal;
  double cachedNormal;
};

template<class RNG> inline Normal<RNG>::Normal
(RNG * gen, double xmean, double xvariance)
{
  pGenerator = gen;
  pMean = xmean;
  pVariance = xvariance;
  pStdDev = sqrt(pVariance);
  haveCachedNormal = 0;
}

template<class RNG> inline double Normal<RNG>::mean() { return pMean; };

template<class RNG> inline double Normal<RNG>::mean(double x)
{
  double t=pMean; pMean = x;
  return t;
}

template<class RNG> inline double Normal<RNG>::variance() { return pVariance; }

template<class RNG> inline double Normal<RNG>::variance(double x)
{
  double t=pVariance; pVariance = x;
  pStdDev = sqrt(pVariance);
  return t;
};


// See Simulation, Modelling & Analysis by Law & Kelton, pp259
// This is the ``polar'' method.

template<class RNG> double Normal<RNG>::operator () ()
{
  if (haveCachedNormal == 1) {
    haveCachedNormal = 0;
    return(cachedNormal * pStdDev + pMean );
  }
  else {
    for(;;) {
      double u1 = pGenerator -> asDouble();
      double u2 = pGenerator -> asDouble();
      double v1 = 2 * u1 - 1;
      double v2 = 2 * u2 - 1;
      double w = (v1 * v1) + (v2 * v2);

      // We actually generate two IID normal distribution variables.
      // We cache the one & return the other.

      if (w <= 1) {
	double y = sqrt( (-2 * log(w)) / w);
	double x1 = v1 * y;
	double x2 = v2 * y;
	haveCachedNormal = 1;
	cachedNormal = x2;
	return(x1 * pStdDev + pMean);
      }
    }
  }
}

#endif
