// Utilities for optimizing spherical and MPSK codes
// using threshold acceptance and steepest descent algorithms

// (C) 1997-1998 Frank Paehlke <frank@paehlke.de>
// Created: 10/23/1996
// $Id: opt-util.cc,v 1.1 2003/10/14 08:59:11 frank Exp $

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// $Log: opt-util.cc,v $
// Revision 1.1  2003/10/14 08:59:11  frank
// moved from . to src/
//
// Revision 1.6  2003/10/14 08:41:58  frank
// ANSI C++ / gcc 3.x compatibility changes
//
// Revision 1.5  2002/08/02 14:45:06  paehlke
// added GPL comment
//
// Revision 1.4  1998/05/28 12:56:32  paehlke
// Final version for masters thesis
//
// Revision 1.3  1997/09/23 13:01:18  frank
// fixed some bugs
//
// Revision 1.2  1997/06/20 14:12:32  frank
// added MPSK support
//
// Revision 1.1  1997/06/20 12:41:00  frank
// Initial revision
//

#include <math.h>
#include <fstream>
#include <iostream>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>

#include "vector.h"
#include "distcode.h"
#include "func1code.h"
#include "distpskcode.h"
#include "optimize.h"

using namespace std;

// int m = 24; // number of code vectors
// int n = 4;  // dimension of the euclidean vectorspace

char codeopt_ta_usage [] =
"Usage: codeopt-ta [OPTION...] CODEFILE\n\n"
"-h          display this help and exit\n"
"-o STRING   set output filename\n"
"-f INT      set function to optimize:\n"
"  -f 0        optimize minimal distance\n"
"  -f 1        optimize sum(exp(-4*d^2))\n"
"-s INT      set random seed\n"
"-t FLOAT    set starting threshold\n"
"-e FLOAT    set final threshold\n"
"-m FLOAT    set threshold decrement\n"
"-d FLOAT    set max. altering distance\n";

int codeopt_ta (int argc, char **argv)
{
  int opt;
  // Parameters of the TA algorithm:
  // threshold value to start with
  double thresh_start = 0.01;
  // threshold value at which the TA algorithm terminates
  double thresh_end = 0.00001;
  // factor by which the threshold gets decremented
  double thresh_mult = 0.95;
  // maximal distance to move a code vector in each TA step
  double alter_dist = 0.05;
  // seed for random number generation
  unsigned int seed = (unsigned int) time (NULL);
  int func = 0;
  ostream * out = &cout;
  istream * in = &cin;
  while ((opt = getopt (argc, argv, "ho:f:s:t:e:m:d:")) != EOF) {
    switch (opt) {
    case 'h': cout << codeopt_ta_usage; return 0;
    case 'o':
      out = new ofstream (optarg);
      if (out->fail ()) {
	cerr << argv[0] << ": error opening output file\n";
	return 1;
      }
      break;
    case 'f':
      func = atoi (optarg);
      if (func < 0 || func > 1) {
	cerr << argv[0] << ": unknown function\n";
	return 1;
      }
    case 's': seed = atoi (optarg); break;
    case 't': thresh_start = atof (optarg); break;
    case 'e': thresh_end = atof (optarg); break;
    case 'm': thresh_mult = atof (optarg); break;
    case 'd': alter_dist = atof (optarg); break;
    default: break;
    }
  }
  if (optind >= argc) {
    cerr << codeopt_ta_usage;
    return 1;
  }
  if (strcmp (argv[optind], "-")) {
    in = new ifstream (argv[optind]);
    if (in->fail ()) {
      cerr << argv[0] << ": error opening input file\n";
      return 1;
    }
  }
  optind++;
  if (optind < argc)
    cerr << argv[0] << ": ignoring extra argument(s)\n";
  // Finished parsing the arguments
  srand (seed);
  Code code (*in);
  switch (func) {
  case 0:
    {
      DistCode code1 (code);
      Optimize_TA (code1, code,
		   thresh_start, thresh_end, thresh_mult, alter_dist);
    } break;
  case 1:
    {
      Func1Code code1 (code);
      Optimize_TA (code1, code,
		   thresh_start, thresh_end, thresh_mult, alter_dist);
    } break;
  default:
    return 1;
  }
  (*out) << code;
  return 0;
}

char codeopt_sa_usage [] =
"Usage: codeopt-sa [OPTION...] CODEFILE\n\n"
"-h          display this help and exit\n"
"-o STRING   set output filename\n"
"-f INT      set function to optimize:\n"
"  -f 0        optimize minimal distance\n"
"  -f 1        optimize sum(exp(-4*d^2))\n"
"-s INT      set random seed\n";

int codeopt_sa (int argc, char **argv)
{
  int opt;
  ostream * out = &cout;
  istream * in = &cin;
  unsigned int seed = (unsigned int) time (NULL);
  int func = 0;
  while ((opt = getopt (argc, argv, "ho:f:s:")) != EOF) {
    switch (opt) {
    case 'h': cout << codeopt_sa_usage; return 0;
    case 'o':
      out = new ofstream (optarg);
      if (out->fail ()) {
	cerr << argv[0] << ": error opening output file\n";
	return 1;
      }
      break;
    case 'f':
      func = atoi (optarg);
      if (func < 0 || func > 1) {
	cerr << argv[0] << ": unknown function\n";
	return 1;
      }
    case 's': seed = atoi (optarg); break;
    default: break;
    }
  }
  if (optind >= argc) {
    cerr << codeopt_sa_usage;
    return 1;
  }
  if (strcmp (argv[optind], "-")) {
    in = new ifstream (argv[optind]);
    if (in->fail ()) {
      cerr << argv[0] << ": error opening input file\n";
      return 1;
    }
  }
  optind++;
  if (optind < argc)
    cerr << argv[0] << ": ignoring extra argument(s)\n";
  // Finished parsing the arguments
  srand (seed);
  Code code (*in);
  switch (func) {
  case 0:
    {
      DistCode code1 (code);
      Optimize_Phase2 (code1, code);
    } break;
  case 1:
    {
    Func1Code code1 (code);
    Optimize_Phase2 (code1, code);
    } break;
  default:
    return 1;
  }
  (*out) << code;
  return 0;
}

char pskopt_ta_usage [] =
"Usage: pskopt-ta [OPTION...] PSKFILE\n\n"
"-h          display this help and exit\n"
"-o STRING   set output filename\n"
"-f INT      set function to optimize:\n"
"  -f 0        optimize minimal distance\n"
// "  -f 1        optimize sum(exp(-4*d^2))\n"
"-s INT      set random seed\n"
"-t FLOAT    set starting threshold\n"
"-e FLOAT    set final threshold\n"
"-m FLOAT    set threshold decrement\n"
"-d FLOAT    set max. altering distance\n";

int pskopt_ta (int argc, char **argv)
{
  int opt;
  // Parameters of the TA algorithm:
  // threshold value to start with
  double thresh_start = 0.01;
  // threshold value at which the TA algorithm terminates
  double thresh_end = 0.00001;
  // factor by which the threshold gets decremented
  double thresh_mult = 0.95;
  // maximal distance to move a code vector in each TA step
  double alter_dist = .05;
  // seed for random number generation
  unsigned int seed = (unsigned int) time (NULL);
  int func = 0;
  ostream * out = &cout;
  istream * in = &cin;
  while ((opt = getopt (argc, argv, "ho:f:s:t:e:m:d:")) != EOF) {
    switch (opt) {
    case 'h': cout << pskopt_ta_usage; return 0;
    case 'o':
      out = new ofstream (optarg);
      if (out->fail ()) {
	cerr << argv[0] << ": error opening output file\n";
	return 1;
      }
      break;
    case 'f':
      func = atoi (optarg);
      if (func < 0 || func > 0) {
	cerr << argv[0] << ": unknown function\n";
	return 1;
      }
    case 's': seed = atoi (optarg); break;
    case 't': thresh_start = atof (optarg); break;
    case 'e': thresh_end = atof (optarg); break;
    case 'm': thresh_mult = atof (optarg); break;
    case 'd': alter_dist = atof (optarg); break;
    default: break;
    }
  }
  if (optind >= argc) {
    cerr << pskopt_ta_usage;
    return 1;
  }
  if (strcmp (argv[optind], "-")) {
    in = new ifstream (argv[optind]);
    if (in->fail ()) {
      cerr << argv[0] << ": error opening input file\n";
      return 1;
    }
  }
  optind++;
  if (optind < argc)
    cerr << argv[0] << ": ignoring extra argument(s)\n";
  // Finished parsing the arguments
  srand (seed);
  PSKCode code (*in);
  switch (func) {
  case 0:
    {
      DistPSKCode code1 (code);
      Optimize_TA (code1, code,
		   thresh_start, thresh_end, thresh_mult, alter_dist);
    } break;
//  case 1:
//    {
//      Func1Code code1 (code);
//      Optimize_TA (code1, code,
//		   thresh_start, thresh_end, thresh_mult, alter_dist);
//    } break;
  default:
    return 1;
  }
  (*out) << code;
  return 0;
}

char pskopt_sa_usage [] =
"Usage: pskopt-sa [OPTION...] PSKFILE\n\n"
"-h          display this help and exit\n"
"-o STRING   set output filename\n"
"-f INT      set function to optimize:\n"
"  -f 0        optimize minimal distance\n"
// "  -f 1        optimize sum(exp(-4*d^2))\n"
"-s INT      set random seed\n";

int pskopt_sa (int argc, char **argv)
{
  int opt;
  ostream * out = &cout;
  istream * in = &cin;
  unsigned int seed = (unsigned int) time (NULL);
  int func = 0;
  while ((opt = getopt (argc, argv, "ho:f:s:")) != EOF) {
    switch (opt) {
    case 'h': cout << pskopt_sa_usage; return 0;
    case 'o':
      out = new ofstream (optarg);
      if (out->fail ()) {
	cerr << argv[0] << ": error opening output file\n";
	return 1;
      }
      break;
    case 'f':
      func = atoi (optarg);
      if (func < 0 || func > 0) {
	cerr << argv[0] << ": unknown function\n";
	return 1;
      }
    case 's': seed = atoi (optarg); break;
    default: break;
    }
  }
  if (optind >= argc) {
    cerr << pskopt_sa_usage;
    return 1;
  }
  if (strcmp (argv[optind], "-")) {
    in = new ifstream (argv[optind]);
    if (in->fail ()) {
      cerr << argv[0] << ": error opening input file\n";
      return 1;
    }
  }
  optind++;
  if (optind < argc)
    cerr << argv[0] << ": ignoring extra argument(s)\n";
  // Finished parsing the arguments
  srand (seed);
  PSKCode code (*in);
  switch (func) {
  case 0:
    {
      DistPSKCode code1 (code);
      Optimize_Phase2 (code1, code);
    } break;
//  case 1:
//    {
//    Func1Code code1 (code);
//    Optimize_Phase2 (code1, code);
//    } break;
  default:
    return 1;
  }
  (*out) << code;
  return 0;
}

int main (int argc, char ** argv)
{
  char * a0 = strrchr (argv[0], '/');
  if (a0 == NULL)
    a0 = argv [0];
  else
    a0++;
  if (!strcmp (a0, "codeopt-ta"))
    return codeopt_ta (argc, argv);
  else if (!strcmp (a0, "codeopt-sa"))
    return codeopt_sa (argc, argv);
  else if (!strcmp (a0, "pskopt-ta"))
    return pskopt_ta (argc, argv);
  else if (!strcmp (a0, "pskopt-sa"))
    return pskopt_sa (argc, argv);
  else {
    cerr << a0 << " cannot be run directly\n";
    return 1;
  }
}
