// Calculation of bit error probabilities of geometrical codes
// on a simulated USC with erasures

// (C) 1997-1998 Frank Paehlke <frank@paehlke.de>
// $Id: pe-algcode.cc,v 1.1 2003/10/14 08:59:11 frank Exp $

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// $Log: pe-algcode.cc,v $
// Revision 1.1  2003/10/14 08:59:11  frank
// moved from . to src/
//
// Revision 1.3  2003/10/14 08:41:58  frank
// ANSI C++ / gcc 3.x compatibility changes
//
// Revision 1.2  2002/08/02 14:45:06  paehlke
// added GPL comment
//
// Revision 1.1  1998/05/28 12:56:32  paehlke
// Final version for masters thesis
//

#include <math.h>
#include <fstream>
#include <iostream>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>

#include "mlcg.h"
#include "rscode.h"
#include "gf.h"

using namespace std;

#ifdef GF16
#include "gf16tables.h"
typedef GF<GF16Tables> GFq;
#else
#ifdef GF32
#include "gf32tables.h"
typedef GF<GF32Tables> GFq;
#else
#ifdef GF64
#include "gf64tables.h"
typedef GF<GF64Tables> GFq;
#else
#ifdef GF128
#include "gf128tables.h"
typedef GF<GF128Tables> GFq;
#else
#include "gf256tables.h"
typedef GF<GF256Tables> GFq;
#endif
#endif
#endif
#endif

#ifdef GBCH256
#define GBCH
#include "gf256tables.h"
typedef GF<GF256Tables> GFqm;
#else
#ifdef GBCH1024
#define GBCH
#include "gf1024tables.h"
typedef GF<GF1024Tables> GFqm;
#else
#ifdef GBCH4096
#define GBCH
#include "gf4096tables.h"
typedef GF<GF4096Tables> GFqm;
#endif
#endif
#endif

#ifdef GBCH
#include "gbchcode.h"
typedef GBCHCode<GFq,GFqm> CODE;
#else
#ifdef XRS
#include "xrscode.h"
typedef XRSCode<GFq> CODE;
#define EXPURG
#else
#ifdef SXRS
#include "sxrscode.h"
typedef SXRSCode<GFq> CODE;
#define EXPURG
#else
#include "rscode.h"
typedef RSCode<GFq> CODE;
#endif
#endif
#endif

typedef double * pDouble;

CODE * code; // algebraical code

MLCG * rg; // random generator

int loops;        // number of loops per SNR value
int d, n, k;      // parameters of algebraical code
#ifdef EXPURG
int xd;
#endif

GFq * ibuf_orig; // original information buffer
GFq * ibuf;      // decoded information buffer
GFq * cbuf;      // code buffer
int * erasvec;   // erasure vector

int eras, werr, serr; // number of erasures/word errors/symbol errors

#ifndef NDEBUG
void PrintVec (char * title, GFq * vec, int len)
{
  int i;
  cerr << title << " = [";
  if (len > 0)
    cerr << (int) vec[0];
  for (i=1; i<len; i++)
    cerr << ',' << (int) vec[i];
  cerr << "]\n";
}
#endif

void InitCodes ()
{
#ifdef EXPURG
  code = new CODE (d, xd, n);
#else
  code = new CODE (d, n);
#endif
  k = code->get_k();
  // allocate buffer memory
  ibuf_orig = new GFq [k];
  ibuf = new GFq [k];
  cbuf = new GFq [n];
  erasvec = new int [n];
}

void ClearCodes ()
{
  delete [] erasvec;
  delete [] cbuf;
  delete [] ibuf;
  delete [] ibuf_orig;
  delete code;
}

void PrintHeader (ostream & out)
{
  out << "# Error probabilities on a symmetric DMC\n"
#ifdef GBCH
      << "# for a generalized BCH code with "
#else
#ifdef XRS
      << "# for an expurgated RS code with "
#else
#ifdef SXRS
      << "# for a systematic expurgated RS code with "
#else
      << "# for a RS code with "
#endif
#endif
#endif
      << "q = " << GFq::q() << ", k = " << k << ", n = " << n
      << ", r = " << code->get_r()
#ifdef EXPURG
      << ", xd = " << xd
#endif
      << ", d = " << d << endl;
  out << "# channel\t\toutput\n"
      << "# ERAS\t SER\t ERAS\t WER\t SER\t BER\n";
}

void PrintOutputLine (ostream & out, double cheras, double chserr)
{
  int symbols = loops * k; // number of transmitted symbols
  out << cheras << '\t' << chserr << '\t'
      << (double)eras / loops << '\t'
      << (double)werr / loops << '\t'
      << (double)serr / symbols << '\t'
      << 0.5 * serr / symbols << endl;
}

void ErrorCurve (ostream & out)
{
  int e1,e2,i,j;
  double cheras,chserr;
  for (e1=4; e1<=20; e1++) {
    cheras = pow (10.0, -0.25*e1);
    for (e2=1; e2<=5; e2++) {
      chserr = pow (10.0, -e2);
      eras = werr = serr = 0;
      for (i=0; i<loops; i++) {
	for (j=0; j<k; j++)
	  ibuf_orig[j] = (int) floor (GFq::q()*rg->asDouble());
	code->Encode (cbuf, ibuf_orig);
	int erasnum = 0;
	for (j=0; j<n; j++) {
	  double r = rg->asDouble();
	  if (r < cheras + chserr) {
	    cbuf[j] += GFq ((int) floor (1 + (GFq::q()-1) * rg->asDouble()));
	    if (r < cheras)
	      erasvec[erasnum++] = j;
	  }
	}
	if (code->Decode (ibuf, cbuf, erasvec, erasnum) < 0)
	  eras++;
	else {
	  int serr0 = serr;
	  for (j=0; j<k; j++)
	    if (ibuf_orig[j] != ibuf[j])
	      serr++;
	  if (serr > serr0)
	    werr++;
	}
      }
      PrintOutputLine (out, cheras, chserr);
    }
  }
}

void SetDefaults ()
{
  loops = 1000;
  n = GFq::q() - 1;
  d = 5; // correct up to 2 errors
#ifdef EXPURG
  xd = 4;
#endif
}

char * usage = // "Usage: " << argv[0] <<
" [OPTION...] CODEFILE\n\n"
"-h         display this help and exit\n"
"-o STRING  set output filename\n"
"-l INT     set # of loops per output line\n"
#ifdef EXPURG
"-c INT[,INT[,INT]]  set code parameters (dmin,xd,n)\n";
#else
"-c INT[,INT]  set code parameters (dmin,n)\n";
#endif

int main (int argc, char ** argv)
{
  int opt;
  int i1,i2;
#ifdef EXPURG
  int i3;
#endif
  char dummy;
  ostream * out = &cout;

  SetDefaults ();
  while ((opt = getopt (argc, argv, "ho:l:c:")) != EOF) {
    switch (opt) {

    case 'h':
      cout << "Usage: " << argv[0] << usage;
      return 0;

    case 'o':
      out = new ofstream (optarg);
      if (out->fail ()) {
	cerr << argv[0] << ": error opening output file " << optarg << "\n";
	return 1;
      }
      break;

    case 'l':
      if (sscanf (optarg, "%d%c", &i1, &dummy) != 1) {
	cerr << argv[0] << ": malformed parameters for option -l: "
	     << optarg << endl;
	return 1;
      }
      if (i1 < 1) {
	cerr << argv[0] << ": option -l: number of loops must be positive"
	  " (was " << i1 << ")\n";
	return 1;
      }
      loops = i1;
      break;

    case 'c':
      // code parameters
#ifdef EXPURG
      switch (sscanf (optarg, "%d,%d,%d%c", &i1, &i2, &i3, &dummy)) {
      case 3:
	if (i3 < 1 || i3 >= GFq::q()) {
	  cerr << argv[0] <<
	    ": option -c: length must be between 1 and " << GFq::q()-1 <<
	    " (was " << i3 << ")\n";
	  return 1;
	}
	n = i3;
      case 2:
	if (i2 <= 1) {
	  cerr << argv[0] <<
	    ": option -c: puncturing distance must be bigger than 1"
	    " (was " << i2 << ")\n";
	  return 1;
	}
	xd = i2;
#else
      switch (sscanf (optarg, "%d,%d%c", &i1, &i2, &dummy)) {
      case 2:
	if (i2 < 1 || i2 >= GFq::q()) {
	  cerr << argv[0] <<
	    ": option -c: length must be between 1 and " << GFq::q()-1 <<
	    " (was " << i2 << ")\n";
	  return 1;
	}
	n = i2;
#endif
      case 1:
	if (i1 < 1) {
	  cerr << argv[0] <<
	    ": option -c: code distance must be positive"
	    " (was " << i1 << ")\n";
	  return 1;
	}
	d = i1;
	break;
      default:
	cerr << argv[0] << ": malformed parameters for option -c: "
	     << optarg << "\n";
	return 1;
      }
      break;

    } // switch
  } // while

  if (optind < argc) {
    cerr << argv[0] << ": ignoring extra argument(s):";
    while (optind < argc)
      cerr << ' ' << argv[optind++];
    cerr << endl;
  }

  // do some consistency checks
  if (d > n) {
    cerr << argv[0] << ": code distance (" << d
	 << ") must not be bigger than code length (" << n
	 << ")\n";
    return 1;
  }

  // initialize random generator
  rg = new MLCG ((_G_int32_t) time (NULL), (_G_int32_t) time (NULL));
  InitCodes ();
  PrintHeader (*out);
  ErrorCurve (*out);
  ClearCodes ();
  delete rg;
}
