// Calculation of the error probability of a concatenation
// of a geometrical code and an algebraical code
// on a simulated time-discrete AWGN channel

// (C) 1997-1998 Frank Paehlke <frank@paehlke.de>
// $Id: pe-geo-alg.cc,v 1.1 2003/10/14 08:59:11 frank Exp $

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// $Log: pe-geo-alg.cc,v $
// Revision 1.1  2003/10/14 08:59:11  frank
// moved from . to src/
//
// Revision 1.5  2003/10/14 08:41:58  frank
// ANSI C++ / gcc 3.x compatibility changes
//
// Revision 1.4  2002/08/02 14:45:06  paehlke
// added GPL comment
//
// Revision 1.3  1998/05/28 12:56:33  paehlke
// Final version for masters thesis
//
// Revision 1.2  1997/11/04 09:33:26  paehlke
// various corrections,
// made consistent with p-error-gbch.cc
//

#include <math.h>
#include <fstream>
#include <iostream>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>

#include "mlcg.h"
#include "normal.h"
#include "code.h"
#include "gmddecoder.h"
#include "gf.h"

using namespace std;

#ifdef GF16
#include "gf16tables.h"
typedef GF<GF16Tables> GFq;
#else
#ifdef GF32
#include "gf32tables.h"
typedef GF<GF32Tables> GFq;
#else
#ifdef GF64
#include "gf64tables.h"
typedef GF<GF64Tables> GFq;
#else
#ifdef GF128
#include "gf128tables.h"
typedef GF<GF128Tables> GFq;
#else
#include "gf256tables.h"
typedef GF<GF256Tables> GFq;
#endif
#endif
#endif
#endif

#ifdef GBCH256
#define GBCH
#include "gf256tables.h"
typedef GF<GF256Tables> GFqm;
#else
#ifdef GBCH1024
#define GBCH
#include "gf1024tables.h"
typedef GF<GF1024Tables> GFqm;
#else
#ifdef GBCH4096
#define GBCH
#include "gf4096tables.h"
typedef GF<GF4096Tables> GFqm;
#endif
#endif
#endif

#ifdef GBCH
#include "gbchcode.h"
typedef GBCHCode<GFq,GFqm> ALGCODE;
#else
#ifdef XRS
#include "xrscode.h"
typedef XRSCode<GFq> ALGCODE;
#define EXPURG
#else
#ifdef SXRS
#include "sxrscode.h"
typedef SXRSCode<GFq> ALGCODE;
#define EXPURG
#else
#include "rscode.h"
typedef RSCode<GFq> ALGCODE;
#endif
#endif
#endif

typedef GMDDecoder<GFq,ALGCODE> GMDD;
typedef double * pDouble;

Code * code;   // geometrical code
ALGCODE * algcode; // algebraical code
GMDD * gmdd; // GMD decoders

MLCG * rg; // random generator
Normal<MLCG> * noise;

double snr_start; // initial SNR value
double snr_incr;  // SNR increment
int snr_num;      // number of SNR values
int snr_db;       // interpret SNR values as dB?
int loops;        // number of loops per SNR value
int d, n, k;      // parameters of algebraical code
#ifdef EXPURG
int xd;
#endif
int d_func;       // distance function for geometrical decoding
int algo;         // algorithm for algebraical decoding
int cl_num;       // # of lists for candidate list decoding
int cl_len;       // list length for candidate list decoding
char codefilename [256];

GFq * infobuf0;  // encoder information buffer
GFq * codebuf0;  // encoder code buffer
Vector * sent;     // sent signal vector
Vector * rcvd;     // received signal vector
GFq * codebuf;   // decoder code buffer
GFq * infobuf;   // decoder information buffer
pDouble * alphabuf; // soft decision buffers
double * betabuf;

int eras, werr, serr; // number of erasures/word errors/symbol errors
double complexity; // summarized complexity per info symbol
// absolute frequency distributions of ld(complexity per info symbol):
int stat_i[32], stat_o[32];

// enumeration of decoding algorithms

enum AlgorithmID {algBM = 0, algCLIST, algPGMD, algGMD, algMVUC};

#ifndef NDEBUG
void PrintVec (char * title, GFq * vec, int len)
{
  int i;
  cerr << title << " = [";
  if (len > 0)
    cerr << (int) vec[0];
  for (i=1; i<len; i++)
    cerr << ',' << (int) vec[i];
  cerr << "]\n";
}
#endif

void Transmission (const Vector & sent, Vector & rcvd)
{
  int i;
  for (i=0; i<code->get_n(); i++) {
    rcvd[i] = sent[i] + (*noise)();
  }
}

void InitCodes ()
{
  int i;
  sent = new Vector [n];
  for (i = 0; i < n; i++)
    sent[i].allocate (code->get_n());
  rcvd = new Vector [n];
  for (i = 0; i < n; i++)
    rcvd[i].allocate (code->get_n());
#ifdef EXPURG
  algcode = new ALGCODE (d, xd, n);
#else
  algcode = new ALGCODE (d, n);
#endif
  k = algcode->get_k();
  if (algo == algCLIST || algo == algPGMD || algo == algGMD)
    gmdd = new GMDD (algcode);
  // allocate buffer memory
  infobuf0 = new GFq [k];
  infobuf = new GFq [k];
  codebuf0 = new GFq [n];
  codebuf = new GFq [n];
  alphabuf = new pDouble [n];
  for (i = 0; i < n; i++)
    alphabuf[i] = new double [GFq::q()];
  betabuf = new double [n];
}

void ClearCodes ()
{
  int i;
  delete [] betabuf;
  for (i = 0; i < n; i++)
    delete [] alphabuf[i];
  delete [] alphabuf;
  delete [] codebuf;
  delete [] codebuf0;
  delete [] infobuf;
  delete [] infobuf0;
  if (algo == algCLIST || algo == algPGMD || algo == algGMD)
    delete gmdd;
  delete algcode;
  for (i = 0; i < n; i++)
    rcvd[i].deallocate();
  delete [] rcvd;
  for (i = 0; i < n; i++)
    sent[i].deallocate();
  delete [] sent;
}

void PrintHeader (ostream & out)
{
  int i;
  double R = code->get_r() * algcode->get_r();
  out << "# Error probabilities for code concatenation with R = "
      << R << endl
      << "# " << loops << " loops per SNR value\n"
      << "# geometrical code: " << codefilename
      << ": m = " << code->get_m() << ", n = " << code->get_n()
      << ", r = " << code->get_r()
      << ", dmin = " << code->MinDist() << endl
#ifdef GBCH
      << "# generalized BCH code: k = " << k
#else
#ifdef EXPURG
      << "# expurgated RS code: k = " << k
#else
      << "# RS code: k = " << k
#endif
#endif
      << ", n = " << n
      << ", r = " << algcode->get_r()
#ifdef EXPURG
      << ", xd = " << xd
#endif
      << ", d = " << d << endl;
  switch (algo) {
  case algCLIST:
    out << "# candidate list decoding (" << cl_num
	<< " lists of length " << cl_len << ")\n";
    break;
  case algPGMD:
    out << "# primitive GMD decoding\n";
    break;
  case algGMD:
    out << "# GMD decoding\n";
  }
  out << "# SNR\t SNR(dB)\t Eb/N0(dB)\t E+E\t ERAS\t WER\t SER\t BER";
  if ((algo == algCLIST) || (algo == algPGMD) || (algo == algGMD)) {
    out << "\t loops";
    int m = 10000 / k;
    for (i = 0; (1<<i) <= m; i++)
      out << "\t ldi=" << i;
  }
  out << endl;
}

void PrintOutputLine (ostream & out, double snr, double snrdb, double ebn0db)
{
  int i;
  int symbols = loops * k; // number of transmitted symbols
  out << snr << '\t' << snrdb << '\t' << ebn0db << '\t'
      << (double)(eras+werr) / loops << '\t'
      << (double)eras / loops << '\t'
      << (double)werr / loops << '\t'
      << (double)serr / symbols << '\t'
      << 0.5 * serr / symbols;
  if ((algo == algCLIST) || (algo == algPGMD) || (algo == algGMD)) {
    out << '\t' << (double)complexity / loops << '\t';
    int m = 10000 / k;
    out << '\t' << stat_i[0];
    for (i=1; (1<<i) <= m; i++) {
      assert (i < 32);
      out << ' ' << stat_i[i];
    }
  }
  out << endl;
}

// DecodeGeometricalCode(pos) decodes the received signal vector src.
// The received symbol is written to dest[]
// and any soft-decision information is written to alpha[].

void DecodeGeometricalCode (const Vector & src, GFq & dest,
			    pDouble alpha, double & beta)
{
  int destsym;
  switch (algo) {
  case algBM:
    code->Decode (src, destsym);
    break;
  case algCLIST:
  case algPGMD:
  case algGMD:
    beta = code->LLHDecode (noise->variance(), src, destsym, alpha);
    break;
  default:
    cerr << "Undefined algorithm ID " << algo << endl;
    exit (1);
  }
  dest = GFq(destsym);
}

int DecodeAlgebraicalCode (GFq * dest, const GFq * src,
			   const double * const * alpha,
			   const double * beta)
{
  int res;
  switch (algo) {
  case algBM:
    res = algcode->Decode (dest, src);
    break;
  case algCLIST:
    res = gmdd->CandidateListDecode (dest, src, alpha, beta,
				     cl_num, cl_len);
    break;
  case algPGMD:
    res = gmdd->PrimitiveDecode (dest, src, beta);
    break;
  case algGMD:
    res = gmdd->Decode (dest, src, alpha, beta);
    break;
  case algMVUC:
    cerr << "Decoding algorithm " << algo << " is not yet implemented "
      "for inner code\n";
    exit (1);
    break;
  default: exit(1);
  }
  if (algo >= algCLIST && algo <= algGMD) {
    // update complexity statistics
    int i = 1;
    double gmdloops = (double)gmdd->GetLoops() / k;
    complexity += gmdloops;
    while ((1<<i) <= gmdloops)
      i++;
    // now i-1 == ld(gmdloops)
    assert ((i > 0) && (i <= 32) &&
	    ((1<<(i-1)) <= 10000 / k));
    stat_i[i-1]++;
  }
  return res;
}

void SimulateCode ()
{
  int i,res;
  // fill info buffer with random symbols
  for (i = 0; i < k; i++)
    infobuf0[i] = (int) floor (GFq::q()*rg->asDouble());
  // apply algebraical code
  algcode->Encode (codebuf0, infobuf0);
  for (i = 0; i < n; i++) {
    // transmit i-th code symbol
    sent[i].copy ((*code)[codebuf0[i]]);
    Transmission (sent[i], rcvd[i]);
    DecodeGeometricalCode (rcvd[i], codebuf[i],
			   alphabuf[i], betabuf[i]);
  }
  res = DecodeAlgebraicalCode (infobuf, codebuf, alphabuf, betabuf);
#if 1
  cerr << "gesendetes Infowort: \\\\\n";
  for (i = 0; i < k; i++)
    cerr << " " << infobuf0[i];
  cerr << "\n\n";
  cerr << "gesendetes Codewort: \\\\\n";
  for (i = 0; i < n; i++)
    cerr << " " << codebuf0[i];
  cerr << "\n\n";
  cerr << "gesendetes Signal: \\\\\n";
  for (i = 0; i < n; i++) {
    cerr << " \\luecke\\mbox{";
    for (int j = 0; j < code->get_n(); j++) {
      double val = sent[i][j];
      if (val > -.001 && val < .001) val = 0;
      if (val < 0)
	cerr << "\\nbalken{" << -10*val << "}";
      else
	cerr << "\\pbalken{" << 10*val << "}";
    }
    cerr << "}%\n";
  }
  cerr << "\n\n";
  cerr << "Rauschen: \\\\\n";
  for (i = 0; i < n; i++) {
    cerr << " \\luecke\\mbox{";
    for (int j = 0; j < code->get_n(); j++) {
      double val = rcvd[i][j] - sent[i][j];
      if (val > -.001 && val < .001) val = 0;
      if (val < 0)
	cerr << "\\nbalken{" << -10*val << "}";
      else
	cerr << "\\pbalken{" << 10*val << "}";
    }
    cerr << "}%\n";
  }
  cerr << "\n\n";
  cerr << "empfangenes Signal: \\\\\n";
  for (i = 0; i < n; i++) {
    cerr << " \\luecke\\mbox{";
    for (int j = 0; j < code->get_n(); j++) {
      double val = rcvd[i][j];
      if (val > -.001 && val < .001) val = 0;
      if (val < 0)
	cerr << "\\nbalken{" << -10*val << "}";
      else
	cerr << "\\pbalken{" << 10*val << "}";
    }
    cerr << "}%\n";
  }
  cerr << "\n\n";
  cerr << "empfangenes Codewort: \\\\\n";
  for (i = 0; i < n; i++) {
    if (codebuf[i] != codebuf0[i])
      cerr << " \\textbf{" << codebuf[i] << "}";
    else
      cerr << " " << codebuf[i];
  }
  cerr << " \\\\\n";
  cerr << "decodiertes Infowort:";
  for (i = 0; i < k; i++) {
    if (infobuf[i] != infobuf0[i])
      cerr << " \\textbf{" << infobuf[i] << "}";
    else
      cerr << " " << infobuf[i];
  }
  cerr << "\n\n";
#endif
  // update error statistics
  if (res < 0)
    eras++;
  else {
    int serr0 = serr;
    for (i = 0; i < k; i++)
      if (infobuf[i] != infobuf0[i])
	serr++;
    if (serr > serr0)
      werr++;
  }
}

void PrintErrorCurve (ostream & out)
{
  int s,i;
  // compute code rate
  double R = code->get_r() * algcode->get_r();
  for (s=0; s<snr_num; s++) {
    // set SNR and Eb/N0
    double snr, snrdb, ebn0, ebn0db;
    if (snr_db) {
      snrdb = snr_start + s * snr_incr;
      snr = pow (10.0, snrdb/10.0);
    }
    else {
      snr = snr_start + s * snr_incr;
      snrdb = 10.0 * log10 (snr);
    }
    ebn0 = snr / (2*R);
    ebn0db = 10.0 * log10 (ebn0);
    noise->variance (1.0/(snr * code->get_n()));
    // init counters and statistics
    eras = werr = serr = 0;
    complexity = 0.0;
    for (i=0; i<32; i++)
      stat_i[i] = stat_o[i] = 0;
    // now do the actual simulation
    for (i = 0; i < loops; i++)
      SimulateCode ();
    PrintOutputLine (out, snr, snrdb, ebn0db);
  }
}

void SetDefaults ()
{
  snr_start = 1.0;
  snr_incr = 0.1;
  snr_num = 1;
  snr_db = 0; // SNR values are not interpreted as dB
  loops = 1000;
  n = GFq::q() - 1;
  d = 5; // correct up to 2 errors
#ifdef EXPURG
  xd = 4;
#endif
  d_func = 0; // use euclidean distance for geometrical decoding
  algo = algBM; // use Berlekamp-Massey for algebraical decoding
  cl_num = 8;
  cl_len = 2;
}

char * usage = // "Usage: " << argv[0] <<
" [OPTION...] CODEFILE\n\n"
"-h   display this help and exit\n"
"-o STRING  set output filename\n"
"-s FLOAT   set initial SNR\n"
"-i FLOAT   set SNR increment\n"
"-n INT     set # of SNR values\n"
"-l INT     set # of loops per SNR value\n"
"-d         interpret SNR values as dB\n"
"-f INT     set distance function for geometrical decoding\n"
"   -f 0       euclidean distance (default)\n"
"   -f 1       spherical distance\n"
"   -f 2       a-posteriori probability\n"
#ifdef EXPURG
"-c INT[,INT[,INT]]  set parameters (dmin,xd,n) of algebraical code\n"
#else
"-c INT[,INT]  set parameters (dmin,n) of algebraical code\n"
#endif
"-a INT     set decoding algorithm for algebraical code:\n"
"   -a 0       Berlekamp-Massey (default)\n"
"   -a 1[,INT[,INT]]    candidate lists (parameters: number,length)\n"
"   -a 2       primitive GMD\n"
"   -a 3       normal GMD\n"
"   -a 4       modified Vucetic\n";

int main (int argc, char ** argv)
{
  int opt;
  int i1,i2,i3;
  char dummy;
  double d1;
  ostream * out = &cout;

  SetDefaults ();
  while ((opt = getopt (argc, argv, "ho:s:i:n:l:df:c:a:")) != EOF) {
    switch (opt) {

    case 'h':
      cout << "Usage: " << argv[0] << usage;
      return 0;

    case 'o':
      out = new ofstream (optarg);
      if (out->fail ()) {
	cerr << argv[0] << ": error opening output file " << optarg << "\n";
	return 1;
      }
      break;

    case 's':
      if (sscanf (optarg, "%lf%c", &d1, &dummy) != 1) {
	cerr << argv[0] << ": malformed parameters for option -s: "
	     << optarg << endl;
	return 1;
      }
      snr_start = d1;
      break;

    case 'i':
      if (sscanf (optarg, "%lf%c", &d1, &dummy) != 1) {
	cerr << argv[0] << ": malformed parameters for option -i: "
	     << optarg << endl;
	return 1;
      }
      snr_incr = d1;
      break;

    case 'n':
      if (sscanf (optarg, "%d%c", &i1, &dummy) != 1) {
	cerr << argv[0] << ": malformed parameters for option -n: "
	     << optarg << endl;
	return 1;
      }
      if (i1 < 1) {
	cerr << argv[0] << ": option -n: number of SNR values must be positive"
	  " (was " << i1 << ")\n";
	return 1;
      }
      snr_num = i1;
      break;

    case 'l':
      if (sscanf (optarg, "%d%c", &i1, &dummy) != 1) {
	cerr << argv[0] << ": malformed parameters for option -l: "
	     << optarg << endl;
	return 1;
      }
      if (i1 < 1) {
	cerr << argv[0] << ": option -l: number of loops must be positive"
	  " (was " << i1 << ")\n";
	return 1;
      }
      loops = i1;
      break;

    case 'd':
      snr_db = 1; // interpret SNR values as dB
      break;

    case 'f':
      // distance function
      if (sscanf (optarg, "%d%c", &i1, &dummy) != 1) {
	cerr << argv[0] << ": malformed parameters for option -f: "
	     << optarg << endl;
	return 1;
      }
      if (i1 < 0 || i1 > 2) {
	cerr << argv[0] <<
	  ": option -f: function ID must be between 0 and 2"
	  " (was " << i1 << ")\n";
	return 1;
      }
      d_func = i1;
      break;

    case 'c':
      // parameters of inner code
#ifdef EXPURG
      switch (sscanf (optarg, "%d,%d,%d%c", &i1, &i2, &i3, &dummy)) {
      case 3:
	if (i3 < 1 || i3 >= GFq::q()) {
	  cerr << argv[0] <<
	    ": option -c: length must be between 1 and " << GFq::q()-1 <<
	    " (was " << i3 << ")\n";
	  return 1;
	}
	n = i3;
      case 2:
	if (i2 <= 1) {
	  cerr << argv[0] <<
	    ": option -c: expurgation distance must be bigger than 1"
	    " (was " << i2 << ")\n";
	  return 1;
	}
	xd = i2;
#else
      switch (sscanf (optarg, "%d,%d%c", &i1, &i2, &dummy)) {
      case 2:
	if (i2 < 1 || i2 >= GFq::q()) {
	  cerr << argv[0] <<
	    ": option -c: length must be between 1 and " << GFq::q()-1 <<
	    " (was " << i2 << ")\n";
	  return 1;
	}
	n = i2;
#endif
      case 1:
	if (i1 < 1) {
	  cerr << argv[0] <<
	    ": option -c: code distance must be positive"
	    " (was " << i1 << ")\n";
	  return 1;
	}
	d = i1;
	break;
      default:
	cerr << argv[0] << ": malformed parameters for option -c: "
	     << optarg << "\n";
	return 1;
      }
      break;

      case 'a':
	// algorithm for algebraical decoding
	switch (sscanf (optarg, "%d,%d,%d%c", &i1, &i2, &i3, &dummy)) {
	case 3:
	  if (i1 == algCLIST)
	    cl_len = i3;
	  else
	    cerr << argv[0] << ": option -a: ignoring second parameter "
		 << i3 << endl;
	case 2:
	  if (i1 == algCLIST) {
	    if (i2 < 1) {
	      cerr << argv[0] << ": option -A: number of lists "
		"must be positive (was " << i2 << ")\n";
	      return 1;
	    }
	    cl_num = i2;
	  }
	  else
	    cerr << argv[0] << ": option -a: ignoring first parameter "
		 << i2 << "\n";
	case 1:
	  if ((i1 < algBM) || (i1 > algMVUC)) {
	    cerr << argv[0] <<
	      ": option -a: invalid algorithm ID" << i1 << ")\n";
	    return 1;
	  }
	  algo = i1;
	  break;
	default:
	  cerr << argv[0] << ": malformed parameters for option -a: "
	       << optarg << "\n";
	  return 1;
	}
	break;

    } // switch
  } // while

  // read geometrical code file
  if (optind >= argc) {
    cerr << "Usage: " << argv[0] << usage;
    return 1;
  }
  if (strcmp (argv[optind], "-")) {
    ifstream in (argv[optind]);
    if (in.fail()) {
      cerr << argv[0] << ": unable to open code file "
	   << argv[optind] << endl;
      return 1;
    }
    code = new Code (in);
    in.close ();
    strncpy (codefilename, argv[optind], 256);
    codefilename[255] = 0;
  }
  else {
    code = new Code (cin);
    strcpy (codefilename, "<stdin>");
  }

  if (++optind < argc) {
    cerr << argv[0] << ": ignoring extra argument(s):";
    while (optind < argc)
      cerr << ' ' << argv[optind++];
    cerr << endl;
  }

  // do some consistency checks
  if (code->get_m() != GFq::q()) {
    cerr << argv[0] << ": geometrical code of size " << code->get_m()
	 << " isn't compatible with GF(" << GFq::q() << ")\n";
    return 1;
  }
  if (d > n) {
    cerr << argv[0] << ": distance of algebraical code (" << d
	 << ") must not be bigger than its length (" << n
	 << ")\n";
    return 1;
  }

  // normalize the avg. codeword energy to 1
  *code *= sqrt (1.0 / code->AvgEnergy ());
  // initialize random generator
  rg = new MLCG ((_G_int32_t) time (NULL), (_G_int32_t) time (NULL));
  noise = new Normal<MLCG> (rg);
  InitCodes ();
  PrintHeader (*out);
  PrintErrorCurve (*out);
  ClearCodes ();
  delete noise;
  delete rg;
  delete code;
}
