// Calculation of the error probability of a concatenation
// of a geometrical code, a systematic expurgated RS code
// and an optional second algebraical code
// on a simulated time-discrete AWGN channel

// (C) 1997-1998 Frank Paehlke <frank@paehlke.de>
// $Id: pe-geo-sxrs-alg.cc,v 1.1 2003/10/14 08:59:11 frank Exp $

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// $Log: pe-geo-sxrs-alg.cc,v $
// Revision 1.1  2003/10/14 08:59:11  frank
// moved from . to src/
//
// Revision 1.4  2003/10/14 08:41:58  frank
// ANSI C++ / gcc 3.x compatibility changes
//
// Revision 1.3  2002/08/02 14:45:06  paehlke
// added GPL comment
//
// Revision 1.2  1998/05/28 12:56:33  paehlke
// Final version for masters thesis
//
// Revision 1.1  1997/11/27 16:22:50  paehlke
// Initial revision
//

#include <math.h>
#include <fstream>
#include <iostream>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>

#include "mlcg.h"
#include "normal.h"
#include "code.h"
#include "sxrscode.h"
#include "gmddecoder.h"
#include "treedecoder.h"
#include "gf.h"

using namespace std;

#ifdef GF16
#include "gf16tables.h"
typedef GF<GF16Tables> GFq;
#else
#ifdef GF32
#include "gf32tables.h"
typedef GF<GF32Tables> GFq;
#else
#ifdef GF64
#include "gf64tables.h"
typedef GF<GF64Tables> GFq;
#else
#ifdef GF128
#include "gf128tables.h"
typedef GF<GF128Tables> GFq;
#else
#include "gf256tables.h"
typedef GF<GF256Tables> GFq;
#endif
#endif
#endif
#endif

typedef SXRSCode<GFq> SXRSC;
typedef GMDDecoder<GFq,SXRSC> GMDD;
typedef TreeDecoder<GFq,SXRSC> TREED;

typedef double * pDouble;

Code * code;   // geometrical code
SXRSC * code_i; // inner SXRS code
SXRSC * code_o; // outer SXRS code
GMDD * gmdd_i; // GMD decoders
GMDD * gmdd_o;
TREED * treed_i; // tree decoders
TREED * treed_o;

MLCG * rg; // random generator
Normal<MLCG> * noise;

double snr_start;    // initial SNR value
double snr_incr;     // SNR increment
int snr_num;         // number of SNR values
int snr_db;          // interpret SNR values as dB?
int loops;           // number of loops per SNR value
int d_i, xd_i, n_i, k_i; // parameters of inner SXRS code
int d_o, xd_o, n_o, k_o; // parameters of outer SXRS code
int intersize;       // interleaver size
int n;               // n = intersize * n_i
int k;               // k = intersize * k_i = step_o * n_o
int k1;              // k1 = step_o * k_o
int feedback;        // number of feedback iterations
int step_o;          // distance of outer code symbols in interleaver
int algo_i, algo_o;  // decoding algorithms for inner and outer SXRS code
double fano_delta_i; // parameter delta of Fano algorithm
double fano_delta_o;
int bstack_termnodes_i;
int bstack_termnodes_o;
int maxnodes_i;      // max # of nodes for sequential decoding
int maxnodes_o;
int cl_num_i;        // # of lists for candidate list decoding
int cl_num_o;
int cl_len_i;        // list length for candidate list decoding
int cl_len_o;
int gmd_check_i;     // do GMD check after inner decoding?
char codefilename [256];

GFq * infobuf0;  // encoder information buffer
GFq * interbuf0; // encoder interleaver buffer
GFq * codebuf0;  // encoder code buffer
Vector * sent;   // sent signal vectors
Vector * rcvd;   // received signal vectors
GFq * codebuf;   // decoder code buffer
GFq * codebuf1;  // corrected by inner decoder
GFq * interbuf;  // decoder interleaver buffer
GFq * infobuf;   // decoder information buffer
pDouble * alphabuf_i; // soft decision buffers for inner code
double * betabuf_i;
int * isknown_i; // information whether interbuf[i] is already known
pDouble * alphabuf_o; // soft decision buffers for outer code
double * betabuf_o;
int * resbuf;    // buffer for inner decoding results (erasures)
GFq * codeword;  // buffer for one outer code word
int * erasures;  // erasure location vector for outer code
pDouble * alphaword; // soft decision buffers for one outer code word
double * betaword;

int eras, werr, serr; // number of erasures/word errors/symbol errors
int eras1, werr1;
double cplxty_i, cplxty_o; // total complexity
// absolute frequency distributions of ld(complexity):
int stat_i[32], stat_o[32];

// enumeration of decoding algorithms

enum AlgorithmID {algBM = 0, algFANO, algFABM, algSTACK, algSTBM,
		  algBSTACK, algBSTBM,
		  algCLIST, algPGMD, algGMD, algMVUC,
		  algSPECIAL = 99};

#ifdef VERBOSE
void PrintVec (char * title, GFq * vec, int len)
{
  int i;
  cerr << title << " = [";
  if (len > 0)
    cerr << (int) vec[0];
  for (i=1; i<len; i++)
    cerr << ',' << (int) vec[i];
  cerr << "]\n";
}
#endif

void Transmission (const Vector & sent, Vector & rcvd)
{
  int i;
  for (i=0; i<code->get_n(); i++) {
    rcvd[i] = sent[i] + (*noise)();
  }
}

void InitCodes (const char * argv0)
{
  int i;
  sent = new Vector [intersize * n_i];
  rcvd = new Vector [intersize * n_i];
  for (i = 0; i < intersize * n_i; i++)
    rcvd[i].allocate (code->get_n());
  code_i = new SXRSC (d_i, xd_i, n_i);
  k_i = code_i->get_k();
  gmdd_i = new GMDD (code_i);
  if ((algo_i >= algFANO) && (algo_i <= algBSTBM)) {
    treed_i = new TREED (*code_i);
    if ((algo_i >= algSTACK) && (algo_i <= algBSTBM))
      treed_i->AllocateStack (maxnodes_i);
  }
  // allocate buffer memory
  codebuf0 = new GFq [intersize * n_i];
  codebuf = new GFq [intersize * n_i];
  codebuf1 = new GFq [intersize * n_i];
  interbuf0 = new GFq [intersize * k_i];
  interbuf = new GFq [intersize * k_i];
  alphabuf_i = new pDouble [intersize * n_i];
  for (i = 0; i < intersize * n_i; i++)
    alphabuf_i[i] = new double [GFq::q()];
  betabuf_i = new double [intersize * n_i];
  resbuf = new int [intersize * k_i];
  if (intersize > 1) {
    // initialize outer code
    if ((intersize * k_i) % n_o != 0) {
      cerr << argv0 << ": total interleaver size (" << intersize * k_i
	   << ") must be a multiple\n\tof outer code length ("
	   << n_o << ")\n";
      exit (1);
    }
    else
      step_o = (intersize * k_i) / n_o;
    code_o = new SXRSC (d_o, xd_o, n_o);
    k_o = code_o->get_k();
    gmdd_o = new GMDD (code_o);
    if ((algo_o >= algFANO) && (algo_o <= algBSTBM)) {
      treed_o = new TREED (*code_o);
      if ((algo_o >= algSTACK) || (algo_o <= algBSTBM))
	treed_o->AllocateStack (maxnodes_o);
    }
    // allocate additional buffer memory
    infobuf0 = new GFq [step_o * k_o];
    infobuf = new GFq [step_o * k_o];
    alphabuf_o = new pDouble [step_o * n_o];
    for (i = 0; i < step_o * n_o; i++)
      alphabuf_o[i] = new double [GFq::q()];
    betabuf_o = new double [step_o * n_o];
    if (feedback > 0)
      isknown_i = new int [intersize * k_i];
    codeword = new GFq [n_o];
    erasures = new int [n_o];
    alphaword = new pDouble [n_o];
    for (i = 0; i < n_o; i++)
      alphaword[i] = new double [GFq::q()];
    betaword = new double [n_o];
  }
}

void ClearCodes ()
{
  int i;
  if (intersize > 1) {
    delete [] betaword;
    for (i = 0; i < n_o; i++)
      delete [] alphaword[i];
    delete [] alphaword;
    delete [] erasures;
    delete [] codeword;
    if (feedback > 0)
      delete [] isknown_i;
    delete [] betabuf_o;
    for (i = 0; i < step_o * n_o; i++)
      delete [] alphabuf_o[i];
    delete [] alphabuf_o;
    delete [] infobuf;
    delete [] infobuf0;
    if ((algo_o >= algFANO) && (algo_o <= algBSTBM)) {
      if ((algo_o >= algSTACK) && (algo_o <= algBSTBM))
	treed_o->DeleteStack ();
      delete treed_o;
    }
    delete gmdd_o;
    delete code_o;
  }
  delete [] resbuf;
  delete [] betabuf_i;
  for (i = 0; i < intersize * n_i; i++)
    delete [] alphabuf_i[i];
  delete [] alphabuf_i;
  delete [] interbuf;
  delete [] interbuf0;
  delete [] codebuf1;
  delete [] codebuf;
  delete [] codebuf0;
  if ((algo_i >= algFANO) && (algo_i <= algBSTBM)) {
    if ((algo_i >= algSTACK) && (algo_i <= algBSTBM))
      treed_i->DeleteStack ();
    delete treed_i;
  }
  delete gmdd_i;
  delete code_i;
  for (i = 0; i < intersize * n_i; i++)
    rcvd[i].deallocate();
  delete [] rcvd;
  delete [] sent;
}

void PrintHeader (ostream & out)
{
  int i;
  double R = code->get_r() * code_i->get_r();
  if (intersize > 1)
    R *= code_o->get_r();
  out << "# Error probabilities for code concatenation with R = "
      << R << endl
      << "# " << loops << " loops per SNR value\n"
      << "# geometrical code: " << codefilename
      << ": m = " << code->get_m() << ", n = " << code->get_n()
      << ", r = " << code->get_r()
      << ", dmin = " << code->MinDist() << endl
      << "# inner SXRS code: k = " << k_i
      << ", n = " << n_i
      << ", r = " << code_i->get_r()
      << ", xd = " << xd_i
      << ", dmin = " << d_i << endl;
  switch (algo_i) {
  case algFANO:
    out << "# Fano decoding (delta = " << fano_delta_i
	<< ", maxnodes = " << maxnodes_i << ")\n";
    break;
  case algFABM:
    out << "# Fano/BM decoding (delta = " << fano_delta_i
	<< ", maxnodes = " << maxnodes_i << ")\n";
    break;
  case algSTACK:
    out << "# Stack decoding (maxnodes = " << maxnodes_i << ")\n";
    break;
  case algSTBM:
    out << "# Stack/BM decoding (maxnodes = " << maxnodes_i << ")\n";
    break;
  case algBSTACK:
    out << "# Bidirectional Stack decoding (maxnodes = "
	<< maxnodes_i << ", maxtermnodes = " << bstack_termnodes_i << ")\n";
    break;
  case algBSTBM:
    out << "# Bidirectional Stack/BM decoding (maxnodes = "
	<< maxnodes_i << ", maxtermnodes = " << bstack_termnodes_i << ")\n";
    break;
  case algCLIST:
    out << "# candidate list decoding (" << cl_num_i
	<< " lists of length " << cl_len_i << ")\n";
    break;
  case algPGMD:
    out << "# primitive GMD decoding\n";
    break;
  case algGMD:
    out << "# GMD decoding\n";
  }
  if (gmd_check_i)
    out << "# GMD check\n";
  if (intersize > 1) {
    out << "# outer SXRS code: k = " << k_o
	<< ", n = " << n_o
	<< ", r = " << code_o->get_r()
	<< ", xd = " << xd_o
	<< ", dmin = " << d_o << endl;
    switch (algo_o) {
    case algFANO:
      out << "# Fano decoding (delta = " << fano_delta_o
	  << ", maxnodes = " << maxnodes_o << ")\n";
      break;
    case algFABM:
      out << "# Fano/BM decoding (delta = " << fano_delta_o
	  << ", maxnodes = " << maxnodes_o << ")\n";
      break;
    case algSTACK:
      out << "# Stack decoding (maxnodes = " << maxnodes_o << ")\n";
      break;
    case algSTBM:
      out << "# Stack/BM decoding (maxnodes = " << maxnodes_o << ")\n";
      break;
    case algBSTACK:
      out << "# Bidirectional Stack decoding (maxnodes = "
	  << maxnodes_o << ", maxtermnodes = " << bstack_termnodes_o << ")\n";
      break;
    case algBSTBM:
      out << "# Bidirectional Stack/BM decoding (maxnodes = "
	  << maxnodes_o << ", maxtermnodes = " << bstack_termnodes_o << ")\n";
      break;
    case algCLIST:
      out << "# candidate list decoding (" << cl_num_o
	  << " lists of length " << cl_len_o << ")\n";
      break;
    case algPGMD:
      out << "# primitive GMD decoding\n";
      break;
    case algGMD:
      out << "# GMD decoding\n";
    }
    out << "# total interleaver size: " << intersize * k_i << " symbols\n";
    if (feedback > 0)
      out << "# " << feedback << " feedback iteration(s)\n";
  }
  out << "# SNR\t SNR(dB)\t Eb/N0(dB)\t E+E\t ERAS\t WER\t SER\t BER";
  if ((algo_i >= algFANO) && (algo_i <= algGMD))
    out << "\t cplxty";
  if (intersize > 1) {
    out << "\t (E+E)'\t WER'";
    if ((algo_o >= algFANO) && (algo_o <= algGMD))
      out << "\t cplxty";
  }
  if ((algo_i >= algFANO) && (algo_i <= algGMD)) {
    int m = (maxnodes_i+1) / k_i;
    for (i = 0; (1<<i) <= m; i++)
      out << "\t ldi=" << i;
  }
  if ((intersize > 1) && (algo_o >= algFANO) && (algo_o <= algGMD)) {
    int m = (maxnodes_o+1) / k_o;
    for (i = 0; (1<<i) <= m; i++)
      out << "\t ldo=" << i;
  }
  out << endl;
}

void PrintOutputLine (ostream & out, double snr, double snrdb, double ebn0db)
{
  int i;
  int symbols; // number of transmitted symbols
  if (intersize > 1)
    symbols = loops * step_o * k_o;
  else
    symbols = loops * k_i;
  out << snr << '\t' << snrdb << '\t' << ebn0db << '\t'
      << (double)(eras+werr) / loops << '\t'
      << (double)eras / loops << '\t'
      << (double)werr / loops << '\t'
      << (double)serr / symbols << '\t'
      << 0.5 * serr / symbols;
  if (algo_i > 0)
    out << '\t' << cplxty_i / loops / intersize << '\t';
  if (intersize > 1) {
    out << '\t' << (double)(eras1+werr1) / loops / step_o
	<< '\t' << (double)werr1 / loops / step_o;
    if (algo_o > 0)
      out << '\t' << (double)cplxty_o / loops / step_o;
  }
  if (algo_i > 0) {
    int m = (maxnodes_i+1) / k_i;
    out << '\t' << stat_i[0];
    for (i=1; (1<<i) <= m; i++) {
      assert (i < 32);
      out << ' ' << stat_i[i];
    }
  }
  if (intersize > 1 && algo_o >  0) {
    int m = (maxnodes_o+1) / k_o;
    out << '\t' << stat_o[0];
    for (i=0; (1<<i) <= m; i++) {
      assert (i < 32);
      out << ' ' << stat_o[i];
    }
  }
  out << endl;
}

// DecodeGeometricalCode(pos) decodes the received signal vector src.
// The received symbol is written to dest[]
// and any soft-decision information is written to alpha[].

void DecodeGeometricalCode (const Vector & src, GFq & dest,
			    pDouble alpha, double & beta)
{
  int destsym;
  switch (algo_i) {
  case algBM:
    code->Decode (src, destsym);
    break;
  case algFANO:
  case algFABM:
  case algSTACK:
  case algSTBM:
  case algBSTACK:
  case algBSTBM:
  case algCLIST:
  case algPGMD:
  case algGMD:
    beta = code->LLHDecode (noise->variance(), src, destsym, alpha);
    break;
  default:
    cerr << "Undefined algorithm ID " << algo_i << endl;
    exit (1);
  }
  dest = GFq(destsym);
}

// PrepareSDOuterCode prepares the soft decision information
// for decoding the outer algebraical code

void PrepareSDOuterCode (const GFq * cword, int res,
			 const double * const * alpha,
			 const double * beta,
			 pDouble const * alpha_o,
			 double * beta_o)
{
  assert ((alpha_o != NULL) && (beta_o != NULL));

  int i1,i2,i3,j;
  i1 = k_i - 1;
  i2 = n_i - 1;
  i3 = (n_i - d_i + 1) % xd_i;
  if (res < 0) {
    while (i1 >= 0) {
      assert (i2 >= d_i - 1);
      if (i3 > 0) {
	const double * ain = alpha[i2];
	double * aout = alpha_o[i1];
	for (j = 0; j < GFq::q(); j++)
	  aout[j] = ain[j];
	beta_o[i1--] = beta[i2--];
	i3--;
      }
      else {
	i2--;
	i3 = xd_i - 1;
      }
    }
  }
  else { // res >= 0
    // calculate the APP of cword[]
    double lp = alpha[0][cword[0]]; // lp = log(APP)
    for (j = 1; j < n_i; j++)
      lp += alpha[j][cword[j]];
    lp = lp * M_LN2 / n_i;
    double p = exp (lp);
    double p1 = - expm1 (lp) / (GFq::q()-1); // p1 = (1 - APP) / (q-1)
    double b; // new value for beta_o[i1]
    if (p1 >= p) {
      p = p1 = 1.0 / GFq::q();
      b = 0.0;
    }
    else
      b = lp - log(p1) / M_LN2;
    while (i1 >= 0) {
      assert (i2 >= d_i - 1);
      if (i3 > 0) {
	double * aout = alpha_o[i1];
	for (j = 0; j < GFq::q(); j++)
	  aout[j] = p1;
	aout[cword[i2]] = p;
	beta_o[i1--] = b;
	i2--;
	i3--;
      }
      else { // i3 <= 0
	i2--;
	i3 = xd_i - 1;
      }
    } // while (i2)
  } // if (res < 0)
}

int DecodeInnerCode (GFq * info, GFq * dest, const GFq * src,
		     const double * const * alpha,
		     const double * beta)
{
  int res;
  switch (algo_i) {
  case algBM:
    res = code_i->Decode (info, src);
    break;
  case algFANO:
    res = treed_i->FanoDecode (info, src, alpha, fano_delta_i, maxnodes_i,
			       0, isknown_i);
    break;
  case algFABM:
    res = treed_i->FanoDecode (info, src, alpha, fano_delta_i, maxnodes_i,
			       1, isknown_i);
    if (res < 0)
      res = code_i->Decode (info, src);
    break;
  case algSTACK:
    res = treed_i->StackDecode (info, src, alpha, 0, isknown_i);
    break;
  case algSTBM:
    res = treed_i->StackDecode (info, src, alpha, 1, isknown_i);
    if (res < 0)
      res = code_i->Decode (info, src);
    break;
  case algBSTACK:
    res = treed_i->BidirStackDecode (info, src, alpha, bstack_termnodes_i,
				     isknown_i);
    break;
  case algBSTBM:
    res = treed_i->BidirStackDecode (info, src, alpha, bstack_termnodes_i,
				     isknown_i);
    if (res < 0)
      res = code_i->Decode (info, src);
    break;
  case algCLIST:
    res = gmdd_i->CandidateListDecode (info, src, alpha, beta,
				       cl_num_i, cl_len_i);
    break;
  case algPGMD:
    res = gmdd_i->PrimitiveDecode (info, src, beta);
    break;
  case algGMD:
    res = gmdd_i->Decode (info, src, alpha, beta);
    break;
  case algMVUC:
  case algSPECIAL:
    cerr << "Decoding algorithm " << algo_i << " is not yet implemented "
      "for inner code\n";
  default: exit(1);
  }
  code_i->Encode (dest, info);
  if (gmd_check_i && (res >= 0)) {
    if (gmdd_i->Reliability (dest, src, alpha, beta) < -4*d_i)
  res = -20;
 }
  if (algo_i >= algFANO && algo_i <= algGMD) {
    // update complexity statistics
    double cplxty;
    if (algo_i >= algFANO && algo_i <= algBSTBM)
      cplxty = treed_i->GetNodes() / k_i; // # of nodes per info symbol
    else
      cplxty = gmdd_i->GetLoops();
    cplxty_i += cplxty;
    int i = 1;
    while ((1<<i) <= cplxty)
      i++;
    // now i-1 == ld(cplxty)
    assert (i > 0 && i <= 32 &&
	    (1<<(i-1)) <= (maxnodes_i+1) / k_i);
    stat_i[i-1]++;
  }
  return res;
}

int DecodeOuterCode (GFq * dest, const GFq * src,
		     const int * erasures, int erasnum,
		     const double * const * alpha,
		     const double * beta)
{
  int res;
  switch (algo_o) {
  case algBM:
    if (erasnum >= d_o) // too many erasures
      res = -20;
    else
      res = code_o->Decode (dest, src, erasures, erasnum);
    break;
  case algFANO:
    res = treed_o->FanoDecode (dest, src, alpha, fano_delta_o, maxnodes_o);
    break;
  case algFABM:
    res = treed_o->FanoDecode (dest, src, alpha, fano_delta_o, maxnodes_o); //
    if (res < 0)
      res = code_o->Decode (dest, src);
    break;
  case algSTACK:
    res = treed_o->StackDecode (dest, src, alpha);
    break;
  case algSTBM:
    res = treed_o->StackDecode (dest, src, alpha);
    if (res < 0)
      res = code_o->Decode (dest, src);
    break;
  case algBSTACK:
    res = treed_o->BidirStackDecode (dest, src, alpha, bstack_termnodes_o);
    break;
  case algBSTBM:
    res = treed_o->BidirStackDecode (dest, src, alpha, bstack_termnodes_o);
    if (res < 0)
      res = code_o->Decode (dest, src);
    break;
  case algCLIST:
    res = gmdd_o->CandidateListDecode (dest, src, alpha, beta,
				       cl_num_o, cl_len_o);
    break;
  case algPGMD:
    res = gmdd_o->PrimitiveDecode (dest, src, beta);
    break;
  case algGMD:
    res = gmdd_o->Decode (dest, src, alpha, beta);
    break;
  case algMVUC:
  case algSPECIAL:
    cerr << "Decoding algorithm " << algo_o << " is not yet implemented "
      "for outer code\n";
  default: exit(1);
  }
  if (algo_o >= algFANO && algo_o <= algGMD) {
    // update complexity statistics
    double cplxty;
    if (algo_o > algFANO && algo_o <= algBSTBM)
      cplxty = treed_o->GetNodes() / k_o; // # of nodes per info symbol
    else
      cplxty = gmdd_o->GetLoops();
    cplxty_o += cplxty;
    int i = 1;
    while ((1<<i) <= cplxty)
      i++;
    // now i-1 == ld(cplxty)
    assert (i > 0 && i <= 32 &&
	    (1<<(i-1)) <= (maxnodes_o+1) / k_o);
    stat_o[i-1]++;
  }
  return res;
}

void SimulateSingleCode ()
{
  int i,res;
  // fill info buffer with random symbols
  for (i = 0; i < k_i; i++)
    interbuf0[i] = (int) floor (GFq::q()*rg->asDouble());
  // apply SXRS code
  code_i->Encode (codebuf0, interbuf0);
  for (i = 0; i < n_i; i++) {
    // transmit i-th code symbol
    sent[i].copy ((*code)[codebuf0[i]]);
    Transmission (sent[i], rcvd[i]);
    DecodeGeometricalCode (rcvd[i], codebuf[i],
			   alphabuf_i[i], betabuf_i[i]);
  }
  res = DecodeInnerCode (interbuf, codebuf1, codebuf, alphabuf_i, betabuf_i);
  // update error statistics
  if (res < 0) {
    eras++;
#if 0
    int j;
    double absval = 0.0;
    cerr << "#Signal:\\\\[1ex]\n";
    for (i = 0; i < n_i; i++)
      for (j = 0; j < code->get_n(); j++)
	cerr << sent[i][j] << '\t';
    cerr << "\n#Rauschen:\\\\[1ex]\n";
    for (i = 0; i < n_i; i++)
      for (j = 0; j < code->get_n(); j++) {
	cerr << rcvd[i][j]-sent[i][j] << '\t';
	absval += (rcvd[i][j]-sent[i][j]) * (rcvd[i][j]-sent[i][j]);
      }
    cerr << "\n#Summe:\\\\[1ex]\n";
    for (i = 0; i < n_i; i++)
      for (j = 0; j < code->get_n(); j++)
	cerr << rcvd[i][j] << '\t';
    cerr << "\n#SNR: " << n_i/absval << "\n#\n#\\hrulefill\n#\n";
#endif
#if 0
    int j;
    double absval,avsum=0.0;
    cerr << "#Quadr. Distanz der Sende-/Empfangsvektoren:\\\\[1ex]\n";
    for (i = 0; i < n_i; i++) {
      absval = 0.0;
      for (j = 0; j < code->get_n(); j++)
	absval += (rcvd[i][j]-sent[i][j]) * (rcvd[i][j]-sent[i][j]);
      cerr << absval << '\t';
      avsum += absval;
    }
    cerr << "\n#SNR: " << n_i/avsum << "\n#\n#\\hrulefill\n#\n";
#endif
#if 0
    int j,num=0;
    for (i = 0; i < n_i; i++) {
      if (codebuf[i] != codebuf0[i])
	num++,cerr << '*';
      else
	cerr << '-';
    }
    cerr << ' ' << num << endl;
#endif
  }
  else {
    int serr0 = serr;
    for (i = 0; i < k_i; i++)
      if (interbuf[i] != interbuf0[i])
	serr++;
    if (serr > serr0) {
#ifdef VERBOSE
      cerr << "Error: GMD = "
	   << gmdd_i->Reliability (codebuf1, codebuf,
				   (const double * const *) alphabuf_i,
				   betabuf_i) << endl;
#endif
      werr++;
    }
  }
}

void SimulateConcatenation ()
{
  int i, j, feed, res, res1, _eras, _werr, _serr;
  // fill info buffer with random symbols
  for (i = 0; i < step_o*k_o; i++)
    infobuf0[i] = (int) floor (GFq::q()*rg->asDouble());
#ifdef VERBOSE
  cerr << "### infobuf0:\n";
  for (i=0;i<step_o;i++) {
    for (j=0;j<k_o;j++)
      cerr << '\t' << (int) infobuf0[i*k_o+j];
    cerr << endl;
  }
#endif
  // apply outer SXRS code
  for (i = 0; i < step_o; i++) {
    code_o->Encode (codeword, &infobuf0[i*k_o]);
    // interleave
    for (j = 0; j < n_o; j++)
      interbuf0[j*step_o + i] = codeword[j];
  }
#ifdef VERBOSE
  cerr << "### interbuf0:\n";
  for (i=0;i<intersize;i++) {
    for (j=0;j<k_i;j++)
      cerr << '\t' << (int) interbuf0[i*k_i+j];
    cerr << endl;
  }
#endif
  // apply inner SXRS code
  for (i = 0; i < intersize; i++)
    code_i->Encode (&codebuf0[i*n_i], &interbuf0[i*k_i]);
#ifdef VERBOSE
  cerr << "### codebuf0:\n";
  for (i=0;i<intersize;i++) {
    for (j=0;j<n_i;j++)
      cerr << '\t' << (int) codebuf0[i*n_i+j];
    cerr << endl;
  }
#endif
  for (i = 0; i < intersize*n_i; i++) {
    // transmit i-th code symbol
    sent[i].copy ((*code)[codebuf0[i]]);
    Transmission (sent[i], rcvd[i]);
    DecodeGeometricalCode (rcvd[i], codebuf[i],
			   alphabuf_i[i], betabuf_i[i]);
  }
#ifdef VERBOSE
  cerr << "### codebuf:\n";
  for (i=0;i<intersize;i++) {
    for (j=0;j<n_i;j++) {
      cerr << "\t " << (int) codebuf[i*n_i+j];
      if (codebuf0[i*n_i+j]!=codebuf[i*n_i+j]) cerr << '*';
    }
    cerr << endl;
    for (j=0;j<n_i;j++)
      (cerr << '\t').form ("%.2lf", betabuf_i[i*n_i+j]);
    cerr << endl;
  }
#endif
  feed = feedback;
  if (feedback > 0)
    // no codesymbols are known at the beginning
    for (i = 0; i < intersize*k_i; i++)
      isknown_i[i] = 0;
  do {
    for (i = 0; i < intersize; i++) {
      res = DecodeInnerCode (&interbuf[i*k_i], &codebuf1[i*n_i],
			     &codebuf[i*n_i],
			     &alphabuf_i[i*n_i], &betabuf_i[i*n_i]);
      PrepareSDOuterCode (&codebuf[i*n_i], res,
			  &alphabuf_i[i*n_i], &betabuf_i[i*n_i],
			  &alphabuf_o[i*k_i], &betabuf_o[i*k_i]);
      for (j = 0; j < k_i; j++)
	resbuf[i*k_i+j] = res;
    }
#ifdef VERBOSE
  cerr << "### interbuf:\n";
  for (i=0;i<intersize;i++) {
    for (j=0;j<k_i;j++) {
      cerr << "\t " << (int) interbuf[i*k_i+j];
      if (resbuf[i*k_i+j] < 0) cerr << '-';
      if (interbuf0[i*k_i+j]!=interbuf[i*k_i+j]) cerr << '*';
    }
    cerr << endl;
    for (j=0;j<k_i;j++)
      (cerr << '\t').form ("%.2lf", betabuf_o[i*k_i+j]);
    cerr << endl;
  }
#endif
    _eras = _werr = _serr = 0;
    // decode outer SXRS code
    assert (algo_o == 0); // sequential decoding not yet implemented
    res = 0;
    for (i = 0; i < step_o; i++) {
      // deinterleave and determine erasures
      int erasnum = 0;
      for (j = 0; j < n_o; j++) {
	codeword[j] = interbuf[j*step_o + i];
	if (resbuf[j*step_o + i] < 0)
	  erasures[erasnum++] = j;
	alphaword[j] = alphabuf_o[j*step_o + i];
	betaword[j] = betabuf_o[j*step_o + i];
      }
      res1 = DecodeOuterCode (&infobuf[i*k_o], codeword, erasures, erasnum,
			      alphaword, betaword);
#if 0
      if (erasnum >= d_o) // too many erasures
	res1 = -20;
      else {
	res1 = code_o->Correct (codeword, codeword, erasures, erasnum);
	code_o->ExtractInformationWord (&infobuf[i*k_o], codeword);
      }
#endif
      if (res1 < 0) {
	res = res1;
	_eras++;
	if (feed > 0) {
	  // feedback of decoding results
	  for (j = 0; j < n_o; j++)
	    isknown_i [j*step_o+i] = 0;
	}
      }
      else { // res1 >= 0
	// update error statistics
	int _serr0 = _serr;
	for (j = 0; j < k_o; j++)
	  if (infobuf[i*k_o + j] != infobuf0[i*k_o + j])
	    _serr++;
	if (_serr > _serr0)
	  _werr++;
	if (feed > 0) {
	  // feedback of decoding results
	  for (j = 0; j < n_o; j++) {
	    isknown_i [j*step_o+i] = 1;
	    interbuf [j*step_o+i] = codeword[j];
	  }
	}
      } // if (res1 < 0)
    } // for (i)
  } while (--feed >= 0);
  // update error statistics
  eras1 += _eras;
  werr1 += _werr;
  if (_eras > 0)
    eras++;
  else if (_werr > 0)
    werr++;
  serr += _serr;
}

void PrintErrorCurve (ostream & out)
{
  int s,i;
  // compute code rate
  double R = code->get_r() * code_i->get_r();
  if (intersize > 1)
    R *= code_o->get_r();
  for (s=0; s<snr_num; s++) {
    // set SNR and Eb/N0
    double snr, snrdb, ebn0, ebn0db;
    if (snr_db) {
      snrdb = snr_start + s * snr_incr;
      snr = pow (10.0, snrdb/10.0);
    }
    else {
      snr = snr_start + s * snr_incr;
      snrdb = 10.0 * log10 (snr);
    }
    ebn0 = snr / (2*R);
    ebn0db = 10.0 * log10 (ebn0);
    noise->variance (1.0/(snr * code->get_n()));
    // init counters and statistics
    eras = werr = serr = eras1 = werr1 = 0;
    cplxty_i = cplxty_o = 0.0;
    for (i=0; i<32; i++)
      stat_i[i] = stat_o[i] = 0;
    // now do the actual simulation
    for (i = 0; i < loops; i++) {
      if (intersize > 1)
	SimulateConcatenation ();
      else
	SimulateSingleCode ();
    }
    PrintOutputLine (out, snr, snrdb, ebn0db);
  }
}

void SetDefaults ()
{
  snr_start = 1.0;
  snr_incr = 0.1;
  snr_num = 1;
  snr_db = 0; // SNR values are not interpreted as dB
  loops = 1000;
  n_i = n_o = GFq::q() - 1;
  d_i = d_o = 5; // correct up to 2 errors
  xd_i = xd_o = 4;
  intersize = 1; // no interleaving
  feedback = 0; // no feedback
  algo_i = algo_o = algBM;
  fano_delta_i = fano_delta_o = 4.0;
  bstack_termnodes_i = bstack_termnodes_o = 1;
  maxnodes_i = maxnodes_o = 1000;
  cl_num_i = cl_num_o = 8;
  cl_len_i = cl_len_o = 2;
  gmd_check_i = 0;
}

char * usage = // "Usage: " << argv[0] <<
" [OPTION...] CODEFILE\n\n"
"-h   display this help and exit\n"
"-o STRING  set output filename\n"
"-s FLOAT   set initial SNR\n"
"-i FLOAT   set SNR increment\n"
"-n INT     set # of SNR values\n"
"-l INT     set # of loops per SNR value\n"
"-d         interpret SNR values as dB\n"
"-c INT[,INT[,INT]]  set parameters (dmin,xd,n) of inner code\n"
"-C INT[,INT[,INT]]  set parameters (dmin,xd,n) of outer code\n"
"-I INT     set interleaver size (1 = no outer code)\n"
"-f INT     use decoding feedback (parameter: # of iterations)\n"
"-a INT     set decoding algorithm for inner code:\n"
"   -a 0       Berlekamp-Massey (default)\n"
"   -a 1[,INT[,FLOAT]]  Fano (parameter: delta,maxnodes)\n"
"   -a 2[,INT[,FLOAT]]  Fano with BM in case of erasures\n"
"   -a 3[,INT]          Stack (parameter: maxnodes)\n"
"   -a 4[,INT]          Stack with BM in case of erasures\n"
"   -a 5[,INT,INT]      BD-Stack (parameter: maxnodes,maxtermnodes)\n"
"   -a 6[,INT.INT]      BD-Stack with BM in case of erasures\n"
"   -a 7[,INT[,INT]]    candidate lists (parameters: number,length)\n"
"   -a 8       primitive GMD\n"
"   -a 9       normal GMD\n"
"   -a 10      modified Vucetic\n"
"   -a 99      special\n"
"-g         apply GMD check after decoding inner code\n"
"-A INT     set decoding algorithm for outer code (see -a)\n";

int main (int argc, char ** argv)
{
  int opt;
  int i1,i2,i3;
  char dummy;
  double d1;
  ostream * out = &cout;

  SetDefaults ();
  while ((opt = getopt (argc, argv, "ho:s:i:n:l:dc:C:I:f:a:gA:")) != EOF) {
    switch (opt) {

    case 'h':
      cout << "Usage: " << argv[0] << usage;
      return 0;

    case 'o':
      out = new ofstream (optarg);
      if (out->fail ()) {
	cerr << argv[0] << ": error opening output file " << optarg << "\n";
	return 1;
      }
      break;

    case 's':
      if (sscanf (optarg, "%lf%c", &d1, &dummy) != 1) {
	cerr << argv[0] << ": malformed parameters for option -s: "
	     << optarg << endl;
	return 1;
      }
      snr_start = d1;
      break;

    case 'i':
      if (sscanf (optarg, "%lf%c", &d1, &dummy) != 1) {
	cerr << argv[0] << ": malformed parameters for option -i: "
	     << optarg << endl;
	return 1;
      }
      snr_incr = d1;
      break;

    case 'n':
      if (sscanf (optarg, "%d%c", &i1, &dummy) != 1) {
	cerr << argv[0] << ": malformed parameters for option -n: "
	     << optarg << endl;
	return 1;
      }
      if (i1 < 1) {
	cerr << argv[0] << ": option -n: number of SNR values must be positive"
	  " (was " << i1 << ")\n";
	return 1;
      }
      snr_num = i1;
      break;

    case 'l':
      if (sscanf (optarg, "%d%c", &i1, &dummy) != 1) {
	cerr << argv[0] << ": malformed parameters for option -l: "
	     << optarg << endl;
	return 1;
      }
      if (i1 < 1) {
	cerr << argv[0] << ": option -l: number of loops must be positive"
	  " (was " << i1 << ")\n";
	return 1;
      }
      loops = i1;
      break;

    case 'd':
      snr_db = 1; // interpret SNR values as dB
      break;

    case 'c':
      // parameters of inner code
      switch (sscanf (optarg, "%d,%d,%d%c", &i1, &i2, &i3, &dummy)) {
      case 3:
	if (i3 < 1 || i3 >= GFq::q()) {
	  cerr << argv[0] <<
	    ": option -c: length must be between 1 and " << GFq::q()-1 <<
	    " (was " << i3 << ")\n";
	  return 1;
	}
	n_i = i3;
      case 2:
	if (i2 <= 1) {
	  cerr << argv[0] <<
	    ": option -c: expurgation distance must be bigger than 1"
	    " (was " << i2 << ")\n";
	  return 1;
	}
	xd_i = i2;
      case 1:
	if (i1 < 1) {
	  cerr << argv[0] <<
	    ": option -c: code distance must be positive"
	    " (was " << i1 << ")\n";
	  return 1;
	}
	d_i = i1;
	break;
      default:
	cerr << argv[0] << ": malformed parameters for option -c: "
	     << optarg << "\n";
	return 1;
      }
      break;

    case 'C':
      // parameters of outer code
      switch (sscanf (optarg, "%d,%d,%d%c", &i1, &i2, &i3, &dummy)) {
      case 3:
	if (i3 < 1 || i3 >= GFq::q()) {
	  cerr << argv[0] <<
	    ": option -C: length must be between 1 and " << GFq::q()-1 <<
	    " (was " << i3 << ")\n";
	  return 1;
	}
	n_o = i3;
      case 2:
	if (i2 <= 1) {
	  cerr << argv[0] <<
	    ": option -C: expurgation distance must be bigger than 1"
	    " (was " << i2 << ")\n";
	  return 1;
	}
	xd_o = i2;
      case 1:
	if (i1 < 1) {
	  cerr << argv[0] <<
	    ": option -C: code distance must be positive"
	    " (was " << i1 << ")\n";
	  return 1;
	}
	d_o = i1;
	break;
      default:
	cerr << argv[0] << ": malformed parameters for option -C: "
	     << optarg << "\n";
	return 1;
      }
      break;

    case 'I':
      if (sscanf (optarg, "%d%c", &i1, &dummy) != 1) {
	cerr << argv[0] << ": malformed parameters for option -I: "
	     << optarg << endl;
	return 1;
      }
      if (i1 < 1) {
	cerr << argv[0] << ": option -I: interleaver size must be positive"
	  " (was " << i1 << ")\n";
	return 1;
      }
      intersize = i1;
      break;

    case 'f':
      // number of feedback iterations
      if (sscanf (optarg, "%d%c", &i1, &dummy) != 1) {
	cerr << argv[0] << ": malformed parameters for option -f: "
	     << optarg << endl;
	return 1;
      }
      if (i1 < 1) {
	cerr << argv[0] << ": option -f: number of iterations must be positive"
	  " (was " << i1 << ")\n";
	return 1;
      }
      feedback = i1;
      break;

    case 'a':
      // decoding algorithm for inner code
      switch (sscanf (optarg, "%d,%d,%lf%c", &i1, &i2, &d1, &dummy)) {
      case 3:
	if (i1 == algCLIST)
	  cl_len_i = (int) floor (d1 + .5);
	else if (i1 == algFANO || i1 == algFABM)
	  fano_delta_i = d1;
	else if (i1 == algBSTACK || i1 == algBSTBM)
	  bstack_termnodes_i = (int) d1;
	else
	  cerr << argv[0] << ": option -a: ignoring second parameter "
	       << d1 << endl;
      case 2:
	switch (i1) {
	case algFANO:
	case algFABM:
	case algSTACK:
	case algSTBM:
	case algBSTACK:
	case algBSTBM:
       	  if (i2 < 1) {
	    cerr << argv[0] << ": option -a: parameter maxnodes "
	      "must be positive (was " << i2 << ")\n";
	    return 1;
	  }
	  maxnodes_i = i2;
	  break;
	case algCLIST:
	  if (i2 < 1) {
	    cerr << argv[0] << ": option -a: number of lists "
	      "must be positive (was " << i2 << ")\n";
	    return 1;
	  }
	  cl_num_i = i2;
	  break;
	default:
	  cerr << argv[0] << ": option -a: ignoring first parameter "
	       << i2 << "\n";
	} // switch (i1)
      case 1:
	if ((i1 < algBM || i1 > algMVUC) && (i1 != algSPECIAL)) {
	  cerr << argv[0] <<
	    ": option -a: invalid algorithm ID " << i1 << ")\n";
	  return 1;
	}
	algo_i = i1;
	break;
      default:
	cerr << argv[0] << ": malformed parameters for option -a: "
	     << optarg << "\n";
	return 1;
      }
      break;

    case 'g':
      gmd_check_i = 1;
      break;

    case 'A':
      // decoding algorithm for outer code
      switch (sscanf (optarg, "%d,%d,%lf%c", &i1, &i2, &d1, &dummy)) {
      case 3:
	if (i1 == algCLIST)
	  cl_len_o = (int) floor (d1 + .5);
	else if (i1 == algFANO || i1 == algFABM)
	  fano_delta_o = d1;
	else if (i1 == algBSTACK || i1 == algBSTBM)
	  bstack_termnodes_o = (int) d1;
	if (i1 != algFANO && i1 != algFABM)
	  cerr << argv[0] << ": option -A: ignoring parameter delta = "
	       << d1 << "\n";
      case 2:
	switch (i1) {
	case algFANO:
	case algFABM:
	case algSTACK:
	case algSTBM:
	case algBSTACK:
	case algBSTBM:
	  if (i2 < 1) {
	    cerr << argv[0] << ": option -A: parameter maxnodes "
	      "must be positive (was " << i2 << ")\n";
	    return 1;
	  }
	  maxnodes_o = i2;
	  break;
	case algCLIST:
	  if (i2 < 1) {
	    cerr << argv[0] << ": option -A: number of lists "
	      "must be positive (was " << i2 << ")\n";
	    return 1;
	  }
	  cl_num_o = i2;
	  break;
	default:
	  cerr << argv[0] << ": option -A: ignoring first parameter "
	       << i2 << "\n";
	} // switch (i1)
      case 1:
	if ((i1 < algBM || i1 > algMVUC) && (i1 != algSPECIAL)) {
	  cerr << argv[0] <<
	    ": option -A: invalid algorithm ID " << i1 << ")\n";
	  return 1;
	}
	algo_o = i1;
	break;
      default:
	cerr << argv[0] << ": malformed parameters for option -A: "
	     << optarg << "\n";
	return 1;
      }
      break;

    } // switch
  } // while

  // read geometrical code file
  if (optind >= argc) {
    cerr << "Usage: " << argv[0] << usage;
    return 1;
  }
  if (strcmp (argv[optind], "-")) {
    ifstream in (argv[optind]);
    if (in.fail()) {
      cerr << argv[0] << ": unable to open code file "
	   << argv[optind] << endl;
      return 1;
    }
    code = new Code (in);
    in.close ();
    strncpy (codefilename, argv[optind], 256);
    codefilename[255] = 0;
  }
  else {
    code = new Code (cin);
    strcpy (codefilename, "<stdin>");
  }

  if (++optind < argc) {
    cerr << argv[0] << ": ignoring extra argument(s):";
    while (optind < argc)
      cerr << ' ' << argv[optind++];
    cerr << endl;
  }

  // do some consistency checks
  if (code->get_m() != GFq::q()) {
    cerr << argv[0] << ": geometrical code of size " << code->get_m()
	 << " isn't compatible with GF(" << GFq::q() << ")\n";
    return 1;
  }
  if (d_i > n_i) {
    cerr << argv[0] << ": distance of inner code (" << d_i
	 << ") must not be bigger than its length (" << n_i
	 << ")\n";
    return 1;
  }
  if (d_o > n_o) {
    cerr << argv[0] << ": distance of outer code (" << d_o
	 << ") must not be bigger than its length (" << n_o
	 << ")\n";
    return 1;
  }
  if (feedback > 0 && (intersize == 1 || algo_i == 0)) {
    cerr << argv[0] << ": feedback is only useful for concatenated codes\n"
	 << "\twith sequential decoding of inner code\n";
    return 1;
  }

  // normalize the avg. codeword energy to 1
  *code *= sqrt (1.0 / code->AvgEnergy ());
  // initialize random generator
  rg = new MLCG ((_G_int32_t) time (NULL), (_G_int32_t) time (NULL));
  noise = new Normal<MLCG> (rg);
  InitCodes (argv[0]);
  PrintHeader (*out);
  PrintErrorCurve (*out);
  ClearCodes ();
  delete noise;
  delete rg;
  delete code;
}
