// Calculation of the error probability of a concatenation
// of a geometrical code and a product code made out of
// two systematic expurgated RS codes
// on a simulated time-discrete AWGN channel

// (C) 1997-1998 Frank Paehlke <frank@paehlke.de>
// $Id: pe-geo-sxrs-prod.cc,v 1.1 2003/10/14 08:59:11 frank Exp $

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// $Log: pe-geo-sxrs-prod.cc,v $
// Revision 1.1  2003/10/14 08:59:11  frank
// moved from . to src/
//
// Revision 1.4  2003/10/14 08:41:58  frank
// ANSI C++ / gcc 3.x compatibility changes
//
// Revision 1.3  2002/08/02 14:45:06  paehlke
// added GPL comment
//
// Revision 1.2  1998/05/28 12:56:35  paehlke
// Final version for masters thesis
//
// Revision 1.1  1998/01/21 13:06:49  paehlke
// Initial revision
//
// Revision 1.1  1997/11/27 16:22:50  paehlke
// Initial revision
//

#include <math.h>
#include <fstream>
#include <iostream>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>

#include "mlcg.h"
#include "normal.h"
#include "code.h"
#include "sxrscode.h"
#include "gmddecoder.h"
#include "treedecoder.h"
#include "gf.h"

using namespace std;

#ifdef GF16
#include "gf16tables.h"
typedef GF<GF16Tables> GFq;
#else
#ifdef GF32
#include "gf32tables.h"
typedef GF<GF32Tables> GFq;
#else
#ifdef GF64
#include "gf64tables.h"
typedef GF<GF64Tables> GFq;
#else
#ifdef GF128
#include "gf128tables.h"
typedef GF<GF128Tables> GFq;
#else
#include "gf256tables.h"
typedef GF<GF256Tables> GFq;
#endif
#endif
#endif
#endif

typedef SXRSCode<GFq> SXRSC;
typedef GMDDecoder<GFq,SXRSC> GMDD;
typedef TreeDecoder<GFq,SXRSC> TREED;

typedef double * pDouble;

Code * code;   // geometrical code
SXRSC * code_i; // horizontal (inner) SXRS code
SXRSC * code_o; // vertical (outer) SXRS code
GMDD * gmdd_i; // GMD decoders
GMDD * gmdd_o;
TREED * treed_i; // tree decoders
TREED * treed_o;

MLCG * rg; // random generator
Normal<MLCG> * noise;

double snr_start;    // initial SNR value
double snr_incr;     // SNR increment
int snr_num;         // number of SNR values
int snr_db;          // interpret SNR values as dB?
int loops;           // number of loops per SNR value
int d_i, xd_i, n_i, k_i; // parameters of horizontal (inner) SXRS code
int d_o, xd_o, n_o, k_o; // parameters of vertical (outer) SXRS code
int feedback;        // number of feedback iterations
int step_o;          // distance of outer code symbols in interleaver
int algo_i, algo_o;  // decoding algorithms for inner and outer SXRS code
double fano_delta_i; // parameter delta of Fano algorithm
double fano_delta_o;
int maxnodes_i;      // max # of nodes for sequential decoding
int maxnodes_o;
int cl_num_i;        // # of lists for candidate list decoding
int cl_num_o;
int cl_len_i;        // list length for candidate list decoding
int cl_len_o;
double gmd_thresh_i; // threshold for GMD check
double gmd_thresh_o;

char codefilename [256];

GFq * infobuf0;  // encoder information buffer
GFq * interbuf0; // encoder interleaver buffer
GFq * codebuf0;  // encoder code buffer
Vector * sent;   // sent signal vectors
Vector * rcvd;   // received signal vectors
GFq * codebuf;   // decoder code buffer
GFq * codebuf_i;  // corrected by horizontal (inner) decoder
GFq * codebuf_o;  // corrected by vertical (outer) decoder
GFq * infobuf;   // decoder information buffer
int * erasbuf;   // decoder erasure buffer
pDouble * alphabuf; // original soft decision buffers
double * betabuf;
pDouble * alphabuf_i; // soft decision buffers after inner decoding
double * betabuf_i;
pDouble * alphabuf_o; // soft decision buffers after outer decoding
double * betabuf_o;
int * resbuf_i;    // buffer for inner decoding results (erasures)
int * resbuf_o;    // buffer for outer decoding results (erasures)
double * relbuf_i; // GMD reliabilities after inner decoding
double * relbuf_o;
GFq * codeword_i;  // buffer for one inner code word
GFq * codeword0_o; // buffers for two outer code words
GFq * codeword_o;
GFq * infoword_o;  // buffer for one outer information word
int * erasures;  // erasure location vector for outer code
pDouble * alphaword; // soft decision buffers for one outer code word
double * betaword;

int eras, werr, serr; // number of erasures/word errors/symbol errors
int eras1, werr1, serr1;
int eras2, serr2;
double cplxty_i, cplxty_o; // total complexity

// enumeration of decoding algorithms

enum AlgorithmID {algBM = 0, algFANO, algFABM, algSTACK, algSTBM,
		  algBSTACK, algBSTBM,
		  algCLIST, algPGMD, algGMD, algMVUC,
		  algSPECIAL = 99};

#ifndef NDEBUG
void PrintVec (char * title, GFq * vec, int len)
{
  int i;
  cerr << title << " = [";
  if (len > 0)
    cerr << (int) vec[0];
  for (i=1; i<len; i++)
    cerr << ',' << (int) vec[i];
  cerr << "]\n";
}
#endif

void Transmission (const Vector & sent, Vector & rcvd)
{
  int i;
  for (i=0; i<code->get_n(); i++) {
    rcvd[i] = sent[i] + (*noise)();
  }
}

void InitCodes ()
{
  int i;
  // initialize horizontal (inner) code
  code_i = new SXRSC (d_i, xd_i, n_i);
  k_i = code_i->get_k();
  gmdd_i = new GMDD (code_i);
  if ((algo_i >= algFANO) && (algo_i <= algBSTBM)) {
    treed_i = new TREED (*code_i);
    if ((algo_i >= algSTACK) && (algo_i <= algBSTBM))
      treed_i->AllocateStack (maxnodes_i);
  }
  // initialize vertical (outer) code
  code_o = new SXRSC (d_o, xd_o, n_o);
  k_o = code_o->get_k();
  gmdd_o = new GMDD (code_o);
  if ((algo_o >= algFANO) && (algo_o <= algBSTBM)) {
    treed_o = new TREED (*code_o);
    if ((algo_o >= algSTACK) && (algo_o <= algBSTBM))
      treed_o->AllocateStack (maxnodes_o);
  }
  // allocate buffer memory for encoder
  infobuf0 = new GFq [k_i * k_o];
  interbuf0 = new GFq [k_i * n_o];
  codebuf0 = new GFq [n_i * n_o];
  sent = new Vector [n_i * n_o];
  // allocate buffer memory for decoder
  rcvd = new Vector [n_i * n_o];
  for (i = 0; i < n_i * n_o; i++)
    rcvd[i].allocate (code->get_n());
  codebuf = new GFq [n_i * n_o];
  codebuf_i = new GFq [n_i * n_o];
  codebuf_o = new GFq [n_i * n_o];
  alphabuf = new pDouble [n_i * n_o];
  for (i = 0; i < n_i * n_o; i++)
    alphabuf[i] = new double [GFq::q()];
  alphabuf_i = new pDouble [n_i * n_o];
  for (i = 0; i < n_i * n_o; i++)
    alphabuf_i[i] = new double [GFq::q()];
  alphabuf_o = new pDouble [n_i * n_o];
  for (i = 0; i < n_i * n_o; i++)
    alphabuf_o[i] = new double [GFq::q()];
  betabuf = new double [n_i * n_o];
  betabuf_i = new double [n_i * n_o];
  betabuf_o = new double [n_i * n_o];
  resbuf_i = new int [n_i * n_o];
  resbuf_o = new int [n_i * n_o];
  relbuf_i = new double [n_i * n_o];
  relbuf_o = new double [n_i * n_o];
  infobuf = new GFq [k_i * k_o];
  erasbuf = new int [k_i * k_o];
  // allocate temporary storage
  codeword_i = new GFq [n_i];
  codeword0_o = new GFq [n_o];
  codeword_o = new GFq [n_o];
  infoword_o = new GFq [k_o];
  erasures = new int [(n_i > n_o)? n_i: n_o];
  alphaword = new pDouble [n_o];
  for (i = 0; i < n_o; i++)
    alphaword[i] = new double [GFq::q()];
  betaword = new double [n_o];
}

void ClearCodes ()
{
  int i;
  delete [] betaword;
  for (i = 0; i < n_o; i++)
    delete [] alphaword[i];
  delete [] alphaword;
  delete [] erasures;
  delete [] infoword_o;
  delete [] codeword_o;
  delete [] codeword0_o;
  delete [] codeword_i;
  delete [] erasbuf;
  delete [] infobuf;
  delete [] relbuf_o;
  delete [] relbuf_i;
  delete [] resbuf_o;
  delete [] resbuf_i;
  delete [] betabuf_o;
  delete [] betabuf_i;
  delete [] betabuf;
  for (i = 0; i < n_i*n_o; i++)
    delete [] alphabuf_o[i];
  delete [] alphabuf_o;
  for (i = 0; i < n_i*n_o; i++)
    delete [] alphabuf_i[i];
  delete [] alphabuf_i;
  for (i = 0; i < step_o * n_o; i++)
    delete [] alphabuf[i];
  delete [] alphabuf;
  delete [] codebuf_o;
  delete [] codebuf_i;
  delete [] codebuf;
  for (i = 0; i < n_i * n_o; i++)
    rcvd[i].deallocate();
  delete [] rcvd;
  delete [] sent;
  delete [] codebuf0;
  delete [] interbuf0;
  delete [] infobuf0;
  if ((algo_o >= algFANO) && (algo_o <= algBSTBM)) {
    if ((algo_o >= algSTACK) && (algo_o <= algBSTBM))
      treed_o->DeleteStack ();
    delete treed_o;
  }
  delete gmdd_o;
  delete code_o;
  if ((algo_i >= algFANO) && (algo_i <= algBSTBM)) {
    if ((algo_i >= algSTACK) && (algo_i <= algBSTBM))
      treed_i->DeleteStack ();
    delete treed_i;
  }
  delete gmdd_i;
  delete code_i;
}

void PrintHeader (ostream & out)
{
  double R = code->get_r() * code_i->get_r() * code_o->get_r();
  out << "# Error probabilities for code concatenation with R = "
      << R << endl
      << "# " << loops << " loops per SNR value\n"
      << "# geometrical code: " << codefilename
      << ": m = " << code->get_m() << ", n = " << code->get_n()
      << ", r = " << code->get_r()
      << ", dmin = " << code->MinDist() << endl
      << "# horizontal (inner) SXRS code: k = " << k_i
      << ", n = " << n_i
      << ", r = " << code_i->get_r()
      << ", xd = " << xd_i
      << ", dmin = " << d_i << endl;
  switch (algo_i) {
  case algFANO:
    out << "# Fano decoding (delta = " << fano_delta_i
	<< ", maxnodes = " << maxnodes_i << ")\n";
    break;
  case algFABM:
    out << "# Fano/BM decoding (delta = " << fano_delta_i
	<< ", maxnodes = " << maxnodes_i << ")\n";
    break;
  case algSTACK:
    out << "# Stack decoding (maxnodes = " << maxnodes_i << ")\n";
    break;
  case algSTBM:
    out << "# Stack/BM decoding (maxnodes = " << maxnodes_i << ")\n";
    break;
  case algBSTACK:
    out << "# Bidirectional stack decoding (maxnodes = " << maxnodes_i
	<< ")\n";
    break;
  case algBSTBM:
    out << "# Bidirectional Stack/BM decoding (maxnodes = " << maxnodes_i
	<< ")\n";
    break;
  case algCLIST:
    out << "# candidate list decoding (" << cl_num_i
	<< " lists of length " << cl_len_i << ")\n";
    break;
  case algPGMD:
    out << "# primitive GMD decoding\n";
    break;
  case algGMD:
    out << "# GMD decoding\n";
  }
  out << "# GMD check threshold: " << gmd_thresh_i << endl;
  out << "# vertical (outer) SXRS code: k = " << k_o
      << ", n = " << n_o
      << ", r = " << code_o->get_r()
      << ", xd = " << xd_o
      << ", dmin = " << d_o << endl;
  switch (algo_o) {
  case algFANO:
    out << "# Fano decoding (delta = " << fano_delta_o
	<< ", maxnodes = " << maxnodes_o << ")\n";
    break;
  case algFABM:
    out << "# Fano/BM decoding (delta = " << fano_delta_o
	<< ", maxnodes = " << maxnodes_o << ")\n";
    break;
  case algSTACK:
    out << "# Stack decoding (maxnodes = " << maxnodes_o
	<< ")\n";
    break;
  case algSTBM:
    out << "# Stack/BM decoding (maxnodes = " << maxnodes_o
	<< ")\n";
    break;
  case algBSTACK:
    out << "# Bidirectional stack decoding (maxnodes = " << maxnodes_o
	<< ")\n";
    break;
  case algBSTBM:
    out << "# Bidirectional Stack/BM decoding (maxnodes = " << maxnodes_o
	<< ")\n";
    break;
  case algCLIST:
    out << "# candidate list decoding (" << cl_num_o
	<< " lists of length " << cl_len_o << ")\n";
    break;
  case algPGMD:
    out << "# primitive GMD decoding\n";
    break;
  case algGMD:
    out << "# GMD decoding\n";
  }
  out << "# GMD check threshold: " << gmd_thresh_o << endl;
  out << "# total code length: " << n_i * n_o * code->get_n()
      << " channel uses\n";
  if (feedback > 0)
    out << "# " << feedback << " feedback iteration(s)\n";
  out << "# SNR\t SNR(dB)\t Eb/N0(dB)\t E+E\t ERAS\t WER\t SER\t BER";
  if ((algo_i >= algFANO) && (algo_i <= algGMD))
    out << "\t cplxty";
  out << "\t (E+E)'\t WER'\t SER'\t (E+E)\"\t SER\"";
  if ((algo_o >= algFANO) && (algo_o <= algGMD))
    out << "\t cplxty";
  out << endl;
}

void PrintOutputLine (ostream & out, double snr, double snrdb, double ebn0db)
{
  int symbols; // number of transmitted symbols
  symbols = loops * k_i * k_o;
  out << snr << '\t' << snrdb << '\t' << ebn0db << '\t'
      << (double)(eras+werr) / loops << '\t'
      << (double)eras / loops << '\t'
      << (double)werr / loops << '\t'
      << (double)serr / symbols << '\t'
      << 0.5 * serr / symbols;
  if (algo_i > 0)
    out << '\t' << cplxty_i / loops / n_o << '\t';
  out << '\t' << (double)(eras1+werr1) / loops / k_i
      << '\t' << (double)werr1 / loops / k_i
      << '\t' << (double)serr1 / symbols
      << '\t' << (double)(eras2+serr2) / symbols
      << '\t' << (double)serr2 / symbols;
  if (algo_o > 0)
    out << '\t' << (double)cplxty_o / loops / n_i;
  out << endl;
}

// DecodeGeometricalCode(pos) decodes the received signal vector src.
// The received symbol is written to dest[]
// and any soft-decision information is written to alpha[].

void DecodeGeometricalCode (const Vector & src, GFq & dest,
			    pDouble alpha, double & beta)
{
  int destsym;
  beta = code->LLHDecode (noise->variance(), src, destsym, alpha);
  dest = GFq(destsym);
}

int DecodeInnerCode (GFq * info, GFq * dest, const GFq * src,
		     const int * erasures, int erasnum,
		     const double * const * alpha,
		     const double * beta)
{
  int res, i;
  switch (algo_i) {
  case algBM:
    if (erasnum >= d_o) // too many erasures
      res = -20;
    else
      res = code_i->Decode (info, src, erasures, erasnum);
    break;
  case algFANO:
    res = treed_i->FanoDecode (info, src, alpha, fano_delta_i, maxnodes_i);
    break;
  case algFABM:
    res = treed_i->FanoDecode (info, src, alpha, fano_delta_i, maxnodes_i);
    if (res < 0)
      res = code_i->Decode (info, src);
    break;
  case algSTACK:
    res = treed_i->StackDecode (info, src, alpha);
    break;
  case algSTBM:
    res = treed_i->StackDecode (info, src, alpha);
    if (res < 0)
      res = code_i->Decode (info, src);
    break;
  case algBSTACK:
    res = treed_i->BidirStackDecode (info, src, alpha);
    break;
  case algBSTBM:
    res = treed_i->BidirStackDecode (info, src, alpha);
    if (res < 0)
      res = code_i->Decode (info, src);
    break;
  case algCLIST:
    res = gmdd_i->CandidateListDecode (info, src, alpha, beta,
				       cl_num_i, cl_len_i);
    break;
  case algPGMD:
    res = gmdd_i->PrimitiveDecode (info, src, beta);
    break;
  case algGMD:
    res = gmdd_i->Decode (info, src, alpha, beta);
    break;
  case algMVUC:
  case algSPECIAL:
    cerr << "Decoding algorithm " << algo_i << " is not yet implemented "
      "for horizontal (inner) code\n";
  default: exit(1);
  }
  if (res >= 0)
    code_i->Encode (dest, info);
  else {
    for (i = 0; i < n_i; i++)
      dest[i] = src[i];
  }
  if (algo_i >= algFANO && algo_i <= algGMD) {
    // update complexity statistics
    double cplxty;
    if (algo_i >= algFANO && algo_i <= algSTBM)
      cplxty = treed_i->GetNodes() / k_i; // # of nodes per info symbol
    else
      cplxty = gmdd_i->GetLoops();
    cplxty_i += cplxty;
  }
  return res;
}

int DecodeOuterCode (GFq * info, GFq * dest, const GFq * src,
		     const int * erasures, int erasnum,
		     const double * const * alpha,
		     const double * beta)
{
  int res, i;
  switch (algo_o) {
  case algBM:
    if (erasnum >= d_o) // too many erasures
      res = -20;
    else
      res = code_o->Decode (info, src, erasures, erasnum);
    break;
  case algFANO:
    res = treed_o->FanoDecode (info, src, alpha, fano_delta_o, maxnodes_o);
    break;
  case algFABM:
    res = treed_o->FanoDecode (info, src, alpha, fano_delta_o, maxnodes_o); //
    if (res < 0)
      res = code_o->Decode (info, src);
    break;
  case algSTACK:
    res = treed_o->StackDecode (info, src, alpha);
    break;
  case algSTBM:
    res = treed_o->StackDecode (info, src, alpha);
    if (res < 0)
      res = code_o->Decode (info, src);
    break;
  case algBSTACK:
    res = treed_o->BidirStackDecode (info, src, alpha);
    break;
  case algBSTBM:
    res = treed_o->BidirStackDecode (info, src, alpha);
    if (res < 0)
      res = code_o->Decode (info, src);
    break;
  case algCLIST:
    res = gmdd_o->CandidateListDecode (info, src, alpha, beta,
				       cl_num_o, cl_len_o);
    break;
  case algPGMD:
    res = gmdd_o->PrimitiveDecode (info, src, beta);
    break;
  case algGMD:
    res = gmdd_o->Decode (info, src, alpha, beta);
    break;
  case algMVUC:
  case algSPECIAL:
    cerr << "Decoding algorithm " << algo_o << " is not yet implemented "
      "for vertical (outer) code\n";
  default: exit(1);
  }
  if (res >= 0)
    code_o->Encode (dest, info);
  else {
    for (i = 0; i < n_o; i++)
      dest[i] = src[i];
  }
  if (algo_o >= algFANO && algo_o <= algGMD) {
    // update complexity statistics
    double cplxty;
    if (algo_o >= algFANO && algo_o <= algSTBM)
      cplxty = treed_o->GetNodes() / k_o; // # of nodes per info symbol
    else
      cplxty = gmdd_o->GetLoops();
    cplxty_o += cplxty;
  }
  return res;
}

// PrepareSDForOuterCode prepares the soft decision information
// for decoding the vertical (outer) SXRS code
// after having decoded the horizontal (inner) SXRS code for the first time

void PrepareSDForOuterCode ()
{
  int i,j;
  for (i = 0; i < n_i*n_o; i++) {
    if ((resbuf_i[i] < 0) || (relbuf_i[i] < gmd_thresh_i)) {
      codebuf_i[i] = codebuf[i];
      for (j = 0; j < GFq::q(); j++)
	alphabuf_i[i][j] = alphabuf[i][j];
      betabuf_i[i] = betabuf[i];
    }
    else { // (resbuf_i[i] >= 0) && (relbuf_i[i] >= gmd_thresh_i)
      for (j = 0; j < GFq::q(); j++)
	alphabuf_i[i][j] = -99.99;
      alphabuf_i[i][codebuf_i[i]] = 0;
      betabuf_i[i] = 99.99;
    }
  }
}

// PrepareFeedbackForInnerCode prepares codebuf_o[]
// for the next decoding iteration

void PrepareFeedbackForInnerCode()
{
  int i, j;
  for (i = 0; i < n_o; i++) {
    if ((resbuf_i[i*n_i] >= 0) && (relbuf_i[i*n_i] < gmd_thresh_i)) {
      // look if any outer code word conflicts with the i-th inner one
      for (j = 0; j < n_i; j++) {
	if ((resbuf_o[j] >= 0) && (codebuf_i[i*n_i+j] != codebuf_o[i*n_i+j]))
	  break;
      }
      if (j < n_i) {
	// there were conflicts
	for (j = 0; j < n_i; j++)
	  resbuf_i[i*n_i+j] = -99;
      }
    }
  }
  for (j = 0; j < n_i; j++) {
    if ((resbuf_o[j] >= 0) && (relbuf_o[j] < gmd_thresh_o)) {
      // look if any inner code word conflicts with the j-th outer one
      for (i = 0; i < n_o; i++) {
	if ((resbuf_i[i*n_i] >= 0) && (codebuf_i[i*n_i+j] != codebuf_o[i*n_i+j]))
	  break;
      }
      if (i < n_o) {
	// there were conflicts
	for (i = 0; i < n_o; i++)
	  resbuf_o[i*n_i+j] = -99;
      }
    }
  }
  for (i = 0; i < n_i*n_o; i++) {
    if (resbuf_i[i] >= 0) {
      codebuf_o[i] = codebuf_i[i];
      for (j = 0; j < GFq::q(); j++)
	alphabuf_o[i][j] = -99.99;
      alphabuf_o[i][codebuf_i[i]] = 0.0;
      betabuf_o[i] = 99.99;
    }
    else if (resbuf_o[i] >= 0) {
      for (j = 0; j < GFq::q(); j++)
	alphabuf_o[i][j] = -99.99;
      alphabuf_o[i][codebuf_o[i]] = 0.0;
      betabuf_o[i] = 99.99;
    }
    else {
      codebuf_o[i] = codebuf[i];
      for (j = 0; j < GFq::q(); j++)
	alphabuf_o[i][j] = alphabuf[i][j];
      betabuf_o[i] = betabuf[i];
    }
  }

  return;
  for (i = 0; i < n_i*n_o ; i++) {
    //     if ((((resbuf_i[i] < 0) || (relbuf_i[i] < gmd_thresh_i))
    // 	 && ((resbuf_o[i] < 0) || (relbuf_o[i] < gmd_thresh_o)))
    // 	|| ((resbuf_i[i] >= 0) && (relbuf_i[i] >= gmd_thresh_i)
    // 	    && (resbuf_o[i] >= 0) && (relbuf_o[i] >= gmd_thresh_o)
    // 	    && (codebuf_i[i] != codebuf_o[i])))
    if (((resbuf_i[i] < 0) || (relbuf_i[i] < gmd_thresh_i))
	&& ((resbuf_o[i] < 0) || (relbuf_o[i] < gmd_thresh_o))
	&& ((resbuf_i[i] < 0) || (resbuf_o[i] < 0)
	    || (resbuf_i[i] != resbuf_o[i])))
      {
	codebuf_o[i] = codebuf[i];
	for (j = 0; j < GFq::q(); j++)
	  alphabuf_o[i][j] = alphabuf[i][j];
	betabuf_o[i] = betabuf[i];
      }
    else {
      if (((resbuf_o[i] < 0) || (relbuf_o[i] < gmd_thresh_o))
	  || ((resbuf_o[i] >= 0) && (relbuf_o[i] >= gmd_thresh_o)
	      && (resbuf_i[i] >= 0) && (relbuf_i[i] >= gmd_thresh_i)
	      && (relbuf_i[i] > relbuf_o[i])))
	codebuf_o[i] = codebuf_i[i];
      for (j = 0; j < GFq::q(); j++)
	alphabuf_o[i][j] = -99.99;
      alphabuf_o[i][codebuf_o[i]] = 0.0;
      betabuf_o[i] = 99.99;
    }
  } // for (i)
}

// PrepareFeedbackForOuterCode prepares codebuf_i[]
// for the next decoding iteration

void PrepareFeedbackForOuterCode()
{
  int i, j;
  for (i = 0; i < n_i*n_o ; i++) {
//     if ((((resbuf_o[i] < 0) || (relbuf_o[i] < gmd_thresh_o))
// 	 && ((resbuf_i[i] < 0) || (relbuf_i[i] < gmd_thresh_i)))
// 	|| ((resbuf_o[i] >= 0) && (relbuf_o[i] >= gmd_thresh_o)
// 	    && (resbuf_i[i] >= 0) && (relbuf_i[i] >= gmd_thresh_i)
// 	    && (codebuf_o[i] != codebuf_i[i])))
    if (((resbuf_o[i] < 0) || (relbuf_o[i] < gmd_thresh_o))
	&& ((resbuf_i[i] < 0) || (relbuf_i[i] < gmd_thresh_i)))
      {
	codebuf_i[i] = codebuf[i];
	for (j = 0; j < GFq::q(); j++)
	  alphabuf_i[i][j] = alphabuf[i][j];
	betabuf_i[i] = betabuf[i];
      }
    else {
      if ((resbuf_i[i] < 0) || (relbuf_i[i] < gmd_thresh_i)
	  || ((resbuf_i[i] >= 0) && (relbuf_i[i] >= gmd_thresh_i)
	      && (resbuf_o[i] >= 0) && (relbuf_o[i] >= gmd_thresh_o)
	      && (relbuf_o[i] > relbuf_i[i])))
	codebuf_i[i] = codebuf_o[i];
      for (j = 0; j < GFq::q(); j++)
	alphabuf_i[i][j] = -99.99;
      alphabuf_i[i][codebuf_i[i]] = 0.0;
      betabuf_i[i] = 99.99;
    }
  } // for (i)
}

// ExtractInformation() extracts the information from codebuf_o[]
// into infobuf[]

void ExtractInformation()
{
  int i1,i2,i3,j1,j2,j3;
  i1 = k_i-1;
  i2 = n_i-1;
  i3 = (n_i - d_i + 1) % xd_i;
  while (i1 >= 0) {
    assert (i2 >= d_i - 1);
    if (i3 > 0) {
      j1 = k_o-1;
      j2 = n_o-1;
      j3 = (n_o - d_o + 1) % xd_o;
      while (j1 >= 0) {
	assert (j2 >= d_o - 1);
	if (j3 > 0) {
	  infobuf[i1*k_o + j1] = codebuf_o[j2*n_i + i2];
	  erasbuf[i1*k_o + j1] = (betabuf_o[j2*n_i + i2] < 99.99);
	  j1--; j2--; j3--;
	}
	else {
	  j2--;
	  j3 = xd_o - 1;
	}
      } // while (j3 > 0)
      i1--; i2--; i3--;
    }
    else { // i3 == 0
      i2--;
      i3 = xd_i - 1;
    }
  } // while (i3 > 0)
}

void UpdateErrorStatistics ()
{
  int i, j, _eras1, _werr1, _serr1, _eras2, _serr2;
  _eras1 = _werr1 = _serr1 = 0;
  for (i = 0; i < k_i; i++) {
    _eras2 = _serr2 = 0;
    for (j = 0; j < k_o; j++) {
      if (erasbuf[i*k_o + j])
	_eras2++;
      else if (infobuf[i*k_o + j] != infobuf0[i*k_o + j])
	_serr2++;
    }
    eras2 += _eras2;
    serr2 += _serr2;
    if (_eras2 > 0) // the i-th outer code word contains erasures
      _eras1++;
    else if (_serr2 > 0) { // the i-th outer code word contains errors
      _werr1++;
      _serr1 += _serr2;
    }
  }
  eras1 += _eras1;
  werr1 += _werr1;
  serr1 += _serr1;
  if (_eras1 > 0) // at least one symbol was erased
    eras++;
  else if (_werr1 > 0) { // at least one symbol is wrong
    werr++;
    serr += _serr1;
  }
}

void SimulateConcatenation ()
{
  int i, j, feed, res, erasnum;
  // fill info buffer with random symbols
  for (i = 0; i < k_i*k_o; i++)
    infobuf0[i] = (int) floor (GFq::q()*rg->asDouble());
#ifdef VERBOSE
  cerr << "### infobuf0:\n";
  for (i=0;i<k_i;i++) {
    for (j=0;j<k_o;j++)
      cerr << '\t' << (int) infobuf0[i*k_o+j];
    cerr << endl;
  }
#endif
  // apply vertical (outer) SXRS code
  for (i = 0; i < k_i; i++) {
    code_o->Encode (codeword_o, &infobuf0[i*k_o]);
    // interleave
    for (j = 0; j < n_o; j++)
      interbuf0[j*k_i + i] = codeword_o[j];
  }
#ifdef VERBOSE
  cerr << "### interbuf0:\n";
  for (i=0;i<n_o;i++) {
    for (j=0;j<k_i;j++)
      cerr << '\t' << (int) interbuf0[i*k_i+j];
    cerr << endl;
  }
#endif
  // apply horizontal (inner) SXRS code
  for (i = 0; i < n_o; i++)
    code_i->Encode (&codebuf0[i*n_i], &interbuf0[i*k_i]);
#ifdef VERBOSE
  cerr << "### codebuf0:\n";
  for (i=0;i<n_o;i++) {
    for (j=0;j<n_i;j++)
      cerr << '\t' << (int) codebuf0[i*n_i+j];
    cerr << endl;
  }
#endif
  for (i = 0; i < n_i*n_o; i++) {
    // transmit i-th code symbol
    sent[i].copy ((*code)[codebuf0[i]]);
    Transmission (sent[i], rcvd[i]);
    DecodeGeometricalCode (rcvd[i], codebuf[i],
			   alphabuf[i], betabuf[i]);
  }
#ifdef VERBOSE
  cerr << "### codebuf:\n";
  for (i=0;i<n_o;i++) {
    for (j=0;j<n_i;j++) {
      cerr << "\t " << (int) codebuf[i*n_i+j];
      if (codebuf0[i*n_i+j]!=codebuf[i*n_i+j]) cerr << '*';
    }
    cerr << endl;
    for (j=0;j<n_i;j++)
      (cerr << '\t').form ("%.2lf", betabuf[i*n_i+j]);
    cerr << endl;
  }
#endif
  // decode the horizontal (inner) SXRS code for the first time
  for (i = 0; i < n_o; i++) {
    res = DecodeInnerCode (codeword_i, &codebuf_i[i*n_i],
			   &codebuf[i*n_i], erasures, 0,
			   &alphabuf[i*n_i], &betabuf[i*n_i]);
    for (j = 0; j < n_i; j++)
      resbuf_i[i*n_i+j] = res;
    double rel =
      gmdd_i->Reliability (&codebuf_i[i*n_i], &codebuf[i*n_i],
			   (const double **) &alphabuf[i*n_i],
			   (const double *) &betabuf[i*n_i]) / d_i;
    for (j = 0; j < n_i; j++)
      relbuf_i[i*n_i+j] = rel;
#ifdef VERBOSE
    if (res >= 0) {
      cerr << rel << '\t';
      int correct=1;
      for (j=0; j<n_i; j++) {
	if (codebuf_i[i*n_i] != codebuf0[i*n_i]) {correct=0;break;}
      }
      cerr << correct << endl;
    }
#endif
  } // for(i)
  // decode the vertical (outer) SXRS code for the first time
  for (i = 0; i < n_i; i++) {
    // deinterleave and determine erasures
    erasnum = 0;
    for (j = 0; j < n_o; j++) {
      codeword0_o[j] = codebuf[j*n_i + i];
      alphaword[j] = alphabuf[j*n_i + i];
      betaword[j] = betabuf[j*n_i + i];
    }
    res = DecodeOuterCode (infoword_o, codeword_o,
			   codeword0_o, erasures, 0,
			   alphaword, betaword);
    for (j = 0; j < n_o; j++)
      codebuf_o[j*n_i + i] = codeword_o[j];
    for (j = 0; j < n_o; j++)
      resbuf_o[j*n_i + i] = res;
    double rel = gmdd_o->Reliability (codeword_o, codeword0_o,
				      (const double **) alphaword,
				      (const double *) betaword) / d_i;
    for (j = 0; j < n_o; j++)
      relbuf_o[j*n_i + i] = rel;
  } // for (i)
#ifdef VERBOSE
    cerr << "### codebuf_i/o:\n";
    for (i=0;i<n_o;i++) {
      for (j=0;j<n_i;j++) {
	cerr << "\t " << (int) codebuf_i[i*n_i+j];
	if (resbuf_i[i*n_i+j] < 0) cerr << '-';
	if (codebuf_i[i*n_i+j]!=codebuf0[i*n_i+j]) cerr << '*';
      }
      cerr << endl;
      for (j=0;j<n_i;j++) {
	cerr << "\t " << (int) codebuf_o[i*n_i+j];
	if (resbuf_o[i*n_i+j] < 0) cerr << '-';
	if (codebuf_o[i*n_i+j]!=codebuf0[i*n_i+j]) cerr << '*';
      }
      cerr << endl;
      for (j=0;j<n_i;j++)
	(cerr << '\t').form ("%.2lf", relbuf_i[i*n_i+j]);
      cerr << endl;
      for (j=0;j<n_i;j++)
	(cerr << '\t').form ("%.2lf", relbuf_o[i*n_i+j]);
      cerr << endl;
    }
#endif
  PrepareFeedbackForInnerCode ();
#ifdef VERBOSE
    cerr << "### codebuf_o:\n";
    for (i=0;i<n_o;i++) {
      for (j=0;j<n_i;j++) {
	cerr << "\t " << (int) codebuf_o[i*n_i+j];
	if (codebuf_o[i*n_i+j]!=codebuf0[i*n_i+j]) cerr << '*';
      }
      cerr << endl;
      for (j=0;j<n_i;j++)
	(cerr << '\t').form ("%.2lf", betabuf_o[i*n_i+j]);
      cerr << endl;
    }
#endif
  for (feed = 0; feed < feedback; feed++) {
    // decode horizontal (inner) SXRS code
    for (i = 0; i < n_o; i++) {
      // determine erasures
      erasnum = 0;
      for (j = 0; j < n_i; j++) {
	if (resbuf_o[i*n_i + j] < 0)
	  erasures[erasnum++] = j;
      }
      res = DecodeInnerCode (codeword_i, &codebuf_i[i*n_i],
			     &codebuf_o[i*n_i], erasures, erasnum,
			     &alphabuf_o[i*n_i], &betabuf_o[i*n_i]);
      for (j = 0; j < n_i; j++)
	resbuf_i[i*n_i+j] = res;
      double rel =
	gmdd_i->Reliability (&codebuf_i[i*n_i], &codebuf[i*n_i],
			     (const double **)&alphabuf[i*n_i],
			     (const double *)&betabuf[i*n_i]) / d_i;
      for (j = 0; j < n_i; j++)
	relbuf_i[i*n_i+j] = rel;
    } // for (i)
    // decode vertical (outer) SXRS code
    for (i = 0; i < n_i; i++) {
      // deinterleave and determine erasures
      erasnum = 0;
      for (j = 0; j < n_o; j++) {
	codeword0_o[j] = codebuf_o[j*n_i + i];
	if (resbuf_o[j*n_i + i] < 0)
	  erasures[erasnum++] = j;
	alphaword[j] = alphabuf_o[j*n_i + i];
	betaword[j] = betabuf_o[j*n_i + i];
      }
      res = DecodeOuterCode (infoword_o, codeword_o,
			      codeword0_o, erasures, erasnum,
			      alphaword, betaword);
      for (j = 0; j < n_o; j++)
	codebuf_o[j*n_i + i] = codeword_o[j];
      for (j = 0; j < n_o; j++)
	resbuf_o[j*n_i + i] = res;
      for (j = 0; j < n_o; j++) {
	codeword0_o[j] = codebuf[j*n_i + i];
	alphaword[j] = alphabuf[j*n_i + i];
	betaword[j] = betabuf[j*n_i + i];
      }
      double rel = gmdd_o->Reliability (codeword_o, codeword0_o,
					(const double **)alphaword,
					(const double *)betaword) / d_i;
      for (j = 0; j < n_o; j++)
	relbuf_o[j*n_i + i] = rel;
    } // for (i)
#ifdef VERBOSE
    cerr << "### codebuf_i/o:\n";
    for (i=0;i<n_o;i++) {
      for (j=0;j<n_i;j++) {
	cerr << "\t " << (int) codebuf_i[i*n_i+j];
	if (resbuf_i[i*n_i+j] < 0) cerr << '-';
	if (codebuf_i[i*n_i+j]!=codebuf0[i*n_i+j]) cerr << '*';
      }
      cerr << endl;
      for (j=0;j<n_i;j++) {
	cerr << "\t " << (int) codebuf_o[i*n_i+j];
	if (resbuf_o[i*n_i+j] < 0) cerr << '-';
	if (codebuf_o[i*n_i+j]!=codebuf0[i*n_i+j]) cerr << '*';
      }
      cerr << endl;
      for (j=0;j<n_i;j++)
	(cerr << '\t').form ("%.2lf", relbuf_i[i*n_i+j]);
      cerr << endl;
      for (j=0;j<n_i;j++)
	(cerr << '\t').form ("%.2lf", relbuf_o[i*n_i+j]);
      cerr << endl;
    }
#endif
    PrepareFeedbackForInnerCode ();
#ifdef VERBOSE
    cerr << "### codebuf_o:\n";
    for (i=0;i<n_o;i++) {
      for (j=0;j<n_i;j++) {
	cerr << "\t " << (int) codebuf_o[i*n_i+j];
	if (codebuf_o[i*n_i+j]!=codebuf0[i*n_i+j]) cerr << '*';
      }
      cerr << endl;
      for (j=0;j<n_i;j++)
	(cerr << '\t').form ("%.2lf", betabuf_o[i*n_i+j]);
      cerr << endl;
    }
#endif
  } // for (feed)
  ExtractInformation ();
#ifdef VERBOSE
  cerr << "### infobuf:\n";
  for (i=0;i<k_i;i++) {
    for (j=0;j<k_o;j++) {
      cerr << "\t " << (int) infobuf[i*k_o+j];
      if (erasbuf[i*k_o+j]) cerr << '-';
      if (infobuf[i*k_o+j]!=infobuf0[i*k_o+j]) cerr << '*';
    }
    cerr << endl;
  }
#endif
  UpdateErrorStatistics ();
}

void PrintErrorCurve (ostream & out)
{
  int s,i;
  // compute code rate
  double R = code->get_r() * code_i->get_r() * code_o->get_r();
  for (s=0; s<snr_num; s++) {
    // set SNR and Eb/N0
    double snr, snrdb, ebn0, ebn0db;
    if (snr_db) {
      snrdb = snr_start + s * snr_incr;
      snr = pow (10.0, snrdb/10.0);
    }
    else {
      snr = snr_start + s * snr_incr;
      snrdb = 10.0 * log10 (snr);
    }
    ebn0 = snr / (2*R);
    ebn0db = 10.0 * log10 (ebn0);
    noise->variance (1.0/(snr * code->get_n()));
    // init counters and statistics
    eras = werr = serr = eras1 = werr1 = 0;
    cplxty_i = cplxty_o = 0.0;
    // now do the actual simulation
    for (i = 0; i < loops; i++)
      SimulateConcatenation ();
    PrintOutputLine (out, snr, snrdb, ebn0db);
  }
}

void SetDefaults ()
{
  snr_start = 1.0;
  snr_incr = 0.1;
  snr_num = 1;
  snr_db = 0; // SNR values are not interpreted as dB
  loops = 1000;
  n_i = n_o = GFq::q() - 1;
  d_i = d_o = 5; // correct up to 2 errors
  xd_i = xd_o = 4;
  feedback = 0; // no feedback
  algo_i = algo_o = algBM;
  fano_delta_i = fano_delta_o = 4.0;
  maxnodes_i = maxnodes_o = 1000;
  cl_num_i = cl_num_o = 8;
  cl_len_i = cl_len_o = 2;
  gmd_thresh_i = gmd_thresh_o = -5.0;
}

char * usage = // "Usage: " << argv[0] <<
" [OPTION...] CODEFILE\n\n"
"-h   display this help and exit\n"
"-o STRING  set output filename\n"
"-s FLOAT   set initial SNR\n"
"-i FLOAT   set SNR increment\n"
"-n INT     set # of SNR values\n"
"-l INT     set # of loops per SNR value\n"
"-d         interpret SNR values as dB\n"
"-c INT[,INT[,INT]]  set parameters (dmin,xd,n) of horizontal (inner) code\n"
"-C INT[,INT[,INT]]  set parameters (dmin,xd,n) of vertical (outer) code\n"
"-f INT     use decoding feedback (parameter: # of iterations)\n"
"-a INT     set decoding algorithm for horizontal code:\n"
"   -a 0       Berlekamp-Massey (default)\n"
"   -a 1[,INT[,FLOAT]]  Fano (parameter: delta,maxnodes)\n"
"   -a 2[,INT]          Stack (parameter: maxnodes)\n"
"   -a 3[,INT[,FLOAT]]  Fano with BM in case of erasures\n"
"   -a 4[,INT]          Stack with BM in case of erasures\n"
"   -a 5[,INT]          BD-Stack (parameter: maxnodes)\n"
"   -a 6[,INT]          BD-Stack with BM in case of erasures\n"
"   -a 7[,INT[,INT]]    candidate lists (parameters: number,length)\n"
"   -a 8       primitive GMD\n"
"   -a 9       normal GMD\n"
"   -a 10      modified Vucetic\n"
"   -a 99      special\n"
"-A INT     set decoding algorithm for outer code (see -a)\n"
"-g FLOAT   set GMD check threshold for horizontal code\n"
"-G FLOAT   set GMD check threshold for outer code\n";

int main (int argc, char ** argv)
{
  int opt;
  int i1,i2,i3;
  char dummy;
  double d1;
  ostream * out = &cout;

  SetDefaults ();
  while ((opt = getopt (argc, argv, "ho:s:i:n:l:dc:C:f:a:A:g:G:")) != EOF) {
    switch (opt) {

    case 'h':
      cout << "Usage: " << argv[0] << usage;
      return 0;

    case 'o':
      out = new ofstream (optarg);
      if (out->fail ()) {
	cerr << argv[0] << ": error opening output file " << optarg << "\n";
	return 1;
      }
      break;

    case 's':
      if (sscanf (optarg, "%lf%c", &d1, &dummy) != 1) {
	cerr << argv[0] << ": malformed parameters for option -s: "
	     << optarg << endl;
	return 1;
      }
      snr_start = d1;
      break;

    case 'i':
      if (sscanf (optarg, "%lf%c", &d1, &dummy) != 1) {
	cerr << argv[0] << ": malformed parameters for option -i: "
	     << optarg << endl;
	return 1;
      }
      snr_incr = d1;
      break;

    case 'n':
      if (sscanf (optarg, "%d%c", &i1, &dummy) != 1) {
	cerr << argv[0] << ": malformed parameters for option -n: "
	     << optarg << endl;
	return 1;
      }
      if (i1 < 1) {
	cerr << argv[0] << ": option -n: number of SNR values must be positive"
	  " (was " << i1 << ")\n";
	return 1;
      }
      snr_num = i1;
      break;

    case 'l':
      if (sscanf (optarg, "%d%c", &i1, &dummy) != 1) {
	cerr << argv[0] << ": malformed parameters for option -l: "
	     << optarg << endl;
	return 1;
      }
      if (i1 < 1) {
	cerr << argv[0] << ": option -l: number of loops must be positive"
	  " (was " << i1 << ")\n";
	return 1;
      }
      loops = i1;
      break;

    case 'd':
      snr_db = 1; // interpret SNR values as dB
      break;

    case 'c':
      // parameters of horizontal (inner) code
      switch (sscanf (optarg, "%d,%d,%d%c", &i1, &i2, &i3, &dummy)) {
      case 3:
	if (i3 < 1 || i3 >= GFq::q()) {
	  cerr << argv[0] <<
	    ": option -c: length must be between 1 and " << GFq::q()-1 <<
	    " (was " << i3 << ")\n";
	  return 1;
	}
	n_i = i3;
      case 2:
	if (i2 <= 1) {
	  cerr << argv[0] <<
	    ": option -c: expurgation distance must be bigger than 1"
	    " (was " << i2 << ")\n";
	  return 1;
	}
	xd_i = i2;
      case 1:
	if (i1 < 1) {
	  cerr << argv[0] <<
	    ": option -c: code distance must be positive"
	    " (was " << i1 << ")\n";
	  return 1;
	}
	d_i = i1;
	break;
      default:
	cerr << argv[0] << ": malformed parameters for option -c: "
	     << optarg << "\n";
	return 1;
      }
      break;

    case 'C':
      // parameters of vertical (outer) code
      switch (sscanf (optarg, "%d,%d,%d%c", &i1, &i2, &i3, &dummy)) {
      case 3:
	if (i3 < 1 || i3 >= GFq::q()) {
	  cerr << argv[0] <<
	    ": option -C: length must be between 1 and " << GFq::q()-1 <<
	    " (was " << i3 << ")\n";
	  return 1;
	}
	n_o = i3;
      case 2:
	if (i2 <= 1) {
	  cerr << argv[0] <<
	    ": option -C: expurgation distance must be bigger than 1"
	    " (was " << i2 << ")\n";
	  return 1;
	}
	xd_o = i2;
      case 1:
	if (i1 < 1) {
	  cerr << argv[0] <<
	    ": option -C: code distance must be positive"
	    " (was " << i1 << ")\n";
	  return 1;
	}
	d_o = i1;
	break;
      default:
	cerr << argv[0] << ": malformed parameters for option -C: "
	     << optarg << "\n";
	return 1;
      }
      break;

    case 'f':
      // number of feedback iterations
      if (sscanf (optarg, "%d%c", &i1, &dummy) != 1) {
	cerr << argv[0] << ": malformed parameters for option -f: "
	     << optarg << endl;
	return 1;
      }
      if (i1 < 1) {
	cerr << argv[0] << ": option -f: number of iterations must be positive"
	  " (was " << i1 << ")\n";
	return 1;
      }
      feedback = i1;
      break;

    case 'a':
      // decoding algorithm for horizontal (inner) code
      switch (sscanf (optarg, "%d,%d,%lf%c", &i1, &i2, &d1, &dummy)) {
      case 3:
	if (i1 == algCLIST)
	  cl_len_i = (int) floor (d1 + .5);
	else if (i1 == algFANO || i1 == algFABM)
	  fano_delta_i = d1;
	else
	  cerr << argv[0] << ": option -a: ignoring second parameter "
	       << d1 << endl;
      case 2:
	switch (i1) {
	case algFANO:
	case algFABM:
	case algSTACK:
	case algSTBM:
	case algBSTACK:
	case algBSTBM:
	  if (i2 < 1) {
	    cerr << argv[0] << ": option -a: parameter maxnodes "
	      "must be positive (was " << i2 << ")\n";
	    return 1;
	  }
	  maxnodes_i = i2;
	  break;
	case algCLIST:
	  if (i2 < 1) {
	    cerr << argv[0] << ": option -a: number of lists "
	      "must be positive (was " << i2 << ")\n";
	    return 1;
	  }
	  cl_num_i = i2;
	  break;
	default:
	  cerr << argv[0] << ": option -a: ignoring first parameter "
	       << i2 << "\n";
	} // switch (i1)
      case 1:
	if ((i1 < algBM || i1 > algMVUC) && (i1 != algSPECIAL)) {
	  cerr << argv[0] <<
	    ": option -a: invalid algorithm ID " << i1 << ")\n";
	  return 1;
	}
	algo_i = i1;
	break;
      default:
	cerr << argv[0] << ": malformed parameters for option -a: "
	     << optarg << "\n";
	return 1;
      }
      break;

    case 'A':
      // decoding algorithm for vertical (outer) code
      switch (sscanf (optarg, "%d,%d,%lf%c", &i1, &i2, &d1, &dummy)) {
      case 3:
	if (i1 == algCLIST)
	  cl_len_o = (int) floor (d1 + .5);
	else if (i1 == algFANO || i1 == algFABM)
	  fano_delta_o = d1;
	if (i1 != algFANO && i1 != algFABM)
	  cerr << argv[0] << ": option -A: ignoring parameter delta = "
	       << d1 << "\n";
      case 2:
	switch (i1) {
	case algFANO:
	case algFABM:
	case algSTACK:
	case algSTBM:
	case algBSTACK:
	case algBSTBM:
	  if (i2 < 1) {
	    cerr << argv[0] << ": option -A: parameter maxnodes "
	      "must be positive (was " << i2 << ")\n";
	    return 1;
	  }
	  maxnodes_o = i2;
	  break;
	case algCLIST:
	  if (i2 < 1) {
	    cerr << argv[0] << ": option -A: number of lists "
	      "must be positive (was " << i2 << ")\n";
	    return 1;
	  }
	  cl_num_o = i2;
	  break;
	default:
	  cerr << argv[0] << ": option -A: ignoring first parameter "
	       << i2 << "\n";
	} // switch (i1)
      case 1:
	if ((i1 < algBM || i1 > algMVUC) && (i1 != algSPECIAL)) {
	  cerr << argv[0] <<
	    ": option -A: invalid algorithm ID " << i1 << ")\n";
	  return 1;
	}
	algo_o = i1;
	break;
      default:
	cerr << argv[0] << ": malformed parameters for option -A: "
	     << optarg << "\n";
	return 1;
      }
      break;

    case 'g':
      if (sscanf (optarg, "%lf%c", &d1, &dummy) != 1) {
	cerr << argv[0] << ": malformed parameters for option -g: "
	     << optarg << "\n";
	return 1;
      }
      gmd_thresh_i = d1;
      break;

    case 'G':
      if (sscanf (optarg, "%lf%c", &d1, &dummy) != 1) {
	cerr << argv[0] << ": malformed parameters for option -G: "
	     << optarg << "\n";
	return 1;
      }
      gmd_thresh_o = d1;
      break;

    } // switch
  } // while

  // read geometrical code file
  if (optind >= argc) {
    cerr << "Usage: " << argv[0] << usage;
    return 1;
  }
  if (strcmp (argv[optind], "-")) {
    ifstream in (argv[optind]);
    if (in.fail()) {
      cerr << argv[0] << ": unable to open code file "
	   << argv[optind] << endl;
      return 1;
    }
    code = new Code (in);
    in.close ();
    strncpy (codefilename, argv[optind], 256);
    codefilename[255] = 0;
  }
  else {
    code = new Code (cin);
    strcpy (codefilename, "<stdin>");
  }

  if (++optind < argc) {
    cerr << argv[0] << ": ignoring extra argument(s):";
    while (optind < argc)
      cerr << ' ' << argv[optind++];
    cerr << endl;
  }

  // do some consistency checks
  if (code->get_m() != GFq::q()) {
    cerr << argv[0] << ": geometrical code of size " << code->get_m()
	 << " isn't compatible with GF(" << GFq::q() << ")\n";
    return 1;
  }
  if (d_i > n_i) {
    cerr << argv[0] << ": distance of horizontal (inner) code (" << d_i
	 << ") must not be bigger than its length (" << n_i
	 << ")\n";
    return 1;
  }
  if (d_o > n_o) {
    cerr << argv[0] << ": distance of vertical (outer) code (" << d_o
	 << ") must not be bigger than its length (" << n_o
	 << ")\n";
    return 1;
  }

  // normalize the avg. codeword energy to 1
  *code *= sqrt (1.0 / code->AvgEnergy ());
  // initialize random generator
  rg = new MLCG ((_G_int32_t) time (NULL), (_G_int32_t) time (NULL));
  noise = new Normal<MLCG> (rg);
  InitCodes ();
  PrintHeader (*out);
  PrintErrorCurve (*out);
  ClearCodes ();
  delete noise;
  delete rg;
  delete code;
}
