// Calculation of bit error probabilities of geometrical codes
// on a simulated time-discrete AWGN channel

// (C) 1997-1998 Frank Paehlke <frank@paehlke.de>
// $Id: pe-geocode.cc,v 1.1 2003/10/14 08:59:11 frank Exp $

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// $Log: pe-geocode.cc,v $
// Revision 1.1  2003/10/14 08:59:11  frank
// moved from . to src/
//
// Revision 1.5  2003/10/14 08:41:58  frank
// ANSI C++ / gcc 3.x compatibility changes
//
// Revision 1.4  2002/08/02 14:45:06  paehlke
// added GPL comment
//
// Revision 1.3  2002/08/02 12:30:33  paehlke
// Imported from RCS to CVS
// On 2000-08-10:
// cmp_symindex: arguments changed from "int *" to "const void *"
//
// Revision 1.2  1998/05/28 12:56:36  paehlke
// Final version for masters thesis
//
// Revision 1.1  1997/06/20 18:08:11  frank
// Initial revision
//

#include <math.h>
#include <fstream>
#include <iostream>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>

#include "mlcg.h"
#include "normal.h"
#include "code.h"

using namespace std;

class Symbol
{
public:
  int err1, err2, err3;
  double beta;
};
typedef Symbol * pSymbol;

Code * code;

MLCG * rg; // random generator
Normal<MLCG> * noise;

double snr_start;   // initial SNR value
double snr_incr;    // SNR increment
int snr_num;        // number of SNR values
int snr_db;         // interpret SNR values as dB?
int loops;          // number of loops per SNR value
int d_func;         // distance function
int algo;           // output style
double eras_thresh; // erasure threshold
int cl2_num;        // # of candidate lists of length 2
int cl3_num;        // # of candidate lists of length 3
int wordlen;        // word length for candidate list decoding

char codefilename [256];

Vector recv;     // received signal vector
Symbol * sym;
int * symindex;
int * cbuf_orig; // original code buffer
int * cbuf;      // decoder code buffer
int eras, werr, serr; // number of erasures/word errors/symbol errors
int serr0, serr1;

void Transmission (const Vector & sent, Vector & recv)
{
  int i;
  for (i=0; i<code->get_n(); i++) {
    recv[i] = sent[i] + (*noise)();
  }
}

void InitCode ()
{
  int i;
  recv.allocate (code->get_n());
  sym = new Symbol [wordlen];
  symindex = new int [wordlen];
  for (i = 0; i < wordlen; i++)
    symindex[i] = i;
  cbuf_orig = new int [wordlen];
  cbuf = new int [wordlen];
}

void ClearCodes ()
{
  delete [] cbuf;
  delete [] cbuf_orig;
  delete [] symindex;
  delete [] sym;
  recv.deallocate();
}

void PrintHeader (ostream & out)
{
  out << "# Symbol error probabilities for geometrical code "
      << codefilename << ":\n"
      << "m = " << code->get_m() << ", n = " << code->get_n()
      << ", r = " << code->get_r()
      << ", dmin = " << code->MinDist() << endl
      << "# " << loops << " loops per SNR value\n";
  switch (algo) {
  case 2:
    out << "# "	<< cl2_num << " candidate lists of length 2\n"
	<< "# word length: " << wordlen << endl;
    break;
  case 3:
    out << "# "	<< cl3_num << " candidate lists of length 3\n"
	<< "# word length: " << wordlen << endl;
  }
  out << "# SNR\t SNR(dB)\t Eb/N0(dB)\t ";
  switch (algo) {
  case 0:
    out << "SER";
    break;
  case 1:
    out << "E+E\t ERAS\t SER";
    break;
  case 2:
  case 3:
    out << "WER\t SER\t SER0\t SER1";
  } // switch
  out << endl;
}

void PrintOutputLine (ostream & out, double snr, double snrdb, double ebn0db)
{
  int symbols = loops * code->get_m(); // number of transmitted symbols
  out << snr << '\t' << snrdb << '\t' << ebn0db << '\t';
  switch (algo) {
  case 0:
    out << (double) serr / symbols;
    break;
  case 1:
    out << (double)(eras+serr) / symbols << '\t'
	<< (double)eras / symbols << '\t'
	<< (double)serr / symbols;
    break;
  case 2:
  case 3:
    out << (double)werr / symbols << '\t'
	<< (double)serr / loops << '\t'
	<< (double)serr0 / loops << '\t'
	<< (double)serr1 / loops << '\t';
  } // switch
  out << endl;
}

void ErrorCurve (ostream & out)
{
  int s,i,j;
  double R = code->get_r(); // code rate
  for (s=0; s<snr_num; s++) {
    double snr, snrdb, ebn0db;
    if (snr_db) {
      snrdb = snr_start + s * snr_incr;
      snr = pow (10.0, snrdb/10.0);
    }
    else {
      snr = snr_start + s * snr_incr;
      snrdb = 10.0 * log10 (snr);
    }
    ebn0db = 10.0 * log10 (snr/(2*R));
    noise->variance (1.0/(snr * code->get_n()));
    serr = 0;
    for (i = 0; i < loops; i++) {
      for (j = 0; j < code->get_m(); j++) {
	Transmission ((*code)[j], recv);
	int sym_recv;
	code->Decode (recv, sym_recv);
	cbuf[j] = sym_recv;
	if (sym_recv != j)
          serr++;
      }
    }
    PrintOutputLine (out, snr, snrdb, ebn0db);
  }
}

void ErrorAndErasureCurve (ostream & out)
{
  int s,i,j;
  double R = code->get_r(); // code rate
  for (s=0; s<snr_num; s++) {
    double snr, snrdb, ebn0db;
    if (snr_db) {
      snrdb = snr_start + s * snr_incr;
      snr = pow (10.0, snrdb/10.0);
    }
    else {
      snr = snr_start + s * snr_incr;
      snrdb = 10.0 * log10 (snr);
    }
    ebn0db = 10.0 * log10 (snr/(2*R));
    noise->variance (1.0/(snr * code->get_n()));
    eras = serr = 0;
    for (i=0; i<loops; i++) {
      for (j=0; j<code->get_m(); j++) {
        Transmission ((*code)[j], recv);
	int sym_recv;
	double beta;
	switch (d_func) {
	case 1: beta = code->DecodeSpherical (recv, sym_recv); break;
	case 2:
	  beta = code->DecodeAPP (noise->variance(), recv, sym_recv);
	  break;
	default: beta = code->Decode (recv, sym_recv);
	}
	if (beta > eras_thresh)
	  eras++;
	else if (sym_recv != j)
          serr++;
      }
    }
    PrintOutputLine (out, snr, snrdb, ebn0db);
  }
}

extern "C" {
  int cmp_symindex (const void * _i, const void * _j)
  {
    int i = * (int *) _i;
    int j = * (int *) _j;
    if (sym[i].beta > sym[j].beta)
      return -1;
    else
      return (sym[i].beta < sym[j].beta);
  }
	   } // extern "C"

void ErrorCurve2 (ostream & out)
{
  int s,i,j;
  double R = code->get_r(); // code rate
  for (s=0; s<snr_num; s++) {
    double snr, snrdb, ebn0db;
    if (snr_db) {
      snrdb = snr_start + s * snr_incr;
      snr = pow (10.0, snrdb/10.0);
    }
    else {
      snr = snr_start + s * snr_incr;
      snrdb = 10.0 * log10 (snr);
    }
    ebn0db = 10.0 * log10 (snr/(2*R));
    noise->variance (1.0/(snr * code->get_n()));
    eras = werr = serr = serr0 = serr1 = 0;
    for (i = 0; i < loops; i++) {
      for (j = 0; j < wordlen; j++) {
	cbuf_orig[j] = (int) floor (code->get_m() * rg->asDouble());
	Transmission ((*code)[cbuf_orig[j]], recv);
	int sym1, sym2, sym3;
	switch (d_func) {
	case 1:
	  sym[j].beta = code->Decode3Spherical (recv, sym1, sym2, sym3);
	  break;
	case 2:
	  sym[j].beta = code->Decode3APP
	    (noise->variance(), recv, sym1, sym2, sym3);
	  break;
	default:
	  sym[j].beta = code->Decode3 (recv, sym1, sym2, sym3);
	}
	if (sym1 != cbuf_orig[j]) {
	  sym[j].err1++;
	  serr0++;
	  if (sym2 != cbuf_orig[j]) {
	    sym[j].err2++;
	    serr1++;
	  }
	}
      }
      qsort (symindex, code->get_m()-1, sizeof(*symindex), cmp_symindex);
      int _serr = serr;
      for (j = 0; j < cl2_num; j++) {
	if (sym[symindex[j]].err2)
	  serr++;
      }
      for (; j < wordlen; j++) {
	if (sym[symindex[j]].err1)
	  serr++;
      }
      if (serr > _serr)
	werr++;
    } // for (i)
    PrintOutputLine (out, snr, snrdb, ebn0db);
  }
}

void ErrorCurve3 (ostream & out)
{
  int s,i,j;
  double R = code->get_r(); // code rate
  for (s = 0; s < snr_num; s++) {
    double snr, snrdb, ebn0db;
    if (snr_db) {
      snrdb = snr_start + s * snr_incr;
      snr = pow (10.0, snrdb/10.0);
    }
    else {
      snr = snr_start + s * snr_incr;
      snrdb = 10.0 * log10 (snr);
    }
    ebn0db = 10.0 * log10 (snr/(2*R));
    noise->variance (1.0/(snr * code->get_n()));
    eras = werr = serr = serr0 = serr1 = 0;
    for (i = 0; i < loops; i++) {
      for (j = 0; j < wordlen; j++) {
	cbuf_orig[j] = (int) floor (code->get_m() * rg->asDouble());
	Transmission ((*code)[cbuf_orig[j]], recv);
	int sym1, sym2, sym3;
	switch (d_func) {
	case 1:
	  sym[j].beta = code->Decode3Spherical (recv, sym1, sym2, sym3);
	  break;
	case 2:
	  sym[j].beta = code->Decode3APP
	    (noise->variance(), recv, sym1, sym2, sym3);
	  break;
	default:
	  sym[j].beta = code->Decode3 (recv, sym1, sym2, sym3);
	}
	if (sym1 != cbuf_orig[j]) {
	  sym[j].err1++;
	  serr0++;
	  if (sym2 != cbuf_orig[j]) {
	    sym[j].err2++;
	    if (sym3 != cbuf_orig[j]) {
	      sym[j].err3++;
	      serr1++;
	    }
	  }
	}
      }
      qsort (symindex, code->get_m()-1, sizeof(*symindex), cmp_symindex);
      int _serr = serr;
      for (j = 0; j < cl3_num; j++) {
	if (sym[symindex[j]].err3)
	  serr++;
      }
      for (; j<code->get_m()-1; j++) {
	if (sym[symindex[j]].err1)
	  serr++;
      }
      if (serr > _serr)
	werr++;
    } // for (i)
    PrintOutputLine (out, snr, snrdb, ebn0db);
  }
}

void SetDefaults ()
{
  snr_start = 1.0;
  snr_incr = 0.1;
  snr_num = 1;
  snr_db = 0; // SNR values are not interpreted as dB
  loops = 1000;
  d_func = 0; // use euclidean distance for geometrical decoding
  algo = 0;   // output errors only
  eras_thresh = 0.5;
  cl2_num = 8;
  cl3_num = 5;
  wordlen = 15;
}

char * usage = // "Usage: " << argv[0] <<
" [OPTION...] CODEFILE\n\n"
"-h   display this help and exit\n"
"-o STRING  set output filename\n"
"-s FLOAT   set initial SNR\n"
"-i FLOAT   set SNR increment\n"
"-n INT     set # of SNR values\n"
"-l INT     set # of loops per SNR value\n"
"-d         interpret SNR values as dB\n"
"-f INT     set distance function"
"   -f 0       euclidean distance (default)"
"   -f 1       spherical distance"
"   -f 2       a-posteriori probability"
"-a INT     set output style:\n"
"   -a 0        errors (default)\n"
"   -a 1,FLOAT  errors and erasures (parameter: threshold)\n"
"   -a 2,INT[,INT]  errors for candidate lists of length 2\n"
"                   (parameters: # of lists, word length)"
"   -a 3,INT[,INT]  errors for candidate lists of length 3"
"                   (parameters: # of lists, word length)";

int main (int argc, char ** argv)
{
  int opt;
  int i1,i2;
  char dummy;
  double d1;
  ostream * out = &cout;

  SetDefaults ();
  while ((opt = getopt (argc, argv, "ho:s:i:n:l:df:a:")) != EOF) {
    switch (opt) {

    case 'h':
      cout << "Usage: " << argv[0] << usage;
      return 0;

    case 'o':
      out = new ofstream (optarg);
      if (out->fail ()) {
	cerr << argv[0] << ": error opening output file " << optarg << "\n";
	return 1;
      }
      break;

    case 's':
      if (sscanf (optarg, "%lf%c", &d1, &dummy) != 1) {
	cerr << argv[0] << ": malformed parameters for option -s: "
	     << optarg << endl;
	return 1;
      }
      snr_start = d1;
      break;

    case 'i':
      if (sscanf (optarg, "%lf%c", &d1, &dummy) != 1) {
	cerr << argv[0] << ": malformed parameters for option -i: "
	     << optarg << endl;
	return 1;
      }
      snr_incr = d1;
      break;

    case 'n':
      if (sscanf (optarg, "%d%c", &i1, &dummy) != 1) {
	cerr << argv[0] << ": malformed parameters for option -n: "
	     << optarg << endl;
	return 1;
      }
      if (i1 < 1) {
	cerr << argv[0] << ": option -n: number of SNR values must be positive"
	  " (was " << i1 << ")\n";
	return 1;
      }
      snr_num = i1;
      break;

    case 'l':
      if (sscanf (optarg, "%d%c", &i1, &dummy) != 1) {
	cerr << argv[0] << ": malformed parameters for option -l: "
	     << optarg << endl;
	return 1;
      }
      if (i1 < 1) {
	cerr << argv[0] << ": option -l: number of loops must be positive"
	  " (was " << i1 << ")\n";
	return 1;
      }
      loops = i1;
      break;

    case 'd':
      snr_db = 1; // interpret SNR values as dB
      break;

    case 'f':
      // distance function
      if (sscanf (optarg, "%d%c", &i1, &dummy) != 1) {
	cerr << argv[0] << ": malformed parameters for option -f: "
	     << optarg << endl;
	return 1;
      }
      if (i1 < 0 || i1 > 2) {
	cerr << argv[0] <<
	  ": option -f: function ID must be between 0 and 2"
	  " (was " << i1 << ")\n";
	return 1;
      }
      d_func = i1;
      break;

    case 'a':
      // set output style
      switch (sscanf (optarg, "%d,%lf,%d%c", &i1, &d1, &i2, &dummy)) {
      case 3:
      	if ((i1 == 2) || (i1 == 3)) {
	  if (i1 < 1) {
	    cerr << argv[0] <<
	      ": option -a: word length must be positive"
	      " (was " << i1 << ")\n";
	    return 1;
	  }
	  wordlen = i1;
	}
	else
	  cerr << argv[0] << ": option -a: ignoring second parameter "
	       << i2 << "\n";
      case 2:
	if (i1 == 1)
	  eras_thresh = d1;
	else if (i1 == 2)
	  cl2_num = (int) floor (d1+.5);
	else if (i1 == 3)
	  cl3_num = (int) floor (d1+.5);
	else
	  cerr << argv[0] << ": option -a: ignoring fkrst parameter "
		 << d1 << "\n";
      case 1:
	if (i1 < 0 || i1 > 3) {
	  cerr << argv[0] <<
	    ": option -a: algorithm ID must be between 0 and 3"
	    " (was" << i1 << ")\n";
	  return 1;
	}
	algo = i1;
	break;
      default:
	cerr << argv[0] << ": malformed parameters for option -a: "
	     << optarg << "\n";
	return 1;
      }
      break;
    } // switch
  } // while

  // read code file
  if (optind >= argc) {
    cerr << "Usage: " << argv[0] << usage;
    return 1;
  }
  if (strcmp (argv[optind], "-")) {
    ifstream in (argv[optind]);
    if (in.fail()) {
      cerr << argv[0] << ": unable to open code file "
	   << argv[optind] << endl;
      return 1;
    }
    code = new Code (in);
    in.close ();
    strncpy (codefilename, argv[optind], 256);
    codefilename[255] = 0;
  }
  else {
    code = new Code (cin);
    strcpy (codefilename, "<stdin>");
  }

  if (++optind < argc) {
    cerr << argv[0] << ": ignoring extra argument(s):";
    while (optind < argc)
      cerr << ' ' << argv[optind++];
    cerr << endl;
  }

  // normalize the avg. codeword energy to 1
  *code *= sqrt (1.0 / code->AvgEnergy ());
  // initialize random generator
  rg = new MLCG ((_G_int32_t) time (NULL), (_G_int32_t) time (NULL));
  noise = new Normal<MLCG> (rg);
  InitCode ();
  PrintHeader (*out);
  switch (algo) {
  case 0: ErrorCurve (*out); break;
  case 1: ErrorAndErasureCurve (*out); break;
  case 2: ErrorCurve2 (*out); break;
  case 3: ErrorCurve3 (*out); break;
  default: exit(1);
  }
  ClearCodes ();
  delete noise;
  delete rg;
  delete code;
}
